' $Id$
' $Source$
' $Copyright$


Option Explicit

On Error Resume Next

' -----------------------------------------------------------------------------
' Note: Functions GetFcpTargetMapping(), WWNToString() and Hex0() are adapted or
' copied directly from the following paper:
' http://download.microsoft.com/download/9/c/5/9c5b2167-8017-4bae-9fde-d599bac8184a/FCTopology.doc
' -----------------------------------------------------------------------------


' ---------------- Begin Global Variables and Constants -----------------------
'
' !!! BUMP THIS VERSION NUMBER EVERY TIME A CHANGE IS MADE TO THIS SCRIPT THAT
' MAKES ITS GENERATED DATA INCOMPATIBLE WITH THE PARSER THAT WILL CONSUME IT.
DIM strDataVersion
strDataVersion = 1.7

' Set the locale to US English (1033).  This is primarily important when
' synthesizing the filename to use when the /f <filename> argument was not
' provided.  Save the previous one to restore at the end for symmetry, even
' though SetLocale() only changes the locale for this instance of the
' scripting engine.
Const US_English = 1033

' strAllStart tracks when the script began running.  strAllEnd will note
' when it finishes.  intAllDiff will compute the difference, and be written
' into the data file.
Dim strAllStart, strAllEnd, intAllDiff
strAllStart = now

Dim intOrigLocale
intOrigLocale = SetLocale(US_English)
' Wscript.Echo "origLocale: '" & intOrigLocale & "'"


DIM strProgName, hntoks
strProgName = WScript.ScriptFullName
' split out the path, and set strProgName to the final component (filename)
hntoks = Split(strProgName, "\")
strProgName = hntoks(UBound(hntoks))


' objOutFile holds the TextStream object to which we write all our output. 
Dim objOutFile

' If RHWindowsScript.vbs is invoked without the /f <filename> flag, it
' synthesizes the filename to which it writes its data.  One component of
' that filename is the hostname.  
' Initially, it synthesizes the filename using the the value of the /s <system>
' flag if specified, or the output of the "hostname.exe" call if running
' locally, i.e., no /s <system> flag was specified.  But the output of
' hostname.exe is a short hostname, and the user may specify a short hostname
' with the /s flag.  But as we fetch the IP addresses, we invoke nslookup
' with each one, mapping it to its FQHN.  So IFF the /f flag wasn't specified,
' we attempt to match the original hostname to a FQHN, and if so, synthesize
' a new filename using that FQHN.  
' Then, at the end of execution, we rename the data file to the new name.
' strFQHNFileName holds the new name, if one is created, and 
' strFQHNPathnFileName holds the path and FQHN filename.
Dim strFQHNFileName, strFQHNPathnFileName
strFQHNFileName = ""
strFQHNPathnFileName = ""

' blnGotFFlag and blnGotDFlag indicate whether the /f and /d flags, resp., were
' specified on the command line.  We use blnGotFFlag in above algorithm, and 
' both flags with probe (/i).
Dim blnGotFFlag, blnGotDFlag
blnGotFFlag = False
blnGotDFlag = False

' When a user configures a remote host, the Collector checks a variety of
' factors to determine whether the host can be Explored; i.e., it performs
' a *probe*.  These checks include verification of account and credential,
' WMI access to both CIMV2 and WMI namespaces, etc.  
' The /i flag specifies that RHWindowsScript.vbs is to do a probe, and that
' flag sets blnProbe to True.
Dim blnProbe
blnProbe = False

' strSystem holds the hostname which we will explore.  Default is "." (System
'           on which the script is running).
' strUser holds the username of the account we'll use.
' strPassword holds the password of that account.
' strOutDir holds the directory path under which we'll write our output file.
'           Default is ".\" (current working directory).
' strOutFile holds the filename specified for the output file.  This is for
'           debugging only.
Dim strSystem, strUser, strPassword, strOutDir, strOutFile, strSystemDir

' Assume defaults for local execution (i.e., fetching WMI data from the 
' host on which the script is running).  This means "." for system, and not
' specifying username or password.  Also assume the CWD for writing the output
' file.
strSystem = "."
strUser = ""
strPassword = ""
strOutDir = ".\"
strOutFile = ""
' strSystemDir is the value of %SystemRoot% on remote hosts, as fetched by WMI.
strSystemDir = ""

' If we are Exploring Windows 2000 or 2003, we use fcinfo to get the target
' mappings.  Otherwise (Windows 2008 and newer), we use WMI. GetOSType() uses
' information from the Win32_GetOperatingSystem class to set "strOSType" to
' the numeric value of the OS (2000, 2003, 2008), and "blnUsefcinfo" to True
' IFF 2003 or 2000, else False.
DIM strOSType, blnUseFCInfo
strOSType = ""
blnUseFCInfo = False


' Some of the discovery is performed by executing commands, like CLIs.
' If this script is run manually or via a cron-like job, it will invoke
' the utilities directly on this host.  If, however, it is run from the
' Collector host and was fed the hostname to explore, the utilities are
' executed on the remote host via a WMI CreateProcess call.  The 
' blnLocalExploration variable tells the script which way to invoke those
' commands.  Default to False, since this script will usually be run
' by the RHCollector.
Dim blnLocalExploration
blnLocalExploration = False

' blnVerbose is set by the /v command-line flag.  If True, more verbose output
' is written to the file, with each line beginning with a "#" character,
' which is interpreted by the Windows Explorer Parser as a comment line.
' Defaults to False, of course.
Dim blnVerbose
blnVerbose = False


' Execute special code to dump the incoming args before our data file name
' has been fetched IFF blnDumpArgs is TRUE.  This will rarely be necessary.
Dim blnDumpArgs
blnDumpArgs = False


' objWMIServices is the object used to connect to the root\WMI WMI namespace.
' objCIMV2Services is the object used to connect to the root\CIMV2 namespace.
' objDefServices is the object used to connect to the root\default namespace.
' Connect() sets up these objects.
Dim objWMIServices, objCIMV2Services, objDefServices

Const CLASSNAME = 0
Const NAMESPACE = 1
Dim strClasses, classArray
' strClasses and classArray help us iterate through the classes from which 
' we wish to fetch data.
' The strClasses string contains a stream of <class.namespace> pairs
' and each pair is separated by ":", like so:
' "<class.nmspc>:<class.nmspc>:<class.nmspc>...".
' So strClasses is split on ":" into classArray to separate the classes to be
' fetched, and each class pair is split on "." into the toks array, putting
' the class name string in the CLASSNAME slot, and the namespace string in the 
' NAMESPACE slot.
' We fetch instances of the following classes, using ":" as separator:
strClasses = _
             "Win32_OperatingSystem.CIMV2:" _
           & "Win32_Processor.CIMV2:" _
           & "Win32_SCSIControllerDevice.CIMV2:" _
           & "Win32_DiskDrive.CIMV2:" _
           & "MSFC_FCAdapterHBAAttributes.WMI:" _
           & "MSFC_FibrePortHBAAttributes.WMI:" _
           & "Win32_DiskDriveToDiskPartition.CIMV2:" _
           & "Win32_DiskPartition.CIMV2:" _
           & "Win32_LogicalDisk.CIMV2:" _
           & "Win32_LogicalDiskToPartition.CIMV2:" _
           & "Win32_Volume.CIMV2:" _
           & "Win32_MountPoint.CIMV2:" _
           & "Win32_Share.CIMV2:" _
           & "Win32_ShareToDirectory.CIMV2:" _
           & "Win32_NetworkAdapterConfiguration.CIMV2"

' split the <classname,namespace> pairs into classArray
classArray = Split(strClasses, ":")


' RunCommand_Local() reads the stdout and stderr of the command it executes
' into arrays.  stdout is read into strOLines(), stderr into strELines().
' The number of valid lines in strOLines is held in intOLineCount;
' intELineCount holds the same for strELines().  The line counts are required
' because the arrays are dynamically ReDimmed as needed, and often are larger
' than the number valid lines they contain.
' The Output class holds these arrays and integers, initializes, ReDims,
' and Resets them.
Class Output
    Public strOLines()
    Public intOLineCount
    Public strELines()
    Public intELineCount

    Public Function ReDimO(intNewSize)
        ReDim preserve strOLines(intNewSize)
        ' if (blnVerbose = True) Then
        '     objOutFile.WriteLine("    ++ Redimming strOLines to " & intNewSize)
        ' End If
    End Function

    Public Function ReDimE(intNewSize)
        ReDim preserve strELines(intNewSize)
        ' if (blnVerbose = True) Then
        '     objOutFile.WriteLine("    ++ Redimming strELines to " & intNewSize)
        ' End If
    End Function

    Public Function Reset()
        ' if (blnVerbose = True) Then
        '     objOutFile.WriteLine("Resetting Output...")
        ' End If
        intOLineCount = 0
	intELineCount = 0
        ReDim strOLines(1)
        ReDim strELines(1)
    End Function

    Private Sub Class_Initialize()
        intOLineCount = 0
        intELineCount = 0
        ReDim strOLines(10)
        ReDim strELines(10)
    End Sub
End Class


' When nslookup fails to resolve an IP addr, getFQHN() returns "UNCONFIGURED".
Const UNCONFIGURED = "UNCONFIGURED"


' The output data file consists of headers of the following types:
'
' |##|Start|##|<Command>|##|SUCCESS|##|  ' marks the start of a command
' |##|Output|##|       ' begin output of command
'  <lines of output>
' |##|Output End|##|   ' end output of command
' |##|Error|##|        ' begin error output of command
'  <error output>
' |##|Error End|##|    ' end error output of command

'  <all above headers comprise a section, which may be repeated >

' |##|EOF|##|          ' end of data file (allows detection of truncated files)
' 
' Define some constants to ease the use of these headings.
Const SEP = "|##|"
Const START = "|##|Start|##|"
Const OPUT = "|##|Output|##|"
Const OPUT_END = "|##|Output End|##|"
Const ERRR = "|##|Error|##|"        ' "ERR" and "ERROR" are reserved words
Const ERRR_END = "|##|Error End|##|"
Const EOF = "|##|EOF|##|"


Dim intFCount
intFCount = 1


' The following veriables are set by GetVxVMInfo(), and used by various
' routines.  
' - blnMPIOInstalled indicates whether the Microsoft MPIO component is
'   installed on the target host.  MPIO is installed by VxDMP or EMC
'   PowerPath on 2003, but MPIO is installed by default on 2008.
' - blnVMInstalled indicates whether any version of VxVM is installed
'   on the target host.  
' - If blnVxVMInstalled is True, then strVxVMVersion is set to the VxVM 
'   version string, and strVxVMInstallDir is the VxVM install directory, 
'   under which the VxVM and VxDMP CLIs are found.
Dim blnMPIOInstalled, blnVxVMInstalled, strVxVMVersion, strVxVMInstallDir

' The next three variables are set by GetEMCPPInfo().
' - blnEMCPPInstalled is set to True if any version of EMC PowerPath is 
'  installed on the target host.  
' - If EMCPP is installed, strEMCPPVersion is set to the version of EMCPP,
'   and strEMCPPInstallDir is set to the path to the directory where the 
'   EMCPP CLI (powermt.exe) is installed on the target host.
Dim blnEMCPPInstalled, strEMCPPVersion, strEMCPPInstallDir

' The VxVCS installation status, VxVCS version and the VCS bin installed
' location.
Dim blnVxVCSInstalled,  strVxVCSVersion, strVxVCSInstallDir

' Default declaration
blnVxVMInstalled = False
blnMPIOInstalled = False
blnEMCPPInstalled = False
blnVxVCSInstalled = False

'
' ------------------ End Global Variables and Constants ----------------------



' Parse the args, set system, user, pw, and outfile directory.
Call ParseArgs()

' strFileName is the filename that we originally synthesized for the data file.
' It may use the short hostname, so it may be superceded by strFQHNFileName.
' strPathnFileName is the full path to the filename, plus the filename.
Dim strFileName
Dim strPathnFileName

' Open the output file for writing.
strFileName = OpenFile(strSystem)

if (blnProbe = True) Then
    Call Do_Probe(strSystem, strUser, strPassword)
    ' Wscript.Quit 0
    ' notreached: Do_Probe() exits.
End If


' Write our version string into the output file.
Call WriteDataVersion(strDataVersion)


' Connect to WMI and CIMV2 namespaces and map admin$ filesystem if remote.
Call Connect(strSystem, strUser, strPassword)

Call GetOSType()

' Determine whether MPIO and/or VxVM are installed (MPIO may be present even
' if VxVM is not); if VxVM is installed, set the VxVM and VxVm installdir.
blnVxVMInstalled = GetVxVMInfo(strVxVMVersion, strVxVMInstallDir, _
                               blnMPIOInstalled)

' Determine whether VCS is installed or not. If installed get the version
' and installed location
blnVxVCSInstalled = GetVxVCSInfo(strVxVCSVersion, strVxVCSInstallDir)

' Determine whether EMCPP is installed
blnEMCPPInstalled = GetEMCPPInfo(strEMCPPVersion, strEMCPPInstallDir)

' Now we fetch a bunch of class instances.
' Interate through the classArray, fetching each class's data in turn.
Dim intI, toks
For intI = 0 to UBound(classArray)
    ' split each <classname,Namespace> pair into toks array by "."
    toks = Split(classArray(intI), ".")
    ' WScript.Echo "Slot " & intI & ": class = " & toks(CLASSNAME) & _
    '                ", NS = " & toks(NAMESPACE)
    Call fetchClass(toks(CLASSNAME), toks(NAMESPACE))
Next

' Get the Fully Qualified HostName associated with each IP Address listed
' in the IPAddress attribute of one instance of 
' Win32_NetworkAdapterConfiguration classes.
Call getFQHNs()

' Fetch the physical memory value from Win32_ComputerSystem class.
' We can't fetch the entire class, because the value of one of its attributes,
' OEMLogoBitmap, may be HUGE.  So since we're fetching just this specific
' attribute, we need a separate routine to do it.
Call getComputerSystem()


Call getMPIO_WMI()

' Get FC-mapping information 
Call GetFcpTargetMapping()


' Get LDM (Microsoft Logical Disk Manager) data.  But if VxVM is installed,
' VxVM overrides LDM: skip LDM discovery.
Call GetLDMData(blnVxVMInstalled)

' Get EMCPP data.
Call GetEMCPPData( blnMPIOInstalled, blnEMCPPInstalled, strEMCPPVersion, _
                   strEMCPPInstallDir)

' Get VxVM and DMP data.
Call GetVxVMData(blnVxVMInstalled, strVxVMVersion, strVxVMInstallDir)

' Get VCS data
Call GetVxVCSData(blnVxVCSInstalled, strVxVCSVersion, strVxVCSInstallDir)

objOutFile.WriteLine(EOF)

objOutFile.Close


' Output of this script is the name of data file we created, but cscript dumps
' a header into stdout, so preface our filename with a distinct marker, and
' the parsing script will look for this.  Use the originally-synthesized
' filename unless strFQHNFileName has been set, in which case we use that.
' Also, if strFQHNFileName has been set, move the data file to the new, FQHN
' filename.
Dim strCmd
If (strFQHNFileName = "") Then
    Wscript.Echo ":=:" & strFileName
Else
    strCmd = "%SystemRoot%\system32\cmd.exe /c move " & strPathnFileName & " " & strFQHNPathnFileName
    ' Wscript.Echo "strCmd: '" & strCmd & "'"
    Dim cOutput, intRetVal
    Set cOutput = New Output
    intRetVal = RunCommand_Local(strCmd, cOutput, True, True)
    if (intRetVal = 0) Then
        Wscript.Echo ":=:" & strFQHNFileName
    Else ' move failed, data file is old name
        Wscript.Echo ":=:" & strFileName
    End If
End If


if (blnLocalExploration = False) Then
    Call UnMap_Filesystem(strSystem, True) ' silent: we've closed objOutFile
End If

' Set Locale back to its original value.
SetLocale(intOrigLocale)

Wscript.Quit 0  ' normal exit status.

' hal-vm4_Solaris_200906241631.txt
' Month: 06, year 2009, day 24, hour 1631 



' #############################################################################
' ######################### Begin Functions and Subs ##########################
' #############################################################################

'
' Command-line Args:
' /s <system>
' /u <user>
' /p <password>
' /d <outfile directory>
' /f <outfile name>
' /v verbose output.
' all arguments are optional.
' strSystem, strUser, strPassword, strOutDir, strOutFile, & blnVerbose are 
' the variables that are set.
'
Function ParseArgs()

On Error Resume Next

    Dim args, strMsg
    Set args = WScript.Arguments
    Dim a0, a1, a3
    Dim acnt, blnGotSys, blnGotUser, blnGotPW

    ' Wscript.Echo "Argcnt: " & args.Count
    ' check for odd number of args.

    acnt = 0
    Do While acnt < args.Count
        ' Wscript.Echo "Arg: " & args.Item(acnt)
        Select Case args.Item(acnt)
            Case "/i"   ' inquire (probe, but /p was taken!)
                blnProbe = True
                acnt = acnt + 1
            Case "/s"   ' system (host)
                strSystem = args.Item(acnt+1)
                acnt = acnt + 2
            Case "/u"   ' username
                strUser = args.Item(acnt+1)
                acnt = acnt + 2
            Case "/p"   ' password
                strPassword = args.Item(acnt+1)
                acnt = acnt + 2
            Case "/d"   ' output directory
                strOutDir = args.Item(acnt+1)
                acnt = acnt + 2
                blnGotDFlag = True
            Case "/f"   ' output filename
                strOutFile = args.Item(acnt+1)
                acnt = acnt + 2
                blnGotFFlag = True
            Case "/v"   ' verbose output
                blnVerbose = True
                acnt = acnt + 1
            Case Else
                usage()
        End Select
    Loop

    ' If user specified the probe flag (/i) but not a filename (/f) and 
    ' output directory (/d), balk.
    If (blnProbe = True) Then
        if (blnGotFFlag = False Or blnGotDFlag = False) Then 
	    strMsg = "Probe (/i) requires filename (/f) and directory (/d)"
            Wscript.Echo strMsg
            usage()
        End If
    End If

    ' If user didn't specify a system, or if they specified "."  or
    ' localhost, we need to fetch the hostname on which we are running in
    ' order to synthesize the filename.  Get that hostname now.
    ' But first remember that we are Exploring the localhost for use 
    ' later in the discovery process.
    if (strSystem = "." Or strSystem = "localhost") Then
        blnLocalExploration = True
        strSystem = getHostname()
        if (blnVerbose = True) Then ' use Wscript.Echo: objOutFile not set yet.
            Wscript.Echo("ParseArgs, modified System: " & strSystem)
        End If
    End If
    
    if blnDumpArgs = True Then
        Dim objafso, Folder, strArgsPathnFileName, objArgsFile
        Set objafso = CreateObject("Scripting.FileSystemObject")
        ' Create Args.txt file
        strArgsPathnFileName = strOutDir & "\Args.txt"
        Set objArgsFile = objafso.CreateTextFile(strArgsPathnFileName, 2)

        ' Write all args into the file.
        Dim strCmd, objTmpArgs, K
        Set objTmpArgs = WScript.Arguments
        For K = 0 to objTmpArgs.Count - 1
            objArgsFile.WriteLine("'" & objTmpArgs(K) & "'")
        Next
        objArgsFile.Close
    End If

End Function ' ParseArgs()


Sub usage()
    Dim strErr
    strErr = "Illegal command line, usage: " & strProgName
    strErr = strErr & " [/u <user>] [/p <password>] [/s <system>] "
    strErr = strErr & "[/d <outputDirectory>]"
    Wscript.Echo strErr
    ' Wscript.Echo "Arguments may appear in any order."
    Wscript.Quit 2
End Sub


'
' OpenFile() opens an output textfile for writing.
' It is passed the hostname of the system being explored, which is used in
' the synthesizing of the filename.  It calls synthesize_Filename() to
' create the output file name, prepends the directory, and opens the file
' for writing.  If the /f <filename> flag was specified on the command line,
' use that filename rather than synthesizing one.  (The /f flag is only for
' parser debugging.)
'
Function OpenFile(strHostname)

On Error Resume Next

    Dim objfso, strFileName
    Set objfso = CreateObject("Scripting.FileSystemObject")

    if (strOutFile <> "") Then
        strFileName = strOutFile
    Else
        strFileName = synthesize_Filename(strHostname)
    End If
    strPathnFileName = strOutDir & "\" & strFileName

    Const ForWriting = 2
    Const ForAppending = 8
    Const InUnicode= True
    Set objOutFile = objfso.CreateTextFile(strPathnFileName, ForWriting, InUnicode)
    if (blnVerbose = True) Then
        Wscript.Echo "# Full filename: " & strPathnFileName
    End If

    OpenFile = strFileName

End Function ' OpenFile


'
' Write the version of this data-gathering script into the output file.
' The parsing script is tightly-coupled to this script, so it will check
' that this version is within the range that its version can handle.
'
Function WriteDataVersion(strDataVersion)

On Error Resume Next

    objOutFile.WriteLine(START & "DataVersion" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
    objOutFile.WriteLine("instance of DataVersion")
    objOutFile.WriteLine("{")
    objOutFile.WriteLine(vbTab & "DataVersion = " & strDataVersion)
    objOutFile.WriteLine("};")
    objOutFile.WriteLine(OPUT_END)

    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

End Function ' WriteDataVersion


'
' Connect to the specified host's root\WMI and root\CIMV2 namespaces
'
Function Connect(strSystem, strUser, strPassword)

On Error Resume Next

    ' Time the Connect().
    Dim strStart, strEnd, intDiff

    Dim strNameSpace, objSWbemLocator
    Dim intCOError, intWMIError, intCIMV2Error, objWMIError, objCIMV2Error
    Dim intDefError, objDefError
    intCOError = 0
    intWMIError = 0
    intCIMV2Error = 0
    intDefError = 0

    objOutFile.WriteLine(START & "Connect" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
    objOutFile.WriteLine("instance of Connect")
    objOutFile.WriteLine("{")

    ' Comments must be within START sections, so dump the args and other
    ' interesting information here if verbose flag was specified.
    strStart = TimeStamp("Start Connect")
    if (blnVerbose = True) Then
        objOutFile.WriteLine("# Final args:")
        objOutFile.WriteLine("# System: " & strSystem)
        objOutFile.WriteLine("# User: " & strUser)
        objOutFile.WriteLine("# Password: " & strPassword)
        objOutFile.WriteLine("# OutputDir: " & strOutDir)
        objOutFile.WriteLine("# OutFile: '" & strOutFile & "'")

        ' Write the user under which we're running into the file.
        ' Create an Output class instance for the output of the command
        Dim cOutput, strCmd, intExitVal, l
        Set cOutput = New Output
        strCmd = "%SystemRoot%\system32\whoami.exe"
        intExitVal = RunCommand_Local(strCmd, cOutput, False, False)

        objOutFile.WriteLine("# whoami?")
        for l = 0 To cOutput.intOLineCount-1
            objOutFile.WriteLine("# '" & cOutput.strOLines(l) & ";")
        Next
    End If

    Set objSWbemLocator = CreateObject("WbemScripting.SWbemLocator")
    if (Err <> 0) Then
        intCOError = Err.Number
        ' wScript.StdErr.WriteLine "CreateObject Err: " & hex(intCOError)
        objOutFile.WriteLine("Connect CreateObject failed: Err = " _
                             & intCOError)
    End If

    ' First connect to the WMI namespace.
    strNameSpace = "root\WMI"
    Set objWMIServices = objSWbemLocator.ConnectServer(strSystem, _
                                                       strNameSpace, _
                                                       strUser, _
                                                       strPassword)
    if (Err <> 0) Then
        intWMIError = Err.Number
        ' Set objWMIError = CreateObject("WbemScripting.SWbemLastError")
        ' wScript.StdErr.WriteLine "WMI ConnectServer Err: " & hex(intWMIError)
        ' objOutFile.WriteLine("WMI ConnectServer Err = " & hex(intWMIError))
    End If
    objWMIServices.Security_.ImpersonationLevel = 3

    ' Then the CIMV2 namespace.
    strNameSpace = "root\CIMV2"
    Set objCIMV2Services = objSWbemLocator.ConnectServer(strSystem, _
                                                       strNameSpace, _
                                                       strUser, _
                                                       strPassword)
    if (Err <> 0) Then
        intCIMV2Error = Err.Number
        ' wScript.StdErr.WriteLine "CIMV2 ConnectServer Err: " & hex(intCIMV2Error)
        ' objOutFile.WriteLine("CIMV2 ConnectServer Err = " & hex(intCIMV2Error))
        ' Set objCIMV2Error = CreateObject("WbemScripting.SWbemLastError")
    End If
    objCIMV2Services.Security_.ImpersonationLevel = 3

    ' Finally the default namespace.
    strNameSpace = "root\default"
    Set objDefServices = objSWbemLocator.ConnectServer(strSystem, _
                                                       strNameSpace, _
                                                       strUser, _
                                                       strPassword)
    if (Err <> 0) Then
        intDefError = Err.Number
        ' wScript.StdErr.WriteLine "default ConnectServer Err: " & hex(intDefError)
        ' objOutFile.WriteLine("default ConnectServer Err = " & hex(intDefError))
        ' Set objDefError = CreateObject("WbemScripting.SWbemLastError")
    End If
    objDefServices.Security_.ImpersonationLevel = 3


    ' If we are doing a remote exploration, map the admin$ filesystem.  
    ' Don't move this step out of this routine, or the verbose output will
    ' violate the constraints of the data file.
    if (blnLocalExploration = False) Then
        Call Map_Filesystem(strSystem, False)
    End If
    strEnd = TimeStamp("End Connect")
    intDiff = TimeStampDiff("Connect duration ", strStart, strEnd, True)

    objOutFile.WriteLine("}")
    objOutFile.WriteLine(OPUT_END)

    objOutFile.WriteLine(ERRR)
    if (intWMIError <> 0) Then
        'objOutFile.WriteLine("Op = " & objWMIError.operation & VBCRLF _
        '             & "ParamInfo = " & objWMIError.ParameterInfo & VBCRLF _
        '             & "ProviderName = " & objWMIError.ProviderName)
        objOutFile.WriteLine("Connect WMI Err = " & hex(intWMIError))
    End If
    if (intCIMV2Error <> 0) Then
        'objOutFile.WriteLine("Op = " & objCIMV2Error.operation & VBCRLF _
        '             & "ParamInfo = " & objCIMV2Error.ParameterInfo & VBCRLF _
        '             & "ProviderName = " & objCIMV2Error.ProviderName)
        objOutFile.WriteLine("Connect CIMV2 Err = " & hex(intCIMV2Error))
    End If
    if (intDefError <> 0) Then
        'objOutFile.WriteLine("Op = " & objDefError.operation & VBCRLF _
        '             & "ParamInfo = " & objDefError.ParameterInfo & VBCRLF _
        '             & "ProviderName = " & objDefError.ProviderName)
        objOutFile.WriteLine("Connect default Err = " & hex(intDefError))
    End If
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

    ' The inability to connect to any of the namespaces is a fatal error:
    ' we will not be able to discovery anything of use.  Exit with a 
    ' non-zero status.
    If (intCOError <> 0 or intWMIError <> 0 or intCIMV2Error <> 0 _
        or intDefError <> 0) Then
        ' Finish up our output file.
        objOutFile.WriteLine(EOF)
        objOutFile.Close
        Wscript.Quit 1  ' Utter cataclysmic failure.
    End If

End Function ' Connect


'
' If we are exploring Windows 2000 or 2003, we must use fcinfo.exe to get the
' FCTargetMapping because the WMI support for this method was unstable.
' If Windows 2008 or newer, we use WMI.
' GetOSType() fetches the "Caption" attribute of the Win32_OperatingSystem
' class which looks something like this: 
' "Microsoft(R) Windows(R) Server 2003, Enterprise Edition"
' ...and searches for the numerical sequence, setting "blnUseFCInfo" to True
' IFF 2000 or 2003, else False, and setting "strOSType" to the numerical
' sequence ("2000", "2003", "2008").
'
Function GetOSType()

On Error Resume Next

    Dim strQuery
    Dim objSet, objWbemObj
    Dim intMyError

    intMyError = 0

    strQuery = "SELECT * FROM Win32_OperatingSystem"
    Set objSet = objCIMV2Services.ExecQuery(strQuery)
        
    For Each objWbemObj In objSet
        if (Err <> 0) Then
            intMyError = Err.Number
        End If

        ' extract Caption attribute
        Dim strOSName, hntoks
        strOSName = objWbemObj.Caption
        ' Wscript.Echo("OSName: '" & strOSName & "'")
	' If the Caption contains "2000" or "2003", use fcinfo.  Otherwise,
	' use WMI.  Set blnUseFCInfo accordingly.
	if (InStr(strOSName, "2000") Or InStr(strOSName, "2003")) Then
            blnUseFCInfo = True
        Else
            blnUseFCInfo = False
        End If
	' Now set strOSType to the numerical value.  InStr finds the token
	' containing the leading "20" in "2000", "2003", or "2008", and
	' Mid extracts only 4 characters, cutting the trailing comma that some
	' Captions have after the number.
	' Split on space
        hntoks = Split(strOSName, " ")
        Dim intI
        For intI = 0 to UBound(hntoks)
            ' Wscript.Echo("tok " & intI & ": '" & hntoks(intI) & "'")
	    if (InStr(hntoks(intI), "20")) Then       ' this is the number
                strOSType = Mid(hntoks(intI), 1, 4)   ' extract 4 chars
            End If
        Next
        Exit For
    Next

    if (blnVerbose = True) Then
        Wscript.Echo("GetOSType, blnUseFCInfo: " & blnUseFCInfo)
        Wscript.Echo("GetOSType, strOSType: " & strOSType)
        Wscript.Echo("GetOSType, Error: " & hex(intMyError))
    End If

    GetOSType = intMyError

End Function ' GetOSType


'
' Fetch the class instances specified by the "strClass" parameter from the
' namespace specified by "strNameSpace".
'
Function fetchClass(strClass, strNameSpace)

On Error Resume Next

    objOutFile.WriteLine(START & strClass & SEP & SEP)
    objOutFile.WriteLine(OPUT)

    Dim objSet, objWbemObj
    Dim intMyError, objLastError
    intMyError = 0

    ' Time each WMI class fetch.
    Dim strStart, strEnd, intDiff
    strStart = TimeStamp("Start " & strClass)

    ' objOutFile.WriteLine("strClass: '" & strClass & "', NS: " & strNameSpace)

    ' Connect() created both namespace objects.  the strNameSpace parameter
    ' tells us which one to use for this class.
    If (strNameSpace = "WMI") Then
        Set objSet = objWMIServices.InstancesOf(strClass)
    Else
        Set objSet = objCIMV2Services.InstancesOf(strClass)
    End If
        
    For Each objWbemObj In objSet
        ' For WMI FC classes, if the FC MOF isn't installed, the above line
        ' will generate an error.  Catch the error and dump the description
        ' into the ERROR section.
        if (Err <> 0) Then
            intMyError = Err.Number
            Set objLastError = CreateObject("WbemScripting.SWbemLastError")
        End If
        objOutFile.WriteLine(objWbemObj.getObjectText_)
    Next
    strEnd = TimeStamp("End " & strClass)
    intDiff = TimeStampDiff(strClass & " duration ", strStart, strEnd, True)

    objOutFile.WriteLine(OPUT_END)

    objOutFile.WriteLine(ERRR)
    if (intMyError <> 0) Then
        objOutFile.WriteLine( "Operation = " & objLastError.operation & VBCRLF _
                     & "ParamInfo = " & objLastError.ParameterInfo & VBCRLF _
                     & "ProviderName = " & objLastError.ProviderName)
        objOutFile.WriteLine("Err = " & hex(intMyError))
    End If
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")
End Function  ' fetchClass()



Function getMPIO_WMI()

On Error Resume Next

    Dim blnMPIOInstalled, strClass

    strClass = "MPIO_GET_DESCRIPTOR"
    objOutFile.WriteLine(START & strClass & SEP & SEP)
    objOutFile.WriteLine(OPUT)

    blnMPIOInstalled = isMPIOInstalled()
    if (blnMPIOInstalled = False) Then   ' Our work is done.
        objOutFile.WriteLine("MPIO_INSTALLED = False")
        objOutFile.WriteLine(OPUT_END)
        objOutFile.WriteLine(ERRR)
        objOutFile.WriteLine("MPIO_INSTALLED = False")
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    End If

    ' WMI MPIO_GET_DESCRIPTOR class data works for some versions of EMC PP,
    ' but not others.  powermt.exe data works for ALL versions of EMC PP.
    ' Therefore, if EMC PowerPath is installed, we'll always use the output
    ' of the powermt CLI instead of MPIO_GET_DESCRIPTOR.
    if (blnEMCPPInstalled = True) Then
        objOutFile.WriteLine("EMC_POWERPATH_INSTALLED = True")
        objOutFile.WriteLine(OPUT_END)
        objOutFile.WriteLine(ERRR)
        objOutFile.WriteLine("EMC_POWERPATH_INSTALLED = True")
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    End If

    ' Another case: some hosts may have the MPIO key but not have any 
    ' MPIO_GET_DESCRIPTOR instances (W2k8 without SFW or EMC PP is this case)
    Dim objSet, objWbemObj
    Dim intMyError, objLastError
    intMyError = 0

    ' Time the WMI class fetch.
    Dim strStart, strEnd, intDiff

    strStart = TimeStamp("Start " & strClass)

    Set objSet = objWMIServices.InstancesOf(strClass)
    For Each objWbemObj In objSet
        ' For WMI FC classes, if the FC MOF isn't installed, the above line
        ' will generate an error.  Catch the error and dump the description
        ' into the ERROR section.
        if (Err <> 0) Then
            intMyError = Err.Number
            Set objLastError = CreateObject("WbemScripting.SWbemLastError")
        End If
        objOutFile.WriteLine(objWbemObj.getObjectText_)
    Next
    strEnd = TimeStamp("End " & strClass)
    
    ' Write the timestamp entry IFF the class fetch succeeded, and tell 'em
    ' that no instances were found if not.
    If (intMyError <> &H80041010) Then
        intDiff = TimeStampDiff(strClass & " duration ",strStart,strEnd,True)
    Else
        objOutFile.WriteLine("No MPIO_GET_DESCRIPTOR instances found")
    End If
    objOutFile.WriteLine(OPUT_END)

    objOutFile.WriteLine(ERRR)
    if (intMyError <> 0) Then
        objOutFile.WriteLine( "Operation = " & objLastError.operation & VBCRLF _
                     & "ParamInfo = " & objLastError.ParameterInfo & VBCRLF _
                     & "ProviderName = " & objLastError.ProviderName)
        objOutFile.WriteLine("Err = " & hex(intMyError))
    End If
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")
    

End Function 'getMPIO_WMI


'
' GetFcpTargetMapping() obtains the FC information for all LUNs that are
' exposed via HBAs. 
' If the host being explored is Windows 2000 or 2003, the target mapping is
' done via fcinfo.exe.  Else it is done via WMI.
' GetOSType() determined the platform type, and set blnUseFCInfo accordingly.
'
Function GetFcpTargetMapping()

On Error Resume Next

   If (blnUseFCInfo = False) Then
       Call WMI_GetFcpTargetMapping()
   Else
       Call FCINFO_GetFcpTargetMapping()
   End If

End Function  ' GetFcpTargetMapping()


'
' The GetFcpTargetMapping method returns 7 in HBAStatus when the inEntryCount
' we specified is smaller than the number of entries it needs to return.
Const HBA_STATUS_ERR_MORE_DATA = 7  ' need more slots for data.


'
' WMI_GetFcpTargetMapping() performs the WMI FC target mapping.  It fetches
' the MSFC_FibrePortHBAAttributes class, and for each of those port WWNs,
' invokes the MSFC_HBAFCPInfo() method, fetching the FC information involving
' that HBA Port.  We're interested in the FCID, ArrayNode and ArrayPort WWNs,
' and the SCSI Bus, Target, and LUN numbers.
' 
Function WMI_GetFcpTargetMapping()

On Error Resume Next

    objOutFile.WriteLine(START & "WMI_GetFcpTargetMapping" & SEP & SEP)

    Dim objSetPorts, objPort, objSetHBAFcpInfo, objHbaFcpInfo
    Dim objInP, objOutP  ' objects holding input and output parameters.
    Dim intInEntryCount, blnRepeat
    Dim intMyError, objLastError
    intMyError = 0

    Set objSetPorts = objWMIServices.InstancesOf("MSFC_FibrePortHBAAttributes")
    Set objSetHBAFcpInfo = objWMIServices.InstancesOf("MSFC_HBAFCPInfo")

    objOutFile.WriteLine(OPUT)

    Dim strStart, strEnd, intDiff
    strStart = TimeStamp("Start WMI_GetFcpTargetMapping")
    ' Loop through each HBA Port's info, and for each port, loop through the
    ' HBAFcpInfo.  Invoke the method for matching Port and FCP InstanceName.
    For Each objPort in objSetPorts
        ' If the FC MOF isn't installed, the above line will generate an error.
        ' Catch it and dump the description into the ERROR section.
        if (Err <> 0) Then
            intMyError = Err.Number
            Set objLastError = CreateObject("WbemScripting.SWbemLastError")
        End If
        For Each objHbaFcpInfo in objSetHBAFcpInfo
            if (Err <> 0) Then
                intMyError = Err.Number
                Set objLastError = CreateObject("WbemScripting.SWbemLastError")
            End If
            ' Both classes return data for all HBAPorts, so we must match 
            ' FCP's and HBAPort's HBA InstanceName
            If (objPort.InstanceName = objHbaFcpInfo.InstanceName) Then 
                ' fetch method object.
                Set objInP = objHbaFcpInfo.Methods_("GetFcpTargetMapping").InParameters.Clone_
                ' Try 10 entries (returned LUNs).  If it's too small, we'll
                ' get HBAStatus of 7 and TotalEntryCount will be set to the
                ' required value.  We'll run it again with the correct size.
                intInEntryCount = 10
                Do           
                    blnRepeat = False
                    ' we know we have matching entries (InstanceName values
                    ' matched above), so construct the InParams for the 
                    ' GetFcpTargetMapping method.
                    objInP.HbaPortWWN = objPort.Attributes.PortWWN
                    objInP.InEntryCount = intInEntryCount
                    ' Execute the GetFcpTargetMapping method.
                    Set objOutP =_
                        objWMIServices.ExecMethod(objHbaFcpInfo.Path_,_
                                                  "GetFcpTargetMapping", _
                                                  objInP) 
                    ' Did the method fail?
                    If (objOutP.HBAStatus <> 0) Then
          '              DIM objLastError
'            Set objLastError = CreateObject("WbemScripting.SWbemLastError")
'objOutFile.WriteLine "Operation = " & objLastError.operation & VBCRLF _
'    & "ParamInfo = " & objLastError.ParameterInfo & VBCRLF _
'    & "ProviderName = " & objLastError.ProviderName

            ' objOutFile.WriteLine("xError " & oLastError.GetObjectText_)
            ' objOutFile.WriteLine("xError " & Err.Number)
                        ' Yup.  Was it because InEntryCount was too small?
                        If (HBA_STATUS_ERR_MORE_DATA <> objOutP.HBAStatus) Then
                            ' Nope.  Some other problem.
                            ' objOutFile.WriteLine("Error " & objOutP.HBAStatus)
                        Else
                            ' Yup.  Set InEntryCount to returned TotalEntryCount
                            ' and rerun.
                            intInEntryCount = objOutP.TotalEntryCount
                            ' objOutFile.WriteLine("Error " & objOutP.HBAStatus)
                            ' objOutFile.WriteLine("Error, need " & intInEntryCount & " entries.")
                            blnRepeat = True
                        End If
                    Else
                        ' Display the returned mappings.
                        If (objOutP.OutEntryCount = 0) Then
                            objOutFile.WriteLine("(No mappings)")
                        Else
                            Dim I, objEntry, objFCPId, objScsiId
                            Dim strHBAPortWWN
                            strHBAPortWWN = WWNToString(objInP.HbaPortWWN)

                            For I = 0 To objOutP.OutEntryCount - 1
                                Set objEntry = objOutP.Entry(I)
                                Call dump_FCInfo(objPort.InstanceName, _
                                                 strHBAPortWWN, _
                                                 objEntry, I)
                            Next ' for I
                        End If ' outEntryCount = 0
                    End If  ' HBAStatus 
                Loop While (blnRepeat = True)
            End If ' objHBAFcpInfo is paired with objPort
        Next ' objHBAFcpInfo
    Next ' oEltPort
    strEnd = TimeStamp("End WMI_GetFcpTargetMapping")
    intDiff = TimeStampDiff("End WMI_GetFcpTargetMapping duration ", _
                            strStart, strEnd, True)

    ' This or FCINFO_GetFcpTargetMapping is the end of the data gathering.
    ' Compute the total duration of the data gathering process, and write
    ' it into this section.
    strAllEnd = now
    Call TimeStampDiff("SCRIPT DURATION ", strAllStart, strAllEnd, True)

    objOutFile.WriteLine(OPUT_END)

    objOutFile.WriteLine(ERRR)
    if (intMyError <> 0) Then
        objOutFile.WriteLine( "Operation = " & objLastError.operation & VBCRLF _
                     & "ParamInfo = " & objLastError.ParameterInfo & VBCRLF _
                     & "ProviderName = " & objLastError.ProviderName)
        objOutFile.WriteLine("Err = " & hex(intMyError))
    End If
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

End Function  ' WMI_GetFcpTargetMapping()


'
' If the platform being discovered is Windows 2000 or 2003, we use fcinfo.exe
' to return the FC TargetMapping data.  FCINFO_GetFcpTargetMapping() performs
' that data collection process.
'
Function FCINFO_GetFcpTargetMapping()

On Error Resume Next

    objOutFile.WriteLine(START & "FCINFO_GetFcpTargetMapping" & SEP & SEP)
    objOutFile.WriteLine(OPUT)

    Dim strStart, strEnd, intDiff
    strStart = TimeStamp("Start FCINFO_GetFcpTargetMapping")

    ' Find out whether fcinfo.exe is installed on the host we're exploring.
    Dim blnfcinfoInstalled
    blnfcinfoInstalled = File_Found("%SystemRoot%\system32", "fcinfo.exe", _
                                    False)

    If (blnfcinfoInstalled = True) Then
        Wscript.Echo("fcinfo.exe is installed")
        objOutFile.WriteLine("# fcinfo.exe is installed")
    Else
        Wscript.Echo("fcinfo.exe is NOT installed")
	FCINFO_GetFcpTargetMapping = 1
        objOutFile.WriteLine("fcinfo.exe is not installed")

        strEnd = TimeStamp("End FCINFO_GetFcpTargetMapping")
        intDiff = TimeStampDiff("End FCINFO_GetFcpTargetMapping duration ", _
                            strStart, strEnd, True)

        ' This or WMI_GetFcpTargetMapping is the end of the data gathering.
        ' Compute the total duration of the data gathering process, and 
        ' write it into this section.
        strAllEnd = now
        Call TimeStampDiff("SCRIPT DURATION ", strAllStart, _
                           strAllEnd, True)

        objOutFile.WriteLine(OPUT_END)
        objOutFile.WriteLine(ERRR)
        objOutFile.WriteLine("fcinfo.exe is not installed")
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    End If

    Dim strCmd, strBaseCmd, cOutput, intRetVal
    Set cOutput = New Output
    strCmd = "%SystemRoot%\system32\fcinfo.exe"
    ' Run the fcinfo command locally if exploring the localhost, else remotely.
    intRetVal = RunCommand(strCmd, cOutput, False, False)
    ' Dump_Output(cOutput)

    if (intRetVal <> 0) Then
        objOutFile.WriteLine("# !!" & strCmd & " Exit Value: " & intRetVal)
        ' XXXX Stop?
    End If

    objOutFile.WriteLine("Command: '" & strCmd & "'")
    Dim I
    For I = 0 To cOutput.intOLineCount-1
        objOutFile.WriteLine(cOutput.strOLines(I))
    Next ' For I

    Dim toks, J, intACnt
    ' Look for line "There are n adapters:" and extract 'n'
    For I = 0 To cOutput.intOLineCount-1
        If (InStr(cOutput.strOLines(I), "There")) Then
            toks = Split(cOutput.strOLines(I), " ")
            ' for J = 0 to UBound(toks)
            '     Wscript.Echo("toks(" & J & "): " & toks(J)) 
            ' Next ' for J
            intACnt = CInt(toks(2))
        End If
    Next ' For I
    ' Wscript.Echo "Adapter count: " & intACnt

    ' Run "fcinfo /mapping /ai:n" for each adapter slot n
    strBaseCmd = "%SystemRoot%\system32\fcinfo.exe /mapping /ai:"
    For I = 0 to intACnt-1
        strCmd = strBaseCmd & I
	' Wscript.Echo "fc cmd: '" & strCmd & "'"
        cOutput.Reset()
        intRetVal = RunCommand(strCmd, cOutput, False, False)
'        if (blnVerbose = True And intRetVal = 0) Then
'            Call Dump_Output(cOutput)
'        End If
        objOutFile.WriteLine("Command: '" & strCmd & "'")
        For J = 0 To cOutput.intOLineCount-1
            objOutFile.WriteLine(cOutput.strOLines(J))
        Next ' For J
    Next ' For I

    strEnd = TimeStamp("End FCINFO_GetFcpTargetMapping")
    intDiff = TimeStampDiff("End FCINFO_GetFcpTargetMapping duration ", _
                            strStart, strEnd, True)

    ' This or WMI_GetFcpTargetMapping is the end of the data gathering.
    ' Compute the total duration of the data gathering process, and write
    ' it into this section.
    strAllEnd = now
    Call TimeStampDiff("SCRIPT DURATION ", strAllStart, strAllEnd, True)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

End Function  ' FCINFO_GetFcpTargetMapping()


'
' GetLDMData() runs diskpart.exe to determine the LDM disk and volume
' configuration on the host.
' DiskPart.exe runs non-interactively when fed an input script with the /s
' flag.  We need the format of that input file to be this:
' list disk
' list volume
' select disk 0
' detail disk
' select disk 1
' detail disk
' <etc. for all LDM disks>
' select volume 0
' detail volume
' select volume 1
' detail volume
' <etc. for all LDM volumes>
' 
'
Function GetLDMData(blnVxVMInstalled)

On Error Resume Next
' On Error goto 0

    objOutFile.WriteLine(START & "LDM" & SEP & SEP)
    objOutFile.WriteLine(OPUT)

    ' If VxVM is installed, then skip LDM discovery.  VxVM overrides LDM, 
    ' even though the LDM CLI (diskpart.exe) is still installed.  VxVM 
    ' will discover everything.  Add descriptive comment, close out the
    ' section, and return.
    if (blnVxVMInstalled = True) Then
        objOutFile.WriteLine("VxVM is installed, skipping LDM")
        Wscript.Echo("VxVM is installed, skipping LDM")
        objOutFile.WriteLine(OPUT_END)

        objOutFile.WriteLine(ERRR)
        objOutFile.WriteLine("VxVM is installed")
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    End If

    ' Start the clock...
    Dim strStart, strEnd, intDiff
    strStart = TimeStamp("Start GetLDMData")
    
    ' Make sure this host has diskpart.exe installed!
    Dim blnDiskpartFound
    blnDiskpartFound = File_Found("%SystemRoot%\system32", "diskpart.exe", _
                                    False)

    ' diskpart.exe is driven by input scripts using the '/s' flag.  If we're
    ' exploring a remote host, create the script textfile on that host.

    ' Wscript.Echo "DP found: " & blnDiskpartFound

    ' 'list disk' command lists the LDM disks.  'list volume' lists the LDM
    ' Volumes.  We'll need the output of each of these commands in order to
    ' derive the disk numbers and volume numbers to fetch with 'detail disk'
    ' and 'detail volume' commands.  So first create a file with "list volume"
    ' and run it, capturing the output.  Then truncate the file and create
    ' with "list disk".  This will be the beginning of the real input file.
    ' (It is much faster (and safer) to ask diskpart to fetch everything in
    ' one script, because it takes care of timing problems.  So we'll do this,
    ' now that we have the list of disks and volumes to fetch.)
    ' Run this short file to get the list of disks, and save the output.
    ' Then add "list volume" command, then lines selecting each of the disks,
    ' followed by "detail disk", and then selecting each of the volumes 
    ' followed by "detail volume".
    Dim strCmd, cOutputDnV, cOutputAll, cOutputJunk, intRetVal
    Set cOutputDnV = New Output
    Set cOutputAll = New Output
    Set cOutputJunk = New Output

    ' diskpart.exe doesn't work correctly when invoked twice without a long
    ' (15 seconds is recommended) sleep between the invocations.  So just
    ' invoke it twice: once with "list disk" and "list volume"--which we use
    ' to determine the set of disks and volumes to detail--and then a second
    ' time with "list disk", "list volume", and all of the detail disk and
    ' detail volume information to be written into the data file.

    ' Write "list disk" and "list volume" into a txt scriptfile and feed it
    ' to diskpart.
    strCmd = "echo list disk > %SystemRoot%\Temp\diskInfo.txt"  ' create
    intRetVal = RunCommand(strCmd, cOutputJunk, False, True)
    strCmd = "echo list volume >> %SystemRoot%\Temp\diskInfo.txt"  ' append
    intRetVal = RunCommand(strCmd, cOutputJunk, False, True)
    ' Now run diskpart.exe, feeding it the script file.  This will fetch the
    ' volume overview into cOutputDnV.
    strCmd = "%SystemRoot%\system32\diskpart.exe /s %SystemRoot%\Temp\diskInfo.txt"
    intRetVal = RunCommand(strCmd, cOutputDnV, False, False)
    ' Wscript.Echo "After hhh We Are!" & "ol: " & cOutputDnV.intOLineCount

    ' Now sleep for 10 seconds to ensure that diskpart is ready to go again.
    ' Wscript.Sleep 10000

    ' The first part of the cOutputDnV output buffer contains the disk 
    ' overview data.  Process it in order to derive the list of valid disks.
    ' We'll feed each of these disk numbers into "select disk n" and "disk 
    ' detail" to get the details of each disk.
    '
    ' diskpart.exe "list disk" output format is this:
    ' Microsoft DiskPart version 6.0.6001'
    ' Copyright (C) 1999-2007 Microsoft Corporation.'
    ' On computer: CCS-DELL2850-7'
    '
    '   Disk ###  Status      Size     Free     Dyn  Gpt'
    '   --------  ----------  -------  -------  ---  ---'
    '   Disk 0    Online       137 GB      0 B         '
    '   Disk 1    Online       137 GB   127 GB   *     '
    '   <etc>
    Dim i, j, toks, blnStart, strTmpStr
    blnStart = False
    For i = 0 To cOutputDnV.intOLineCount-1
        ' Wscript.Echo "dp line: '" & cOutputDnV.strOLines(i) & "'"
        If (InStr(cOutputDnV.strOLines(i), "Disk")) Then
            strTmpStr = cOutputDnV.strOLines(i)
            strTmpStr = FixForSplit(strTmpStr) ' cut multiple space chars.
            toks = split(strTmpStr, " ")
            ' for j = 0 to UBound(toks)
            '     Wscript.Echo("vtoks(" & j & "): " & toks(j)) 
            ' Next ' for j
            ' "Disk" is in toks(0), "###" or disk number is in toks(1)
            if (toks(1) = "###") Then ' all "Disk" lines after this are numbers
                blnStart = True
            ElseIf (blnStart = True) Then
                if (toks(0) = "Disk") Then
                    ' Wscript.Echo("    Got disk " & toks(1))
                    strCmd = "echo select disk " & toks(1) & _
                         " >> %SystemRoot%\Temp\diskInfo.txt"
                    cOutputJunk.Reset()
                    intRetVal = RunCommand(strCmd, cOutputJunk, False, True)
                    strCmd = "echo detail disk >> %SystemRoot%\Temp\diskInfo.txt"
                    cOutputJunk.Reset()
                    intRetVal = RunCommand(strCmd, cOutputJunk, False, True)
                End If 
            End If
            ' Wscript.Echo ""
        End If
    Next

    ' Now process the volume overview we fetched above, also in cOutputDnV: 
    ' the second part of the data.  
    ' "list volume" output is of this format:
    ' Volume ###  Ltr  Label        Fs     Type        Size     Status     Info
    ' ----------  ---  -----------  -----  ----------  -------  ------     ----
    '   Volume 0  G    Simple_10GB  NTFS   Simple        10 GB  Healthy 
    '   Volume 1       Stripe_15,1  NTFS   Stripe      2042 MB  Healthy
    '   Volume 2  I    Mir_14,16_C  NTFS   Mirror      1021 MB  Healthy    
    '   Volume 3  F    Span-8,10Hi  NTFS   Spanned     3042 MB  Healthy    
    '   Volume 4  H    RAID5_2,3,4  NTFS   RAID-5      4090 MB  Healthy    
    '   Volume 5  E                       DVD-ROM         0 B  No Media
    '   Volume 6  C                 NTFS   Partition     99 GB  Healthy  System
    '   Volume 7  D    Ghost Image  NTFS   Partition     37 GB  Healthy     
    '
    ' diskpart.exe will abort if we attempt to fetch details of volumes with 
    ' Status values of "No Media", so we must skip those.  Only feed "Healthy"
    ' volumes into "detail volume".
    blnStart = False
    For i = 0 To cOutputDnV.intOLineCount-1
        ' Wscript.Echo "v line: '" & cOutputDnV.strOLines(i) & "'"
        If (InStr(cOutputDnV.strOLines(i), "Volume")) Then
            strTmpStr = cOutputDnV.strOLines(i)
            strTmpStr = FixForSplit(strTmpStr)
            toks = split(strTmpStr, " ")

            ' for j = 0 to UBound(toks)
            '     Wscript.Echo("vtoks(" & j & "): " & toks(j)) 
            ' Next ' for j

            ' "Volume" is in toks(0), "###" or volume number is in toks(1)
            ' all "Volume" lines after this precede volume numbers
            if (toks(0) = "Volume" And toks(1) = "###") Then 
                blnStart = True
            Elseif (blnStart = True) Then
                ' Here it gets tricky.  As shown above, the Ltr, Label, and
                ' Fs fields may be blank, so we can't pre-determine tok 
                ' indices.  So look for "MB" or "GB" or "B", and the next
                ' token is the Status field.
                for j = 0 to UBound(toks)
                    if (toks(j) = "B" or toks(j) = "MB" or toks(j) = "GB") Then
                        ' Wscript.Echo(" vol " & toks(1) & " stat " & toks(j+1))
                        if (toks(j+1) = "Healthy") Then ' good volume
                            strCmd = "echo select volume " & toks(1) & _
                                    " >> %SystemRoot%\Temp\diskInfo.txt"
                            cOutputJunk.Reset()
                            intRetVal = RunCommand(strCmd, cOutputJunk, _
			                           False, True)
                            strCmd = "echo detail volume >> %SystemRoot%\Temp\diskInfo.txt"
                            cOutputJunk.Reset()
                            intRetVal = RunCommand(strCmd, cOutputJunk, _
			                           False, True)
                            Exit For
                        End If
                    End If
                Next ' for j

            End If
            ' Wscript.Echo ""

        End If
    Next

    ' Ok, we've created the script file with all of the commands in it.
    ' Now run diskpart.exe, feeding it that file, to get the whole enchilada.
    strCmd = "%SystemRoot%\system32\diskpart.exe /s %SystemRoot%\Temp\diskInfo.txt"
    intRetVal = RunCommand(strCmd, cOutputAll, False, False)

    ' Write all output into the data file.
    For J = 0 To cOutputAll.intOLineCount-1
        objOutFile.WriteLine(cOutputAll.strOLines(J))
        ' Wscript.Echo (cOutputAll.strOLines(J))
    Next ' For J

    ' Now delete the script file, diskInfo.txt
    strCmd = "del %SystemRoot%\Temp\diskInfo.txt"
'    intRetVal = RunCommand(strCmd, cOutputAll, False, True)

    strEnd = TimeStamp("End GetLDMData")
    intDiff = TimeStampDiff("GetLDMData duration ", strStart, strEnd, True)

    ' Close out the section.
    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

End Function 'GetLDMData


'
' GetVxVMData() runs all of the CLIs on the target host required to gather the
' VxVM data.
'
Function GetVxVMData(blnVxVMInstalled, strVxVMVersion, strVxVMInstallDir)

On Error Resume Next

    objOutFile.WriteLine(START & "VxVM" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
 
    Dim strStart, strEnd, intDiff
    strStart = TimeStamp("Start GetVxVMData")

    ' Wscript.Echo "GetVxVMData: VxVMInstalled " & blnVxVMInstalled
    ' Wscript.Echo "GetVxVMData: strVxVMVersion " & strVxVMVersion
    ' Wscript.Echo "GetVxVMData: strVxVMInstallDir " & strVxVMInstallDir
    
    if (blnVxVMInstalled = True) Then
        Wscript.Echo "VxVM is Installed!"
        Wscript.Echo "Version: " & strVxVMVersion
    Else
        Wscript.Echo "VxVM is not Installed!"
    End If

    ' If SFW isn't installed, close out this section and return.
    if (blnVxVMInstalled = False) Then
        ' Wscript.Echo "Not Installed!"
        strEnd = TimeStamp("End GetVxVMData")
        intDiff = TimeStampDiff("GetVxVMData duration ", strStart, strEnd, True)
        objOutFile.WriteLine(OPUT_END)
        objOutFile.WriteLine(ERRR)
        objOutFile.WriteLine("# VxVM is not installed")
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    End If

    objOutFile.WriteLine("Version=" & strVxVMVersion)
    objOutFile.WriteLine("INSTALLDIR=" & strVxVMInstallDir)

    Dim blnCLIsFound, intMissingCLICnt, arrMissingCLIs()
    blnCLIsFound = VerifyVxVMCLIs(strVxVMInstallDir, intMissingCLICnt, _
                                  arrMissingCLIs)
    if (blnCLIsFound = False) Then
        strEnd = TimeStamp("End GetVxVMData")
        intDiff = TimeStampDiff("GetVxVMData duration ", strStart, strEnd, True)
        objOutFile.WriteLine(OPUT_END)
        objOutFile.WriteLine(ERRR)
        For i = 0 to intMissingCLICnt-1
            objOutFile.WriteLine("# ! " & arrMissingCLIs(i) & " was not found on host")
        Next ' for i
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    Else
        Wscript.Echo "# All CLIs found"
    End If

    ' Wscript.Echo "VxVMInstallDir: '" & strVxVMInstallDir & "'"
    doGetVxVMData(strVxVMInstallDir)

    strEnd = TimeStamp("End GetVxVMData")
    intDiff = TimeStampDiff("GetVxVMData duration ", strStart, strEnd, True)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    ' If VM is installed but not a version that uses MPIO, flag as an error
    ' so that the parser will know that we don't support this configuration.
    if (blnVxVMInstalled = True) and (blnMPIOInstalled = False) Then
        objOutFile.WriteLine("VM is installed but MPIO is NOT installed")
    End If
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

End Function 'GetVxVMData

'
' GetVxVCSData() runs all of the CLIs on the target host required to gather the
' VxVCS data.
'
Function GetVxVCSData(blnVxVCSInstalled, strVxVCSVersion, strVxVCSInstallDir)

On Error Resume Next

    objOutFile.WriteLine(START & "VxVCS" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
 
    Dim strStart, strEnd, intDiff
    strStart = TimeStamp("Start GetVxVCSData")

    ' Wscript.Echo "GetVxVCSData: VxVCSInstalled " & blnVxVCSInstalled
    ' Wscript.Echo "GetVxVCSData: strVxVCSVersion " & strVxVCSVersion
    ' Wscript.Echo "GetVxVCSData: strVxVCSInstallDir " & strVxVCSInstallDir
    
    if (blnVxVCSInstalled = True) Then
        Wscript.Echo "VxVCS is Installed!"
        Wscript.Echo "Version: " & strVxVCSVersion
    Else
        Wscript.Echo "VxVCS is not Installed!"
    End If

    ' If VCS isn't installed, close out this section and return.
    if (blnVxVCSInstalled = False) Then
        ' Wscript.Echo "Not Installed!"
        strEnd = TimeStamp("End GetVxVCSData")
        intDiff = TimeStampDiff("GetVxVCSData duration ", strStart, strEnd, True)
        objOutFile.WriteLine(OPUT_END)
        objOutFile.WriteLine(ERRR)
        objOutFile.WriteLine("# VxVCS is not installed")
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    End If

    objOutFile.WriteLine("Version=" & strVxVCSVersion)
    objOutFile.WriteLine("INSTALLDIR=" & strVxVCSInstallDir)

    Dim blnCLIsFound, intMissingCLICnt, arrMissingCLIs()
    blnCLIsFound = VerifyVxVCSCLIs(strVxVCSInstallDir, intMissingCLICnt, _
                                  arrMissingCLIs)
    ' If VerifyVxVCSCLIS returns false, then the below if condition
    ' prints the missing cli's and ends the output with OPUT_END
    '
    ' If VerifyVxVCSCLIS returns true, then the OPUT_END following the 'if'
    ' condition ends the output
    if (blnCLIsFound = False) Then
        strEnd = TimeStamp("End GetVxVCSData")
        intDiff = TimeStampDiff("GetVxVCSData duration ", strStart, strEnd, True)
        objOutFile.WriteLine(OPUT_END)
        objOutFile.WriteLine(ERRR)
        For i = 0 to intMissingCLICnt-1
            objOutFile.WriteLine("# ! " & arrMissingCLIs(i) & " was not found on host")
        Next ' for i
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    Else
        Wscript.Echo "# All CLIs found"
    End If

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

    ' Wscript.Echo "VxVCSInstallDir: '" & strVxVCSInstallDir & "'"
    doGetVxVCSData(strVxVCSInstallDir)

    strEnd = TimeStamp("End GetVxVCSData")
    intDiff = TimeStampDiff("GetVxVCSData duration ", strStart, strEnd, True)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

End Function 'GetVxVCSData


'
' GetEMCPPData() runs all of the CLIs on the target host required to gather the
' EMC PowerPath data.
'
Function GetEMCPPData(blnMPIOInstalled, blnEMCPPInstalled, strEMCPPVersion, _
                      strEMCPPInstallDir)

On Error Resume Next

    Dim strVersion, i

    Wscript.Echo "GetEMCPPData: blnMPIOInstalled " & blnMPIOInstalled
    Wscript.Echo "GetEMCPPData: EMCPPInstalled " & blnEMCPPInstalled
    Wscript.Echo "GetEMCPPData: strEMCPPVersion " & strEMCPPVersion
    Wscript.Echo "GetEMCPPData: strEMCPPInstallDir " & strEMCPPInstallDir

    objOutFile.WriteLine(START & "EMCPP" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
 
    Dim strStart, strEnd, intDiff
    strStart = TimeStamp("Start GetEMCPPData")
    
    ' GetEMCPPInfo() has already determined whether MPIO is installed, and 
    ' whether EMCPP is installed and where.
    If (blnMPIOInstalled = True) Then
        objOutFile.WriteLine("MPIO_INSTALLED = True")
    Else
        objOutFile.WriteLine("MPIO_INSTALLED = False")
    End If
    If (blnEMCPPInstalled = True) Then
        Wscript.Echo "EMCPP is Installed!"
        Wscript.Echo "Version: " & strEMCPPVersion
    Else
        Wscript.Echo "EMCPP is not Installed!"
    End If

    ' If EMC PP isn't installed, close out this section and return.
    if (blnEMCPPInstalled = False) Then
        Wscript.Echo "Not Installed!"
        strEnd = TimeStamp("End GetEMCPPData")
        intDiff = TimeStampDiff("GetEMCPPData duration ",strStart,strEnd,True)
        objOutFile.WriteLine(OPUT_END)
        objOutFile.WriteLine(ERRR)
        objOutFile.WriteLine("# EMCPP is not installed")
        objOutFile.WriteLine(ERRR_END)
        objOutFile.WriteLine("")
        Exit Function
    End If

    doGetEMCPPData(strEMCPPInstallDir)

    strEnd = TimeStamp("End GetEMCPPData")
    intDiff = TimeStampDiff("GetEMCPPData duration ", strStart, strEnd, True)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    ' If EMCPP is installed but not a version that uses MPIO, flag as an error
    ' so that the parser will know that we don't support this configuration.
    if (blnEMCPPInstalled = True) and (blnMPIOInstalled = False) Then
        objOutFile.WriteLine("EMCPP is installed but MPIO is NOT installed")
    End If
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

End Function 'GetEMCPPData



'
' GetVxVMInfo() consults the target host's registry to find out whether SFW
' is installed, and if so, what version is installed and where (i.e., the
' value of 'InstallDir'.
' - If VxVM is installed, the function returns True and sets 'strVersion' to
'   the VxVM version, and 'strInstallDir' to the VxVM InstallDir location,
'   as read from the registry, and sets blnMPIOInstalled to True IFF MPIO
'   is installed on the host, False if not.
' - If VxVM is not installed, GetVxVMInfo() returns False, and leaves 
'   strVersion, strInstallDir, and blnMPIOInstalled unchanged.
'
Function GetVxVMInfo(strVersion, strInstallDir, blnMPIOInstalled)

On Error Resume Next

    Dim strKeyPath, arrSubKeys, subkey
    Dim blnVmInstalled, blnIs64Bit, strVMKey
    blnVmInstalled = False

    ' First find out whether MPIO is installed on the host.  At least EMC
    ' PowerPath (in addition to newer versions of VxDMP) uses (and installs)
    ' MPIO, so it may be present even if VxDMP isn't installed.
    blnMPIOInstalled = isMPIOInstalled()

    ' Next determine whether the host is running a 32- or 64-bit OS.  If
    ' 64-bit, all of the keys we expect will be under "Software\Wow6432Node"
    ' instead of "Software".
    blnIs64Bit = Is64BitOS()
    if (blnIs64Bit = True) Then
        Wscript.Echo "It is 64-bit."
    Else
        Wscript.Echo "It is 32-bit."
    End If

    if (blnIs64Bit) Then
        strKeyPath = "Software\Wow6432Node\"
    Else
        strKeyPath = "Software\"
    End If

    ' First look for a key like "Volume Manager" under <strKeyPath>\Veritas
    ' E.g. "Volume Manager 5.1".
    strKeyPath = strKeyPath & "Veritas\"
    arrSubKeys = RegEnumKeys(objDefServices, strKeyPath)
    For Each subkey In arrSubKeys
        ' Wscript.Echo subkey
        If (InStr(subkey, "Volume Manager")) Then
            blnVmInstalled = True
            strVMKey = subkey
            ' Wscript.Echo "Installed!"
            Exit For
        End If
    Next
    GetVxVMInfo = blnVmInstalled

    ' If SFW isn't installed we're finished.
    if (blnVmInstalled = False) Then
        ' Wscript.Echo "Not Installed!"
        Exit Function
    End If

    ' Wscript.Echo "Fetch the value of VM 5.1 key..."
    ' SFW is installed: fetch its Version key value, and write it into the
    ' Output section.
    Dim strKeyName, strKeyValue
    strKeyPath = strKeyPath & strVMKey
    strKeyName = "Version"
    strKeyValue = RegRead(objDefServices, strKeyPath, strKeyName)
    ' Wscript.Echo "KeyName: " & strKeyName & ", KeyValue: " & strKeyValue
    strVersion = strKeyValue

    ' Now fetch the InstallDir value.  This is where we'll find the VxVM
    ' and VxDMP CLIs.
    strKeyName = "INSTALLDIR"
    strKeyValue = RegRead(objDefServices, strKeyPath, strKeyName)
    ' Wscript.Echo "KeyName: " & strKeyName & ", KeyValue: " & strKeyValue
    strInstallDir = strKeyValue

End Function 'GetVxVMInfo


'
' GetVxVCSInfo() consults the target host's registry to find out whether VxVCS
' is installed, and if so, what version is installed and where (i.e., the
' value of 'InstallDir'.
' - If VxVCS is installed, the function returns True and sets 'strVersion' to
'   the VxVCS version, and 'strInstallDir' to the VxVCS InstallDir location,
' - If VxVCS is not installed, GetVxVCSInfo() returns False, and leaves 
'   strVersion, strInstallDir, and blnMPIOInstalled unchanged.
'
Function GetVxVCSInfo(strVersion, strInstallDir)

On Error Resume Next

    Dim strKeyPath, arrSubKeys, subkey
    Dim blnVCSInstalled, blnIs64Bit, strVCSKey
    blnVCSInstalled = False

    ' determine whether the host is running a 32- or 64-bit OS.  If
    ' 64-bit, all of the keys we expect will be under "Software\Wow6432Node"
    ' instead of "Software".
    blnIs64Bit = Is64BitOS()
    if (blnIs64Bit = True) Then
        Wscript.Echo "It is 64-bit."
    Else
        Wscript.Echo "It is 32-bit."
    End If

    if (blnIs64Bit) Then
        strKeyPath = "Software\Wow6432Node\"
    Else
        strKeyPath = "Software\"
    End If

    ' look for a key like "VCS" under <strKeyPath>\Veritas
    strKeyPath = strKeyPath & "Veritas\"
    arrSubKeys = RegEnumKeys(objDefServices, strKeyPath)
    For Each subkey In arrSubKeys
        ' Wscript.Echo subkey
        If (InStr(subkey, "VCS")) Then
            blnVCSInstalled = True
            strVCSKey = subkey & "\Base"
            ' Wscript.Echo "Installed!"
            Exit For
        End If
    Next
    GetVxVCSInfo = blnVCSInstalled

    ' If SFW isn't installed we're finished.
    if (blnVCSInstalled = False) Then
        ' Wscript.Echo "Not Installed!"
        Exit Function
    End If

    ' If VCS is installed: fetch its Version key value, and write it into the
    ' Output section.
    Dim strKeyName, strKeyValue
    strKeyPath = strKeyPath & strVCSKey
    strKeyName = "Version"
    strKeyValue = RegRead(objDefServices, strKeyPath, strKeyName)
    ' Wscript.Echo "KeyName: " & strKeyName & ", KeyValue: " & strKeyValue
    strVersion = strKeyValue

    ' Now fetch the InstallDir of VCS CLI
    strKeyName = "HomeDirectory"
    strKeyValue = RegRead(objDefServices, strKeyPath, strKeyName)
    ' Wscript.Echo "KeyName: " & strKeyName & ", KeyValue: " & strKeyValue
    strInstallDir = strKeyValue & "\bin\"

End Function 'GetVxVCSInfo


'
' GetEMCPPInfo() consults the target host's registry to find out whether EMCPP
' is installed, and if so, what version is installed and where (i.e., the
' value of 'InstallDir'.
' - If EMCPP is installed, the function returns True and sets 'strVersion' to
'   the EMCPP version, and 'strInstallDir' to the EMCPP InstallDir location,
'   as read from the registry
' - If EMCPP is not installed, GetEMCPPInfo() returns False, and leaves 
'   strVersion, strInstallDir unchanged.
'
Function GetEMCPPInfo(strVersion, strInstallDir)

On Error Resume Next

    Dim strKeyPath 
    Dim blnEMCPPInstalled 
    Dim strRegPath, strRegKey
    Dim keyToCheck, cliName

    blnEMCPPInstalled = False
    bGotEmcppPath = False

    ' First find out whether MPIO is installed on the host.  At least EMC
    ' PowerPath (in addition to newer versions of VxDMP) uses (and installs)
    ' MPIO, so it may be present even if VxDMP isn't installed.
    blnMPIOInstalled = isMPIOInstalled()

    ' Check with 32-bit registry key first. 
    strRegPath = "Software\Wow6432Node\EMC\PowerPath"
    ' Wscript.Echo "EMC PP registry path: " & strRegPath
    keyToCheck = "Emcp_mp_rtl.dll"
    cliName = "powermt.exe"
    blnEMCPPInstalled = getEmcPPPath(strRegPath, keyToCheck, cliName, strVersion, strInstallDir)

    if (blnEMCPPInstalled = False) Then
        ' If we have got here means that we haven't got PowerPath
        ' details under Wow6432Node. Now check in the normal path
        strRegPath = "Software\EMC\PowerPath"
        ' Wscript.Echo "EMC PP registry path: " & strRegPath
        blnEMCPPInstalled = getEmcPPPath(strRegPath, keyToCheck, cliName, strVersion, strInstallDir)
    End If

    GetEMCPPInfo = blnEMCPPInstalled

    ' If EMCPP isn't installed we're finished.
    if (blnEMCPPInstalled = False) Then
        Wscript.Echo "EMCPP Not Installed!"
        strVersion = Null
        strInstallDir = Null
        Exit Function
    End If

    ' Wscript.Echo " powermt install location is: " & strInstallDir

End Function 'GetEMCPPInfo


' getEmcPPPath() consults target host's registry
' with registry paths to find out whether
' EMC PowerPath is installed or not.
' - If EMC PowerPath is installed, function returns true
'   and sets 'strVersion' to the PP version
'   and 'strInstallDir' to the path where powermt is available
' - If EMC PowerPath is not installed, function returns
'   false and leaves 'strVersion' and 'strInstallDir' unchanged
'
Function getEmcPPPath(emcppRegPath, keyToCheck, cliName, strVersion, strInstallDir)
    Dim bGotEmcppPath
    Dim strRegPath, strRegKey
    Dim strKeyName, strKeyValue
    Dim bCLIExist

    bGotEmcppPath = False

    ' Wscript.Echo "getEmcPPPath: emcppRegPath is: " & emcppRegPath
    strRegKey = "Version"
    bGotEmcppPath = checkRegPathAndKey(emcppRegPath, strRegKey)
    If (bGotEmcppPath = True) Then
        strVersion = RegRead(objDefServices, emcppRegPath, strRegKey)
        ' Wscript.Echo "getEmcPPPath: PowerPath version is: " & strVersion
        bGotEmcppPath = checkRegPathAndKey(emcppRegPath, keyToCheck)
        If (bGotEmcppPath = True) Then
            strKeyValue = RegRead(objDefServices, emcppRegPath, keyToCheck)
            strInstallDir = left( strKeyValue, len(strKeyValue) - (len(keyToCheck) + 1))
            bCLIExist = File_Found(strInstallDir, cliName, False)
            If (bCLIExist = True) Then
                ' Wscript.Echo "powermt.exe install path: " & strInstallDir
                getEmcPPPath = True
                Exit Function
            End If
            ' Wscript.Echo "getEmcPPPath: Couldn't find powermt at : " & strInstallDir
        End If
        ' Wscript.Echo "getEmcPPPath: Couldn't find dll: " & keyToCheck
    End If

    ' Wscript.Echo "getEmcPPPath: Got here means powermt.exe not found"
    getEmcPPPath = False
    strVersion = Null
    strInstallDir = Null
    Exit Function

End Function 'getEmcPPPath    


' checkRegPathAndKey() simple function to check if
' at the given registry path 'strRegPath' the Registry
' key 'strRegKey' exists. It does this by reading the
' registry key value and checking for a non-null value
'
' Returns true if the Key exists and a non-null value
' is found
'
' Returns false if the value cannot be read at the given
' path and key
Function checkRegPathAndKey(strRegPath, strRegKey)
    Dim arrSubKeys, keyValue
    Dim bKeyExists
    bKeyExists = False
    ' Wscript.Echo "checkRegPathAndKey: Path " & strRegPath & ", Key: " & strRegKey

    keyValue = RegRead(objDefServices, strRegPath, strRegKey)
    ' Check if the value returned is not null. If not null
    ' we have found the key under the given path
    If (keyValue <> "") Then
        bKeyExists = True
    End If
    checkRegPathAndKey = bKeyExists
End Function 'checkRegPathandKey


'
' isMPIOInstalled() checks for the presence of a specific registry section
' named "mpio" to determine whether a DMP mechanism using MPIO is installed
' on the host.  
' If the registry section is found, isMPIOInstalled returns True.  
' Else it returns False.
'
Function isMPIOInstalled()

    isMPIOInstalled = False    ' assume not installed by default.
    Dim strKeyPath, arrSubKeys, subkey
    ' We determine whether MPIO is installed on the host by looking for an
    ' "mpio" section under SYSTEM\CurrentControlSet\Services
    strKeyPath = "SYSTEM\CurrentControlSet\Services"
    arrSubKeys = RegEnumKeys(objDefServices, strKeyPath)
    For Each subkey In arrSubKeys
        ' Wscript.Echo subkey
        If (InStr(subkey, "mpio")) Then
            isMPIOInstalled = True
            Wscript.Echo "MPIO is Installed!"
            Exit For
        End If
    Next

End Function 'isMPIOInstalled

'
' VerifyVxVCSCLIs() ensures that all of the CLIs required for VxVCS and VxDMP
' discovery are present on the target machine.  By the time VerifyVxVCSCLIs()
' has been called, the script already verified that SFW is installed.
' VerifyVxVCSCLIs returns True IFF all CLIs are found.
' If any of the CLIs are not found, it returns False and fills the 
' arrMissingCLIs array with the string names of each missing CLI. 
' We need hares.exe, hasys.exe, hagrp.exe and hastatus.exe 
' in order to do the VxVCS and VxDMP discovery.
'
Function VerifyVxVCSCLIs(strInstallDir, intMissingCLICnt, arrMissingCLIs)

    DIM arrDiskGroups, blnSuccess

    DIM arrCLIs, j
    arrCLIs = Array("hares.exe", "hasys.exe", "hagrp.exe", "hastatus.exe")
    DIM strCLI
    intMissingCLICnt = 0

    VerifyVxVCSCLIs = True   ' assume they're all present
    ' ...But make the arrMissingCLIs array handle the case of all CLIs missing.
    ReDim arrMissingCLIs(UBOUND(arrCLIs))

    Dim blnCLIInstalled
    for j = 0 to UBOUND(arrCLIs)
        strCLI = arrCLIs(j)
	' Wscript.Echo "Looking for '" & strCLI & "' in " & strInstallDIR

        blnCLIInstalled = File_Found(strInstallDir, strCLI, False)
        if (blnCLIInstalled = False) Then
            Wscript.Echo "Didn't find " & strCLI
            arrMissingCLIs(intMissingCLICnt) = strCLI
	    intMissingCLICnt = intMissingCLICnt + 1
            VerifyVxVCSCLIs = False
        End If
    Next ' for j

End Function 'VerifyVxVCSCLIs


'
' VerifyVxVMCLIs() ensures that all of the CLIs required for VxVM and VxDMP
' discovery are present on the target machine.  By the time VerifyVxVMCLIs()
' has been called, the script already verified that SFW is installed.
' VerifyVxVMCLIs returns True IFF all CLIs are found.
' If any of the CLIs are not found, it returns False and fills the 
' arrMissingCLIs array with the string names of each missing CLI. 
' We need vxassist.exe, vxdg.exe, vxprint.exe, vxvol.exe and vxdisk.exe
' in order to do the VxVM and VxDMP discovery.
' 
' We cannot rely on vxdmpadm.exe as in case of Windows 2008, with SF 5.1
' there can be case where vxdmpadm bits are not installed and instead
' windows MPIO is installed
'
Function VerifyVxVMCLIs(strInstallDir, intMissingCLICnt, arrMissingCLIs)

    DIM arrDiskGroups, blnSuccess

    DIM arrCLIs, j
    arrCLIs = Array("vxassist.exe", "vxdg.exe", "vxprint.exe", _
                    "vxvol.exe", "vxdisk.exe")
    DIM strCLI
    intMissingCLICnt = 0

    VerifyVxVMCLIs = True   ' assume they're all present
    ' ...But make the arrMissingCLIs array handle the case of all CLIs missing.
    ReDim arrMissingCLIs(UBOUND(arrCLIs))

    Dim blnCLIInstalled
    for j = 0 to UBOUND(arrCLIs)
        strCLI = arrCLIs(j)
	' Wscript.Echo "Looking for '" & strCLI & "' in " & strInstallDIR

        blnCLIInstalled = File_Found(strInstallDir, strCLI, False)
        if (blnCLIInstalled = False) Then
            Wscript.Echo "Didn't find " & strCLI
            arrMissingCLIs(intMissingCLICnt) = strCLI
	    intMissingCLICnt = intMissingCLICnt + 1
            VerifyVxVMCLIs = False
        End If
    Next ' for j

End Function 'VerifyVxVMCLIs


' Set the dynamic array to 10 slots initially; we'll REDIM as necessary.
Const INIT_ARR_SIZE = 10
' If we have to REDIM arrDiskGroups array, bump it substantially.
Const ARR_REDIM_INC = 40

'
' doGetVxVMData() runs all of the CLIs required to gather the VxVM data.
' It writes the data into the VxVM section of the data file.
'
Function doGetVxVMData(strInstallDir)

    DIM blnSuccess

    ' First fetch the VxVM version string.
    blnSuccess = doGetVxVMVersion(strInstallDir)
    
    ' Then get the disk groups.  Save these in an array, we'll use them.
    DIM intDiskGroupCnt, arrDiskGroups()

    ReDim arrDiskGroups(INIT_ARR_SIZE)
    intDiskGroupCnt = 0
    blnSuccess = doGetVxVMDGs(strInstallDir, intDiskGroupCnt, arrDiskGroups)

    Dim j
    ' if (blnSuccess = True) Then
    '     for j = 0 to intDiskGroupCnt-1
    '         Wscript.Echo "DG[" & j & "]: " & arrDiskGroups(j)
    '     Next
    ' End If

    ' Next get the volumes in each of these disk groups.
    blnSuccess = doGetVxVMDGVols(strInstallDir, intDiskGroupCnt, arrDiskGroups)

    ' And the disks in each of these disk groups.
    blnSuccess = doGetVxVMDGDisks(strInstallDir, intDiskGroupCnt, arrDiskGroups)

    blnSuccess = doGetVxVMHardDiskInfo(strInstallDir)


End Function 'doGetVxVMData


'
' doGetVxVCSData() runs all of the CLIs required to gather the VxVM data.
' It writes the data into the VxVM section of the data file.
'
Function doGetVxVCSData(strInstallDir)

    DIM blnSuccess


    Dim strCmd, strFullCmd, strDCmd, cOutput, intRes
    Set cOutput = New Output

    objOutFile.WriteLine(START & "haclus_display" & SEP & SEP)
    objOutFile.WriteLine(OPUT)

    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "haclus.exe" & chr(34) & " -display -localclus"
    strFullCmd = chr(34) & strInstallDir & strCmd

    ' We want to run the command *on the host being explored*, so invoke
    ' RunCommand().  It'll run the command locally if we're running on the
    ' host to be explored, or remotely if exploring a different host.
    Call RunCommand(strFullCmd, cOutput, False, False)
    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal
    intRes = Write_Output(strDCmd, cOutput, False)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

    objOutFile.WriteLine(START & "hasys_display" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "hasys.exe" & chr(34) & " -display -localclus"
    strFullCmd = chr(34) & strInstallDir & strCmd
    
    Call RunCommand(strFullCmd, cOutput, False, False)
    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal
    intRes = Write_Output(strDCmd, cOutput, False)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

    objOutFile.WriteLine(START & "hagrp_display" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "hagrp.exe" & chr(34) & " -display -localclus"
    strFullCmd = chr(34) & strInstallDir & strCmd
    
    Call RunCommand(strFullCmd, cOutput, False, False)
    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal
    intRes = Write_Output(strDCmd, cOutput, False)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

    objOutFile.WriteLine(START & "hares_dep" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "hares.exe" & chr(34) & " -dep"
    strFullCmd = chr(34) & strInstallDir & strCmd
    
    Call RunCommand(strFullCmd, cOutput, False, False)
    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal
    intRes = Write_Output(strDCmd, cOutput, False)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("") 


    objOutFile.WriteLine(START & "hastatus_summary" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
    
    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "hastatus.exe" & chr(34) & " -summary"
    strFullCmd = chr(34) & strInstallDir & strCmd
    
    Call RunCommand(strFullCmd, cOutput, False, False)
    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal
    intRes = Write_Output(strDCmd, cOutput, False)

    objOutFile.WriteLine(OPUT_END)
    objOutFile.WriteLine(ERRR)
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("") 

    objOutFile.WriteLine(START & "hares_display" & SEP & SEP)
    objOutFile.WriteLine(OPUT)
    
    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "hares.exe" & chr(34) & " -display -attribute Group Type State VMDGResName VolumeName -type MountV -localclus"
    strFullCmd = chr(34) & strInstallDir & strCmd
    
    Call RunCommand(strFullCmd, cOutput, False, False)
    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal
    intRes = Write_Output(strDCmd, cOutput, False)

    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "hares.exe" & chr(34) & " -display -attribute Group Type State DiskGroupName -type VMDg -localclus"
    strFullCmd = chr(34) & strInstallDir & strCmd
    
    Call RunCommand(strFullCmd, cOutput, False, False)
    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal
    intRes = Write_Output(strDCmd, cOutput, False)
End Function 'doGetVxVCSData


'
' doGetEMCPPData() runs the CLI required to gather the EMC PowerPath data
' ("powermt.exe display dev=all").  
' It writes the data into the "EMCPP" section of the data file.
'
'C:\Documents and Settings\administrator.SPCQADOM>powermt display dev=all|more
'Pseudo name=harddisk1
'CLARiiON ID=CK200051200424 [SG_Dell285011]
'Logical device ID=600601609BCE14000A80AE1D31D7DB11 [VxVm5.0LUN 121]
'state=alive; policy=CLAROpt; priority=0; queued-IOs=0
'Owner: default=SP B, current=SP A
'==============================================================================
'---------------- Host ---------------   - Stor -   -- I/O Path -  -- Stats ---
'### HW Path                 I/O Paths    Interf.   Mode    State  Q-IOs Errors
'==============================================================================
'   2 port2\path0\tgt0\lun0     c2t0d0    SP A0     active  alive      0      0
'   3 port3\path0\tgt0\lun0     c3t0d0    SP A0     active  alive      0      0

'Pseudo name=harddisk2
'CLARiiON ID=CK200051200424 [SG_Dell285011]
'Logical device ID=600601609BCE14000B80AE1D31D7DB11 [VxVm5.0LUN 122]
'state=alive; policy=CLAROpt; priority=0; queued-IOs=0
'Owner: default=SP B, current=SP A
'==============================================================================
'---------------- Host ---------------   - Stor -   -- I/O Path -  -- Stats ---
'### HW Path                 I/O Paths    Interf.   Mode    State  Q-IOs Errors
'==============================================================================
'   2 port2\path0\tgt0\lun1     c2t0d1    SP A0     active  alive      0      0
'   3 port3\path0\tgt0\lun1     c3t0d1    SP A0     active  alive      0      0
'
Function doGetEMCPPData(strInstallDir)

    DIM blnSuccess
    Dim strFullCmd, strDCmd, strPPVersionCmd
    Set cOutput = New Output

    strFullCmd = chr(34) & strInstallDir  & chr(92) & "powermt" & chr(34) & " display dev=all"
    strPPVersionCmd = chr(34) & strInstallDir  & chr(92) & "powermt" & chr(34) & " version"

    strDCmd = "powermt version"
    Call RunCommand(strPPVersionCmd, cOutput, False, False)
    blnSuccess = Write_Output(strDCmd, cOutput, False)

    strDCmd = "powermt display dev=all"
    ' Wscript.Echo "powermt path " & strFullCmd
    Call RunCommand(strFullCmd, cOutput, False, False)

    blnSuccess = Write_Output(strDCmd, cOutput, False)

End Function 'doGetEMCPPData


'
' doGetVxVMVersion() invokes 'vxassist version' on the target machine to
' retrieve the VxVM version, and writes the result into the VxVM section
' of the data file, using VxVM's and VxDMP's internal section format.
' Note that by the time doGetVxVMVersion() is called, we've already 
' determined that VxVM is indeed installed.
'
Function doGetVxVMVersion(strInstallDir)

    Dim strCmd, strFullCmd, strDCmd, cOutput, intRes
    Set cOutput = New Output

    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "vxassist.exe" & chr(34) & " version"
    strFullCmd = chr(34) & strInstallDir & strCmd
    strDCmd = "vxassist.exe version"

    ' We want to run the command *on the host being explored*, so invoke
    ' RunCommand().  It'll run the command locally if we're running on the
    ' host to be explored, or remotely if exploring a different host.
    Call RunCommand(strFullCmd, cOutput, False, False)

    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal
    intRes = Write_Output(strDCmd, cOutput, False)

    doGetVxVMVersion = True

End Function 'doGetVxVMVersion


'
' doGetVxVMDGs() executes "vxdg.exe list" to fetch all of the VxVM disk groups.
' First it writes the output to the data file, but then it strains through
' the output, and writes the disk group names into the arrDiskGroups array.
' intDiskGroupCnt tracks the number of DiskGroups in the array, because the
' array may be larger than the number of occupied slots.
'
' strInstallDir tells us where the vxdg.exe command is installed on the host.
'
Function doGetVxVMDGs(strInstallDir, intDiskGroupCnt, arrDiskGroups)

    Dim strCmd, strFullCmd, strDCmd, cOutput, intRes
    Set cOutput = New Output

    doGetVxVMDGs = True  ' We hope for success.

    ' chr(34) brackets path & filename in dbl-quotes in case it contains spaces
    strCmd = "vxdg.exe" & chr(34) & " list"
    strFullCmd = chr(34) & strInstallDir & strCmd
    strDCmd = "vxdg.exe list"

    Call RunCommand(strFullCmd, cOutput, False, False)
    intRes = Write_Output(strDCmd, cOutput, False)

    ' If the vxdg command succeeded, extract the diskgroup names from the
    ' output, and save them in arrDiskGroups array.
    if (intRes = 0) Then
        Dim i, toks, strLine, strstr
        For I = 0 To cOutput.intOLineCount-1
            If (InStr(cOutput.strOLines(I), "Name")) Then
		strLine = cOutput.strOLines(I)
		strLine = Left(strLine, Len(strLine) - 1) ' There's only a LF
                ' Wscript.Echo "Nameline: '" & strLine & "'"
		' Line looks like: "Name     :   BasicGroup"
                toks = split(strLine, ":")
		' Wscript.Echo "0: " & toks(0) & ", 1: " & toks(1)
		strstr = Trim(toks(1))
                ' Wscript.Echo "DG Name: '" & strstr & "'"
		if (intDiskGroupCnt >= UBOUND(arrDiskGroups)) Then
                    ' Wscript.Echo "Redimming for " & strstr & " " & intDiskGroupCnt
                    ReDim preserve arrDiskGroups(intDiskGroupCnt+ARR_REDIM_INC)
                End If
		arrDiskGroups(intDiskGroupCnt) = strstr
		intDiskGroupCnt = intDiskGroupCnt + 1
            End If
        Next ' For I
    Else
        doGetVxVMDGs = False  ' dang.
    End If

End Function 'doGetVxVMDGs

'
' doGetVxVMDGVols() executes "vxprint -g -vn" and "vxdg.exe list" on each 
' diskgroup passed in via arrDiskGroups to fetch all of the VxVM disk group
' Volumes.
'
' Next it executes "vxprint -g #DGNAME# -vl to get the details of all the
' volumes under the DiskGroup. It gets detailed information about each volume.
' From the output of this command, we extract the "DeviceName" and prepare an
' array. This is next passed to doGetVxVMVolDiskInfo to get the details of
' Disks under each volume
' This has to be done because the volume-names have to be unique in a DG but
' can have same name across diskgroups
'
' strInstallDir tells us where the vxdg.exe command is installed on the host.
' intDiskGroupCnt tells us how many valid diskgroups are in arrDiskGroups.
' The list of DiskGroups is passed in via arrDiskGroups array.
'
Function doGetVxVMDGVols(strInstallDir, intDiskGroupCnt, arrDiskGroups)

    Dim strCmd, strFullCmd, strDCmd, cOutput, intRes, blnSuccess
    DIM intVolumeCnt, arrVolumes()
    ReDim arrVolumes(INIT_ARR_SIZE)
    Set cOutput = New Output

    doGetVxVMDGVols = True  ' We hope for success.

    Dim i, j, strDG, strLine, strDN, arrToks
    For i = 0 To intDiskGroupCnt-1
        strDG = arrDiskGroups(i)
        strCmd = "vxprint.exe" & chr(34) & " -g " & strDG & " -vn"
        strFullCmd = chr(34) & strInstallDir & strCmd
        strDCmd = "vxprint.exe -g " & strDG & " -vn"
        Call RunCommand(strFullCmd, cOutput, False, False)
	' don't include the path in the command in the header.
        intRes = Write_Output(strDCmd, cOutput, False)

        strCmd = "vxdg.exe" & chr(34) & " dginfo -g " & strDG
        strFullCmd = chr(34) & strInstallDir & strCmd
        strDCmd = "vxdg.exe dginfo -g " & strDG
        Call RunCommand(strFullCmd, cOutput, False, False)
        intRes = Write_Output(strDCmd, cOutput, False)

        strCmd = "vxprint.exe" & chr(34) & " -g " & strDG & " -vl"
        strFullCmd = chr(34) & strInstallDir & strCmd
        strDCmd = "vxprint.exe -g " & strDG & " -vl"
        Call RunCommand(strFullCmd, cOutput, False, False)
        intRes = Write_Output(strDCmd, cOutput, False)
        
        ' Now extract the "DeviceName" for each volume
        intVolumeCnt = 0
        For j = 0 To cOutput.intOLineCount-1
            If (InStr(cOutput.strOLines(j), "DeviceName")) Then
               ' Wscript.Echo "Got DeviceName -- " & cOutput.strOLines(j)
               strLine = cOutput.strOLines(j)
               strLine = Left(strLine, Len(strLine) - 1) ' There's only a LF
               arrToks = split(strLine, ":")
               strDN = Trim(arrToks(1))
               If (intVolumeCnt >= UBOUND(arrVolumes)) Then
                  ReDim preserve arrVolumes(intVolumeCnt+ARR_REDIM_INC)
               End If
               arrVolumes(intVolumeCnt) = strDN
               intVolumeCnt = intVolumeCnt + 1
            End If
        Next ' For j
        cOutput.Reset()

        ' For j = 0 to intVolumeCnt-1
        '    Wscript.Echo "DN [" & j & "]: " & arrVolumes(j)
        ' Next
        blnSuccess = doGetVxVMVolDiskInfo(strInstallDir, intVolumeCnt, arrVolumes)
    Next ' For I

    ' XXXXX I don't handle errors here.

End Function 'doGetVxVMDGVols


'
' doGetVxVMDGDisks() executes "vxprint -g -dn" and "vxprint -g -dn" on each 
' diskgroup passed in via arrDiskGroups to fetch all of the VxVM disk group
' Disks.
'
' strInstallDir tells us where the vxdg.exe command is installed on the host.
' intDiskGroupCnt tells us how many valid diskgroups are in arrDiskGroups.
' The list of DiskGroups is passed in via arrDiskGroups array.
'
Function doGetVxVMDGDisks(strInstallDir, intDiskGroupCnt, arrDiskGroups)

    Dim strCmd, strFullCmd, strDCmd, cOutput, intRes
    Set cOutput = New Output

    doGetVxVMDGDisks = True

    Dim i, strDG
    For i = 0 To intDiskGroupCnt-1
        strDG = arrDiskGroups(i)
        strCmd = "vxprint.exe" & chr(34) & " -g " & strDG & " -dn"
        strFullCmd = chr(34) & strInstallDir & strCmd
        strDCmd = "vxprint.exe -g " & strDG & " -dn"
        Call RunCommand(strFullCmd, cOutput, False, False)
        intRes = Write_Output(strDCmd, cOutput, False)
    Next ' For I

    ' XXXXX I don't handle errors here.

End Function 'doGetVxVMDGDisks


'
' doGetVxVMVolDiskInfo() executes "vxvol volinfo -v" on each volume passed in 
' via arrVolumes array. The array contains the DeviceName of the volumes.
' This commands gets the Volume DiskInfo.
'
' strInstallDir tells us where the vxdg.exe command is installed on the host.
' intVolumeCnt tells us how many valid volumes are in the arrVolumes array.
' The list of Volumes is passed in via arrVolumes array.
'
Function doGetVxVMVolDiskInfo(strInstallDir, intVolumeCnt, arrVolumes)

    Dim strCmd, strFullCmd, cOutput, intRes, strDCmd
    Set cOutput = New Output

    doGetVxVMVolDiskInfo = True

    ' Wscript.Echo "in doGetVxVMVolDiskInfo"

    Dim i, strVol 
    For i = 0 To intVolumeCnt-1
        strVol = arrVolumes(i)
        ' Wscript.Echo "in VolDiskInfo DN[" & i & "]: " & strVol
        strCmd = "vxvol.exe" & chr(34) & " volinfo " & strVol & " -v"
        strFullCmd = chr(34) & strInstallDir & strCmd
        strDCmd = "vxvol.exe volinfo " & strVol & " -v"
        Call RunCommand(strFullCmd, cOutput, False, False)
        intRes = Write_Output(strDCmd, cOutput, False)
        cOutput.Reset()
    Next ' For I

    ' XXXXX I don't handle errors here.

End Function 'doGetVxVMVolDiskInfo


'
' doGetVxVMHardDiskInfo() fetches the list of harddisks by running 
' "vxdisk list -v", then feeds the hardisk names into "vxdisk diskinfo"
' to fetch the detailed information about each disk.
' Also execute "vxdmpadm arrayinfo" and "vxdmpadm pathinfo" to fetch 
' the detailed DMP information about each harddisk.
'
' strInstallDir tells us where the vxdg.exe command is installed on the host.
'
Function doGetVxVMHardDiskInfo(strInstallDir)

    Dim strCmd, strFullCmd, strDCmd, cOutput, cDOutput, intRes
    Set cOutput = New Output
    Set cDOutput = New Output

    doGetVxVMHardDiskInfo = True

    ' Wscript.Echo "in doGetVxVMHardDiskInfo"

    Dim i, strLine, strDisk, arrToks
    Dim strSubCmd, strSubFullCmd, strSubDCmd, intSubRes

    ' Call "vxdisk list" to get a list of the disks, cut column 1 to get
    ' the harddisk names, and feed each of those into "vxdisk diskinfo",
    ' "vxdmpadm pathinfo", and "vxdmpadm arrayinfo".
    strCmd = "vxdisk.exe" & chr(34) & " list -v"
    strFullCmd = chr(34) & strInstallDir & strCmd
    strDCmd = "vxdisk.exe list -v"
    Call RunCommand(strFullCmd, cOutput, False, False)
    intRes = Write_Output(strDCmd, cOutput, False)

    For i = 0 To cOutput.intOLineCount-1
        If (InStr(cOutput.strOLines(i), "Harddisk")) Then
            strLine = cOutput.strOLines(i)
            strLine = Left(strLine, Len(strLine) - 1) ' There's only a LF
            ' Wscript.Echo "dline: '" & strLine & "'"
            arrtoks = Split(strLine, " ", 2)
            strDisk = Trim(arrtoks(0))
            ' Wscript.Echo "disk: '" & strDisk & "'"

            strSubCmd = "vxdisk.exe" & chr(34) & " diskinfo " & strDisk
            strSubFullCmd = chr(34) & strInstallDir & strSubCmd
            strSubDCmd = "vxdisk.exe diskinfo " & strDisk
            Call RunCommand(strSubFullCmd, cDOutput, False, False)
            intSubRes = Write_Output(strSubDCmd, cDOutput, False)
            cDOutput.Reset()

            strSubCmd = "vxdmpadm.exe" & chr(34) & " arrayinfo " & strDisk
            strSubFullCmd = chr(34) & strInstallDir & strSubCmd
            strSubDCmd = "vxdmpadm.exe arrayinfo " & strDisk
            Call RunCommand(strSubFullCmd, cDOutput, False, False)
            intSubRes = Write_Output(strSubDCmd, cDOutput, False)
            cDOutput.Reset()

            strSubCmd = "vxdmpadm.exe" & chr(34) & " pathinfo " & strDisk
            strSubFullCmd = chr(34) & strInstallDir & strSubCmd
            strSubDCmd = "vxdmpadm.exe pathinfo " & strDisk
            Call RunCommand(strSubFullCmd, cDOutput, False, False)
            intSubRes = Write_Output(strSubDCmd, cDOutput, False)
            cDOutput.Reset()
        End If
    Next ' For i

    ' XXXXX I don't handle errors here.

End Function 'doGetVxVMHardDiskInfo


'
' WMI returns WWNs in an 8-slot array, with each slot containing the string
' representation of a decimal value.  WWNToString() converts each string to
' the string representation in hex, and packs it into the format our Explorer
' expects of WWNs.
'
Function WWNToString(arrayWWN)

On Error Resume Next

    Dim I

    WWNToString = Hex0(arrayWWN(0))
    For I = 1 To 7
        WWNToString = WWNToString & Hex0(arrayWWN(I))
    Next
End Function


'
' Hex0() accepts the string representation of a decimal value and converts it
' into the string representation of the number in hex.  If the resulting hex
' digit is only one character long, it pads it with a leading "0", to yield
' the two hex digit representation required for packing WWNs.
'
Function Hex0(intN)

On Error Resume Next

    Hex0 = Hex(intN)

    If (intN < &h10) Then Hex0 = "0" & Hex0

End Function


'
' dump_FCInfo() writes the WMI FC TargetMapping data for one LUN into the
' data file in the expected format.
'
Function dump_FCInfo(strHBAPortInstanceName, strHBAPortWWN, objEntry, I)

On Error Resume Next

    Dim objFCPId, objScsiId

    Set objFCPId = objEntry.FCPId

    objOutFile.WriteLine("instance of GetFcpTargetMapping")
    objOutFile.WriteLine("{")
    objOutFile.WriteLine(CHR(09) & "HBAKey = " & UCase(strHBAPortInstanceName))
    objOutFile.WriteLine(CHR(09) & "HBAPortWWN = " & strHBAPortWWN)
    objOutFile.WriteLine(CHR(09) & "FCID   = x" & Hex0(objFCPId.Fcid))
    objOutFile.WriteLine(CHR(09) & "FcpLun = " & objFCPId.FcpLun)
    objOutFile.WriteLine(CHR(09) & "NodeWWN = " & WWNToString(objFCPId.NodeWWN))
    objOutFile.WriteLine(CHR(09) & "PortWWN = " & WWNToString(objFCPId.PortWWN))

    Set objScsiId = objEntry.ScsiId
    objOutFile.WriteLine(CHR(09) & "ScsiBusNumber    = " & objScsiId.ScsiBusNumber)
    objOutFile.WriteLine(CHR(09) & "ScsiTargetNumber = " & _
        objScsiId.ScsiTargetNumber)
    objOutFile.WriteLine(CHR(09) & "ScsiOSLun        = " & objScsiId.ScsiOSLun)
'    ' If (vbNull = VarType(objScsiId.OSDeviceName)) Then
'    '     objOutFile.WriteLine(CHR(09) & "OS Device Name     = NULL")
'    ' Else
'    '   objOutFile.WriteLine("\tOS Device Name     = " & objScsiId.OSDeviceName)
'    '     objOutFile.WriteLine(CHR(09) & "OS Device Name     = " )
'    ' End If
    objOutFile.WriteLine("}")

End Function ' dump_FCInfo


'
' synthesize_Filename() creates the name of the output file to be written.
' The filename is of the form:
' <hostname>_<OS>_YYYYMMDDHHMM
' For example: CCS-DELL2850-7_Win_200907151342
' CCS-DELL2850-7 is hostname, "Win" is OS, year is 2009, month is 07, 
' day is 15, time is 1342 hours (24-hour clock).
' Note that the locale was set to US_English at the beginning of the script,
' so we can rely on the displayed date and time formats.
'
Function synthesize_Filename(strHostname)

On Error Resume Next

    Const strOS = "Win"
    Dim strFileName

    ' The synthesized filenames contain the time of the discovery in GMT.
    ' The ActiveTimeBias registry key is the number of minutes the timezone
    ' of this host is off of GMT, negative or positive.  Fetch it; we'll 
    ' use it to adjust the current time to GMT.
    Dim oShell, strRegKey, intOffsetMin
    Set oShell = CreateObject("WScript.Shell") 
    strRegKey = "HKEY_LOCAL_MACHINE\System\CurrentControlSet\" &_ 
          "Control\TimeZoneInformation\ActiveTimeBias" 
    intOffsetMin = oShell.RegRead(strRegKey) 

    ' Now fetch the current time, adjust by the Bias, and dump out in 
    ' the desired formats.
    Dim HereNow, GMTNow, strRawDate, strRawTime
    Dim toks, strTime, strDate, strDateTime

    ' Find out what time it is in this timezone right now.
    HereNow = Now()

    ' Adjust by the minutes we are off of GMT.
    GMTNow = DateAdd("n", intOffsetMin, HereNow)

    ' Save date and time in short formats.
    strRawDate = FormatDateTime(GMTNow, vbShortDate)
    ' strRawDate looks like this: 11/20/2009
    ' Wscript.Echo "ShortGMTDate: '" & strRawDate & "'"
    toks = Split(strRawDate, "/")
    ' toks(0) has month, 1 has day, 2 has year

    ' Must pad single-digit months or days with leading zero
    if (Len(toks(0)) = 1) Then
        toks(0) = "0" & toks(0)
    End If
    if (Len(toks(1)) = 1) Then
        toks(1) = "0" & toks(1)
    End If

    strDate = toks(2) & toks(0) & toks(1)
    ' strDate looks like this: '20091120', or this: '20100105'
    ' Wscript.Echo "strDate: '" & strDate & "'"

    strRawTime = FormatDateTime(GMTNow, vbShortTime)
    ' Wscript.Echo "ShortGMTTime: '" & strRawTime & "'"
    toks = Split(strRawTime, ":")
    strTime = toks(0) & toks(1)
    ' strTime looks like this: '0532'
    ' Wscript.Echo "strTime: '" & strTime & "'"

    ' Now concatenate Date and Time
    strDateTime = strDate & strTime
    ' strDateTime looks like this: '200911200532'
    ' Wscript.Echo "strDateTime: '" & strDateTime & "'"

    ' We have everything we need.  Synthesize the filename.
    ' <Hostname>_Win_
    strFileName = strHostname & "_" & strOS & "_"
    ' append strDateTime and ".txt" suffix
    strFileName = strFilename & strDateTime & ".txt"

    if (blnVerbose = True) Then
        objOutFile.WriteLine("# synthesize_Filename, GMT Bias = " & _
	                     intOffsetMin)
        objOutFile.WriteLine("#                      FileName: '" & _
	                     strFileName & "'")
    End If
    synthesize_Filename = strFileName

End Function ' synthesize_Filename


'
' If the user doesn't specify a hostname, our discovery will take place on
' the host on which this script is running.  getHostname() fetches that
' hostname.
'
Function getHostname()

On Error Resume Next

    Dim strCmd, strHostname, cOutput, intRetVal
    Set cOutput = New Output
    strCmd = "%SystemRoot%\system32\hostname.exe"
    intRetVal = RunCommand_Local(strCmd, cOutput, False, False)
    strHostname = cOutput.strOLines(0)
    if (blnVerbose = True) Then
        objOutFile.WriteLine "# GetHostname: hostname rval " & intRetVal & _
	                     "#  linecnt: " & cOutput.intOLineCount
        objOutFile.WriteLine "# tmpHostname: '" & strHostname & "'"
    End If

    getHostname = strHostname
End Function ' getHostname


'
' getFQHNs() fetches the Win32_NetworkAdapterConfiguration WMI class, and
' scans through it, looking for the single instance of the "IPAddress"
' attribute in one instance of the class.  The value of this attribute is
' a set of IP Addresses, each of which may be IPv4 or IPv6 format, in the
' following format:
' { "<IPAddr>", "<IPAddr>" ... };
' Once this value is obtained, the braces and semicolon are cut, it is split
' on ",", and each IP Address is fed to nslookup, mapping it to the Fully
' Qualified HostName (FQHN) to which it refers.
'
Function getFQHNs()

On Error Resume Next

    Dim objSet, objWbemObj
    Dim intMyError, objLastError
    Dim strQuery
    intMyError = 0


    ' Write our own "class" instance into the file, with each attr an 
    ' IP Address, and its value the FQHN associated with that IP.
    objOutFile.WriteLine(START & "FQHN" & SEP & SEP)
    objOutFile.WriteLine(OPUT)

    Dim strStart, strEnd
    strStart = TimeStamp("Start getFQHNs")

    objOutFile.WriteLine("instance of FQHN")
    objOutFile.WriteLine("{")

    ' IPEnabled=TRUE will select the instance that has, or instances that
    ' have, the IP information.
    ' On some hosts, we'll have one instance with IPEnabled=TRUE, and that
    ' instance will have an array of multiple IP Addresses.  On other hosts,
    ' we'll have *multiple* instances of IPEnabled classes, where each 
    ' instance has an array of a single IP Addr.  Allow for each case, and
    ' this will also handle multiple instances with arrays of multiple IPs.
    ' So, write *all* IPAddrs from *all* instances into a single FQHN struct.

    strQuery = "SELECT * FROM Win32_NetworkAdapterConfiguration WHERE IPEnabled=TRUE"
    Set objSet = objCIMV2Services.ExecQuery(strQuery)

    For Each objWbemObj In objSet
        if (Err <> 0) Then
            intMyError = Err.Number
            Set objLastError = CreateObject("WbemScripting.SWbemLastError")
        End If

        Dim intIPCnt
        intIPCnt = UBound(objWbemObj.IPAddress)

        Dim j, strIPAddr, strFQHN, strStr
        For j = 0 To intIPCnt
            ' objOutFile.WriteLine(objWbemObj.IPAddress(j))
            strIPAddr = objWbemObj.IPAddress(j)
            if (blnVerbose = True) Then
                Dim strCmd
                strCmd = "# getFQHNs: call nslookup with '" & strIPAddr & "'"
                objOutFile.WriteLine(strCmd)
            End If
            if (strIPAddr <> "") Then
                strFQHN = getFQHN(strIPAddr)
                ' <IPAddress> = "<FQHN>"
                strStr = vbTab & strIPAddr & " = " & chr(34) & strFQHN & chr(34)
                objOutFile.WriteLine(strStr)
                ' If we didn't get a /f flag, see if this FQHN is the one
                ' we need to synthesize a FQHN filename.
                If (blnGotFFlag = False) Then
                    Call checkName(strFQHN)
                End If
            End If
        Next
    Next
    ' close out the FQHN struct.
    objOutFile.WriteLine("};")

    strEnd = TimeStamp("End getFQHNs")
    Call TimeStampDiff("End getFQHNs duration ", strStart, strEnd, True)

    objOutFile.WriteLine(OPUT_END)

    ' and write any errors.
    objOutFile.WriteLine(ERRR)
    if (intMyError <> 0) Then
        objOutFile.WriteLine( "Operation = " & objLastError.operation & VBCRLF _
                     & "ParamInfo = " & objLastError.ParameterInfo & VBCRLF _
                     & "ProviderName = " & objLastError.ProviderName)
        objOutFile.WriteLine("Err = " & hex(intMyError))
    End If
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")

End Function  ' getFQHNs()


'
' getFQHN() invokes nslookup to map the IP address specified in the
' strIP parameter to its Fully Qualified HostName.  
' getFQHN() returns the hostname as a string.  If nslookup doesn't find a
' match, getFQHN() returns the string "UNCONFIGURED".
'
Function getFQHN(strIP)

On Error Resume Next

    Dim strCmd, cOutput, intRetVal
    Set cOutput = New Output
    strCmd = "%SystemRoot%\system32\nslookup.exe " & strIP
    ' If the script is exploring a remote host, run 'nslookup.exe' on that
    ' host, in order to have the best chance of resolving the IP to a valid
    ' FQHN.  If it is Exploring the localhost, run the nslookup locally.
    intRetVal = RunCommand(strCmd, cOutput, False, False)

    Dim j, toks
    getFQHN = ""
    for j = 0 To cOutput.intOLineCount-1
        ' objOutFile.WriteLine "line: '" & strLines(j) & "'"
        toks = Split(cOutput.strOLines(j), ":")
        ' objOutFile.WriteLine "toks[0]: '" & toks(0) & "'"
        ' objOutFile.WriteLine "ubound of toks: " & UBound(toks)
        If (UBound(toks) = 1) Then   ' This is an <attr>:<value> line
            ' objOutFile.WriteLine "t(0): '" & toks(0) & "', t(1): '" & toks(1) & "'"
            if (toks(0) = "Name") Then
                Dim strFQHN
		strFQHN = Trim(toks(1))
                if (blnVerbose = True) Then
                    objOutFile.WriteLine("# setting FQHN '" & strFQHN & "'")
                End If
                getFQHN = strFQHN
                Exit For
            End If
        End If
    Next
    if (getFQHN = "") Then
        getFQHN = UNCONFIGURED
    End If

End Function ' getFQHN


'
' When running in manual mode, exploring the host on which we are running,
' we won't receive the FQHN to use for the data file name.  'hostname' returns
' just the shortname, and we use that to generate the initial filename to open,
' but after we fetch the IP Addr list, we feed each one to 'nslookup', mapping
' the addr to a FQHN.  Then, we check each of those names to see if a component
' of the name matches the short hostname; if so, we have our FQHN to use for
' the filename.  In that case, at the end of the data collection after the
' original data file is closed, we 'move' the file to a new filename, which
' is synthesized using this FQHN.  checkName() performs that pattern matching
' to determine whether strFQHN is the FQHN equivalent of strSystem; if so,
' it uses that FQHN to synthesize a new filename.
'
Function checkName(strFQHN)

On Error Resume Next

    if (blnVerbose = True) Then
        objOutFile.WriteLine "# checkName: FQHN '" & strFQHN & "'"
        objOutFile.WriteLine "# checkName: strSystem '" & strSystem & "'"
    End If
    ' If we've already synthesized a new name, return immediately.
    If (strFQHNFileName <> "") Then
        Exit Function
    End If

    Dim toks, i
    ' objOutFile.WriteLine "FQHN: '" & strFQHN & "'"
    toks = Split(strFQHN, ".")
    ' for i = 0 To UBound(toks)
    '     objOutFile.WriteLine "FQHN toks(" & i & "): '" & toks(i) & "'"
    ' Next
    ' If the first component of the FQHN matches our short system name, 
    ' then we have the FQHN with which we need to generate the new hostname.
    if (toks(0) = strSystem) Then
        strFQHNFileName = synthesize_Filename(strFQHN)
        strFQHNPathnFileName = strOutDir & "\" & strFQHNFileName
    Else
        if (blnVerbose = True) Then
            objOutFile.WriteLine "# toks(0) (" & toks(0) & _
	                         "#  didn't match strSystem (" & strSystem & ")"
        End If
    End If

    if (blnVerbose = True) Then
        objOutFile.WriteLine "# strFQHNFileName: " & strFQHNFileName
        objOutFile.WriteLine "# strFQHNPathnFileName: '" & _
	                      strFQHNPathnFileName & "'"
    End If

End Function ' checkName


'
' getComputerSystem() fetches several attributes needed to synthesize Host
' data from the Win32_ComputerSystem class, and writes the data into a struct
' named ComputerSystem, which will be picked up by the parser.  We may add more
' attrs to this class as needed, but they must be fetched one by one,
' because one of the Win32_ComputerSystem attributes is OEMLogoBitmap, whose
' value may be HUGE, and which we do not need.
'
Function getComputerSystem()

On Error Resume Next

    Dim objSet, objWbemObj
    Dim intMyError, objLastError
    Dim strQuery
    intMyError = 0

    strQuery = "SELECT * FROM Win32_ComputerSystem"
    Set objSet = objCIMV2Services.ExecQuery(strQuery)
        
    For Each objWbemObj In objSet
        if (Err <> 0) Then
            intMyError = Err.Number
            Set objLastError = CreateObject("WbemScripting.SWbemLastError")
        End If

        ' Write our own Win32_ComputerSystem class instance into the file,
        ' currently with one attribute: "PhysicalMemory".
        objOutFile.WriteLine(START & "Win32_ComputerSystem" & SEP & SEP)
        objOutFile.WriteLine(OPUT)

        Dim strStart, strEnd
        strStart = TimeStamp("Start getComputerSystem")

        objOutFile.WriteLine("instance of Win32_ComputerSystem")
        objOutFile.WriteLine("{")

        ' TotalPhysicalMemory
        Dim strPhysMem, strLPs, strPs, strSysType, strStr
        strPhysMem = objWbemObj.TotalPhysicalMemory
        strStr = vbTab & "PhysicalMemory = " & chr(34) & strPhysMem & chr(34)
        objOutFile.WriteLine(strStr)
        ' NumberOfLogicalProcessors may not exist on host, but we catch errors.
        strLPs = objWbemObj.NumberOfLogicalProcessors
        strStr = vbTab & "NumberOfLogicalProcessors = " & chr(34) & strLPs & chr(34)
        objOutFile.WriteLine(strStr)
        ' NumberOfProcessors
        strPs = objWbemObj.NumberOfProcessors
        strStr = vbTab & "NumberOfProcessors = " & chr(34) & strPs & chr(34)
        objOutFile.WriteLine(strStr)
        ' SystemType
        strSysType = objWbemObj.SystemType
        strStr = vbTab & "SystemType = " & chr(34) & strSysType & chr(34)
        objOutFile.WriteLine(strStr)

        objOutFile.WriteLine("};")

        strEnd = TimeStamp("End getComputerSystem")
        Call TimeStampDiff("End getComputerSystem duration ", _
                           strStart, strEnd, True)

        objOutFile.WriteLine(OPUT_END)
    Next

    objOutFile.WriteLine(ERRR)
    if (intMyError <> 0) Then
        objOutFile.WriteLine( "Operation = " & objLastError.operation & VBCRLF _
                     & "ParamInfo = " & objLastError.ParameterInfo & VBCRLF _
                     & "ProviderName = " & objLastError.ProviderName)
        objOutFile.WriteLine("Err = " & hex(intMyError))
    End If
    objOutFile.WriteLine(ERRR_END)
    objOutFile.WriteLine("")
End Function  ' getComputerSystem()



' --------------------------------------------------------------------------
' --------------------------------------------------------------------------
' --------------------------------------------------------------------------
' --------------------------------------------------------------------------


'
' RunCommand() runs the command specified by 'strCmd' on either the local
' host or the host specified by strSystem, depending upon whether we are
' performing local or remote discovery (i.e., whether 'blnLocalExploration'  
' is True or False).  This class of commands consist of those that perform
' the discovery or map IP addrs to FQHNs, like 'fcinfo.exe' and 'nslookup'.
' - cOutput class instance will receive the output and error output of the cmd.
' - blnSilent suppresses all output IFF true (primarily to ensure that section-
'   parsing requirements are honored).
' - blnNoOutput tells the called routines that the command will not return
'   any data.  This prevents unnecessary attempts to fetch nonexistent data
'   across the network, but is primarily due to the fact that remote commands
'   are executed by redirecting the output to a file, which prevents commands
'   that also require output redirection from working correctly.  For example,
'   the LDM code creates files of commands on the remote host via "echo > file"
'   commands, and these commands do not work correctly if the remote execution
'   logic also redirects to file.
' 
' Note that some commands must *always* be run on the local host, and for
' those commands we directly invoke RunCommand_Local().
'
Function RunCommand(strCmd, ByRef cOutput, blnSilent, blnNoOutput)

On Error Resume Next

    Dim intRetVal

    if (blnLocalExploration = False) Then
        if (blnVerbose = True And blnSilent = False) Then
            objOutFile.WriteLine "#  RunCommand Remote Command: '" & _
	                         strCmd & "'"
        End If
        intRetVal = RunCommand_Remote(strCmd, strSystem, cOutput, blnSilent, _
	                              blnNoOutput)
    Else
        if (blnVerbose = True And blnSilent = False) Then
            objOutFile.WriteLine "#  RunCommand Local Command: '" & _
	                         strCmd & "'"
        End If
        intRetVal = RunCommand_Local(strCmd, cOutput, blnSilent, blnNoOutput)
    End If

    if (blnVerbose = True And blnSilent = False And blnNoOutput <> True) then
        if (intRetVal = 0) Then
            Call Dump_Output(cOutput)
        Else
            objOutFile.WriteLine("RunCommand '" & strCmd & "' rc: " & intRetVal)
        End If
    End If

    RunCommand = intRetVal

End Function ' RunCommand


'
' RunCommand_Local() executes the command specified by the 'strCommand'
' parameter on the local host, returning the stdout and stderr output from
' the command in the arrays contained in the Output class, which is passed
' by reference to the routine.  stdout is read into 'strOLines'; stderr into
' 'strELines'.
' RunCommand_Local() ReDims the arrays as needed to accommodate the number
' of lines returned by the command, so they may be larger than the number
' of valid lines they contain.  Therefore, it sets 'intOLineCount' and
' 'intELineCount' to the number of valid lines in strOLines() and 
' strELines(), resp.
' If blnSilent is true, no messages are logged even if blnVerbose is True.
'
' RunCommand_Local() returns the Exit status (ExitCode) of the command.
'
Function RunCommand_Local(strCommand, ByRef pOutput, blnSilent, blnNoOutput)

On Error Resume Next

    Dim strFullCmd
    strFullCmd = "%SystemRoot%\system32\cmd.exe /c " & strCommand

    if (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine "#  RC_Local Raw Command: '" & strCommand & "'"
        objOutFile.WriteLine "#  RC_L Full Command: '" & strFullCmd & "'"
    End If

    Dim objWshShell, objExec, strLine
    Dim intInitSize, intCurOSize, intCurESize, intInc, i
    intInc = 40       ' bump the array size this much for each ReDim()
    intInitSize = 40  ' initial size for pOutput and Error arrays
    intCurOSize = intInitSize
    intCurESize = intInitSize

    pOutput.ReDimO(intInitSize)
    pOutput.ReDimE(intInitSize)
    Set objWshShell = WScript.CreateObject("WScript.Shell")

    ' Run the command.
    Set objExec = objWshShell.Exec(strFullCmd)
    if (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine "#  RC_L Exec status: " & objExec.Status
        objOutFile.WriteLine "#  RC_L ProcessID: " & objExec.ProcessID
    End If

    ' If we're supposed to capture the output...
    if (blnNoOutput <> True) Then
        i = 0
        ' ...read stdout into strOLines()
        Do While objExec.StdOut.AtEndOfStream <> True
            strLine = objExec.StdOut.Readline()
            ' objOutFile.WriteLine "line: '" & strLine & "'"
            if (i >= intCurOSize) Then
                ' objOutFile.WriteLine "bumping by " & intInc
                pOutput.ReDimO(intCurOSize+intInc)
                intCurOSize = intCurOSize + intInc
        	End If
            pOutput.strOLines(i) = strLine
            i = i + 1
        Loop
        pOutput.intOLineCount = i

        ' Then read stderr into strELines()
        i = 0
        Do While objExec.StdErr.AtEndOfStream <> True
            strLine = objExec.StdErr.Readline()
            ' objOutFile.WriteLine "line: '" & strLine & "'"
            if (i >= intCurESize) Then
                ' objOutFile.WriteLine "bumping by " & intInc
                pOutput.ReDimE(intCurESize+intInc)
                intCurESize = intCurESize + intInc
        	End If
            pOutput.strELines(i) = strLine
            i = i + 1
        Loop
        pOutput.intELineCount = i
    End If

    ' Status is set to zero when the command is spawned, and turns non-zero
    ' when the command has exited, after which we can get a valid ExitCode.
    Do While objExec.Status = 0
        'objOutFile.WriteLine "Sleep..."
        Wscript.Sleep 100  ' 100 milliseconds == 1/10 of a second.
    Loop

    RunCommand_Local = objExec.ExitCode
    
    If (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine "#  RC_L ExitCode: " & objExec.ExitCode
        objOutFile.WriteLine("#  RC_L OLine count: " & pOutput.intOLineCount)
        objOutFile.WriteLine("#  RC_L ELine count: " & pOutput.intELineCount)
        Dump_Output(pOutput)
    End If

End Function ' RunCommand_Local
'
' RunCommand_Remote() executes the command specified by the 'strCommand'
' parameter on the host specified by 'strHostname', returning the stdout
' and stderr output from the command in the arrays contained in the Output
' class, which is passed by reference to the routine.  stdout is read into
' 'strOLines'; stderr into 'strELines'.
' If blnSilent is true, no messages are logged even if blnVerbose is True.
'
' XXXXXX RunCommand_Remote() returns some status of the invoked command, 
' although that isn't the exitstatus of the process unfortunately (yet!).
'
const TYPE_RETRIES = 120  ' retry the "type" command for 120 seconds (2 mins)

Function RunCommand_Remote(strCommand, strHostname, ByRef pOutput, _
	                   blnSilent, blnNoOutput)

On Error Resume Next

    if (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine "#  RC_Remote: Command: '" & strCommand & "'"
        objOutFile.WriteLine "#             System: '" & strHostname & "'"
    End If

    ' Randomizing the output filenames seems to moderate the problem when 
    ' MS and remote host are both 2008, and the 'type' command can't find
    ' the output filename.  Seems to be some OS caching problem.
    DIM strOutPnF, strOutFile, strErrPnF, strErrFile
    intFCount = intFCount + 1
    strOutFile = "ranitOut" & intFCount & ".txt"
    strErrFile = "ranitErr" & intFCount & ".txt"
    ' Wscript.Echo "outfile: '" & strOutFile & "'"
    ' Wscript.Echo "errfile: '" & strErrFile & "'"

    ' We always write the file in the remote host's %SystemRoot%\Temp
    ' directory, because 1) that always exists, and 2) it is directly
    ' under Admin$ filesystem, which we will use to snarf the file.
    strOutPnF = "%SystemRoot%\Temp\" & strOutFile
    strErrPnF = "%SystemRoot%\Temp\" & strErrFile
    ' if (blnVerbose = True And blnSilent = False) Then
    '     objOutFile.WriteLine "#  (RC_R Output path+file: '" & strOutPnF & "')")
    '     objOutFile.WriteLine "#  (RC_R err path+file: '" & strErrPnF & "')")
    ' End If

    Dim intORes, intCmdRes, strCmd, k
    ' Run command on remote host, creating, not appending to, the output file.
    intCmdRes = ExecRemote(strCommand, strHostname, strOutPnF, strErrPnF, _
                           False, blnSilent, blnNoOutput)

    ' Now, IFF blnNoOutput is False, type the output and error files into 
    ' our Class arrays and return.
    If (blnNoOutput <> True) Then
        ' First fetch the contents of the stdout file.
        strCmd = "type \\" & strHostname & "\admin$\Temp\" & strOutFile
        if (blnVerbose = True And blnSilent = False) Then
            objOutFile.WriteLine("#  (RC_R: type Ocmd: '" & strCmd & "')")
        End If

        ' When both the MS and the remote host are Windows 2008, sometimes the
        ' output file isn't available on the admin$ share for the 'type' 
        ' command even after the executing process has exited.  It appears to
        ' be some kind of OS caching problem (which never happens when either
        ' MS or remote host or both are 2003).  So if the 'type' command fails,
        ' retry it in a loop, sleeping for 1 second between tries.
        Dim blnOSucceeded, intORetries
        blnOSucceeded = False
        intORetries = 0
        For k = 0 to TYPE_RETRIES
            intORes = RunCommand_Local(strCmd, pOutput, blnSilent, False)
            ' Wscript.Echo "type res: " & intORes
            If (intORes <> 0) Then  ' failed: sleep for second & then try again.
                intORetries = intORetries + 1
                ' Wscript.Echo "type failed."
                Wscript.Sleep 1000   ' 1000 ms == 1 second.
                pOutput.Reset()
            Else ' succeeded: pop out of the loop.
                ' Wscript.Echo "type succeeded."
                blnOSucceeded = True
                Exit For
            End If
        Next

        ' Then the contents of the stderr file.
        strCmd = "type \\" & strHostname & "\admin$\Temp\" & strErrFile
        if (blnVerbose = True And blnSilent = False) Then
            objOutFile.WriteLine("#  (RC_R: type Ecmd: '" & strCmd & "')")
        End If

        Dim intERes, blnESucceeded, intERetries
        blnESucceeded = False
        intERetries = 0
        Dim cErrOutput
        Set cErrOutput = New Output
        For k = 0 to TYPE_RETRIES
            intERes = RunCommand_Local(strCmd, cErrOutput, blnSilent, False)
            ' Wscript.Echo "type res: " & intERes
            If (intERes <> 0) Then  ' failed: sleep for second & then try again.
                intERetries = intERetries + 1
                ' Wscript.Echo "type failed."
                Wscript.Sleep 1000   ' 1000 ms == 1 second.
                cErrOutput.Reset()
            Else ' succeeded: pop out of the loop.
                ' Wscript.Echo "type succeeded."
                blnESucceeded = True
                Exit For
            End If
        Next
        ' The contents of the stderr file on the remote machine is now in
        ' cErrOutput.strOLines().  Copy the data from cErrOutput.strOLines()
        ' to pOutput.strELines(), and set pOutput.intELineCount.

        ' Wscript.Echo "copying lines: " & cErrOutput.intELineCount
        Dim x
        ' Make sure the pOutput.strELines() array is large enough.
        pOutput.ReDimE(cErrOutput.intOLineCount)
        For x = 0 to cErrOutput.intOLineCount-1
            pOutput.strELines(x) = cErrOutput.strOLines(x)
        Next
        pOutput.intELineCount = cErrOutput.intOLineCount

        ' If blnSilent is True, we can only return success or an error.  
        ' But if blnSilent is False, then: 
        '  - if the 'type' command failed, or succeeded after initially failing,
        '    display an error message.
        '  - if it succeeded the first time, display a message IFF blnVerbose
        '    is True.
        If (blnSilent = False) then
            if (blnOSucceeded = False) then
                objOutFile.WriteLine("#!!RC_R: type of '" & strCommand & _
                                     "' failed after " & _
                                     intORetries & " retries, err " & intORes)
            Elseif (intORetries <> 0) Then
                objOutFile.WriteLine("#!!RC_R: type of '" & strCommand & _
                                     "' succeeded after " & _
                                     intORetries & " retries")
            Elseif (blnVerbose = True) Then
                objOutFile.WriteLine("#  (RC_R: type of '" & strCommand & _
                                     "' succeeded)")
            End If
        End If

        ' Now that we have the contents of the remote files, delete them.
        ' Create a new Output class for the output of the commands which we
        ' will run, but from which we only need the exit status, not output.
        Dim cDiscard, intDres
        ' First the stdout file
        Set cDiscard = New Output
        strCmd = "del \\" & strHostname & "\admin$\Temp\" & strOutFile
        intDRes = RunCommand_Local(strCmd, cDiscard, blnSilent, True)
        if (intDRes <> 0) Then
            if (blnSilent = False) Then
                objOutFile.WriteLine("#  !!RC_R: Remote OFile delete failed, err " & _
                                     intDRes)
            End If
        Else
            if (blnVerbose = True And blnSilent = False) Then
                objOutFile.WriteLine("#  (RC_R: Remote OFile delete succeeded)")
            End If
        End If ' intDRes <> 0
        ' Now the stderr file.
        cDiscard.Reset()
        strCmd = "del \\" & strHostname & "\admin$\Temp\" & strErrFile
        intDRes = RunCommand_Local(strCmd, cDiscard, blnSilent, True)
        if (intDRes <> 0) Then
            if (blnSilent = False) Then
                objOutFile.WriteLine("#  !!RC_R: Remote EFile delete failed, err " & _
                                     intDRes)
            End If
        Else
            if (blnVerbose = True And blnSilent = False) Then
                objOutFile.WriteLine("#  (RC_R: Remote EFile delete succeeded)")
            End If
        End If ' intDRes <> 0

    End If  ' blnNoOutput <> True

    RunCommand_Remote = intCmdRes

End Function 'RunCommand_Remote


'
' ExecRemote() executes the command contained in 'strCommand' on the remote
' host 'strHostname', writing the output to a file on the remote host whose
' path and name are specified by 'strOutPnF'.
' If blnAppend is True, the output of the command is *appended* to the file;
' if False, the file is truncated and written with this new output only.
' If blnSilent is true, no messages are logged even if blnVerbose is True.
'
' ExecRemote() returns either the result code of the process creation (which
' is *not* the exit status of the executed command!) or the result of the
' Wait() command if that command returned non-zero (i.e., the process hung).
'
Function ExecRemote(strCommand, strHostname, strOutPnF, strErrPnF, _
                    blnAppend, blnSilent, blnNoOutput)

On Error Resume Next
' On Error goto 0

    if (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine "#   ER Raw Command: '" & strCommand & "'"
        objOutFile.WriteLine "#   ER Hostname: '" & strHostname & "'"
        objOutFile.WriteLine "#   ER OutPnF: '" & strOutPnF & "'"
        objOutFile.WriteLine "#   ER Append: " & blnAppend
    End If

    ' We have a chicken-and-egg problem.  exe's must run in a cmd.exe shell,
    ' so we have to find cmd.exe.  It is under %SystemDrive%, but that is
    ' an environment variable, and only cmd.exe on the remote machine can
    ' expand it, but we can't find cmd.exe to do the expansion.  But we have
    ' WMI!!  WMI's Win32_OperatingSystem has the SystemDirectory value,
    ' which is %SystemDrive%.  Fetch it and use it as the path to cmd.exe.
    ' Save it in a global variable so that we don't have to fetch it again
    ' for subsequent remote commands.
    DIM strQuery, objSet, objWbemObj
    if (strSystemDir = "") Then
        strQuery = "SELECT * FROM Win32_OperatingSystem"
        Set objSet = objCIMV2Services.ExecQuery(strQuery)
        
	DIM intMyError, objLastError
        For Each objWbemObj In objSet
            if (Err <> 0 And blnSilent = False) Then
                intMyError = Err.Number
                Set objLastError = CreateObject("WbemScripting.SWbemLastError")
                objOutFile.WriteLine("Op = " & objLastError.operation & VBCRLF _
                     & "ParamInfo = " & objLastError.ParameterInfo & VBCRLF _
                     & "ProviderName = " & objLastError.ProviderName)
                objOutFile.WriteLine("Err = " & hex(intMyError))
		' Failed!!  Our attempt to run the process probably will fail
		' also, but let's try the normal default of C:\WINDOWS\system32.
		strSystemDir = "C:\WINDOWS\system32"
                objOutFile.WriteLine "# Try strSystemDir='" & strSystemDir & "'" 
            Else
                strSystemDir = objWbemObj.SystemDirectory
                if (blnVerbose = True And blnSilent = False) Then
                    objOutFile.Writeline "# strSystemDir='" & strSystemDir & "'" 
                End If
            End If
	    Exit For
	Next
    End If

    Const SW_NORMAL = 1

    DIM strFullCmd
    strFullCmd = strSystemDir & "\cmd.exe /c " & strCommand
    ' If blnNoOutput is true, either the command won't generate any output,
    ' or else the user doesn't care about the output.  Usually this is because
    ' the command itself redirects output to a file, which breaks if we also
    ' redirect the output.
    If (blnNoOutput <> True) Then
        ' Wscript.Echo "APPEND!: " & strCommand
        ' The caller may want us to append to the file rather than create it
	' anew in order to add the output of several commands before 
	' retrieving it.  Conversely, they may need to get the output of this
	' command back in order to know what command to invoke next.  
	' blnAppend tells which to do.
        If (blnAppend = True) Then
            strFullCmd = strFullCmd & " >>" & strOutPnF & " 2>>" & strErrPnF
        else
            strFullCmd = strFullCmd & " >" & strOutPnF & " 2>" & strErrPnF
        End If
    ' Else
    ' Wscript.Echo "DO NOT APPEND!: " & strCommand
    End If
    if (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine "#   ER Full Command: '" & strFullCmd & "'"
    End If

    DIM objStartup, objConfig, objProcess, intReturn, intProcessID, intRC

    Set objStartup = objCIMV2Services.Get("Win32_ProcessStartup")
    Set objConfig = objStartup.SpawnInstance_
    objConfig.ShowWindow = SW_NORMAL

    ' Create process and run command
    Set objProcess = objCIMV2Services.Get("Win32_Process")
    intReturn = objProcess.Create _
        (strFullCmd, strSystemDir, objConfig, intProcessID)
    If (intReturn <> 0) then
        if (blnSilent = False) Then
            objOutFile.WriteLine "# !! ER Remote process couldn't be created." & _
                     vbNewLine & "#   ER Command line: " & strCommand & _
                     vbNewLine & "#   ER Return value: " & intReturn
        End If
    Else
        if (blnVerbose = True And blnSilent = False) Then
            objOutFile.WriteLine "#   ER Remote process " & intProcessID & _
                                 " created." & vbNewLine & _
                                 "#   ER Command line: " & strCommand 
        End If
    End If

    ' Wait for the remote process to exit before returning
    Dim intWRetVal
    intWRetVal = Wait(strHostname, intProcessID, blnSilent)
    ' Wscript.Echo "Wait returned " & intWRetVal

    if (intWRetVal = 0) Then
        intRC = intReturn
    Else
        intRC = intWRetVal
    End If

    if (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine("#   ER process " & intProcessID & _
                             " Exitval = " & intRC)
    End If

    ExecRemote = intRC

End Function ' ExecRemote


'
' Wait() determines when the process running on 'strHostname' whose ProcessID
' is 'intPID' has finished and exited.  It does this by repeatedly fetching
' that PID via the Win32_Process WMI call.  When it no longer exists, we
' know that the process has exited.
' This is necessary because the Win32_ProcessStartup call creates the remote
' process and returns immediately.  
' Wait() implements a polling interval at which cadence it checks for the
' existence of the PID and then sleeps; and a timeout, after which it returns
' to the caller even though the process is still running.  This prevents 
' our script from hanging if commands it attempts to execute hang.
'
' If blnSilent is true, no messages are logged even if blnVerbose is True.
'
' Wait() returns 0 if the process exits within the timeout period, else 1.
'
' Currently, the timeout period is 8 minutes (Minute*8)
'
Function Wait(strHostname, intPID, blnSilent)

On Error Resume Next

    const Tenth_Second=100    ' 100 milliseconds is 1/10 of a second
    const Second=1000         ' 1000 milliseconds is a second
    const Minute=60000        ' 60 * 1000 milliseconds is a minute

    DIM strQuery, objSet, objWbemObj, intProcCount, i, intRetVal
    intRetVal = 1  ' set to 0 when the process exits.

    if (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine("#     -> Wait for process " & intPID & _
	                     " on host " & strHostname)
    End If
    ' We'll wait for 8 minutes for the remote process to exit.  Every 1/10
    ' of a second, invoke a remote Win32_Process query, selecting for the
    ' Process ID.  If the remote process no longer exists, return 0.
    ' If it still exists, sleep for another tenth of a second and check 
    ' again. If we count 8 minutes and the process is still running, assume
    ' it is hung and return 1.
    For i = 0 To (Minute*8) Step Tenth_Second
        strQuery = "SELECT * FROM Win32_Process where ProcessId=" & intPID
        Set objSet = objCIMV2Services.ExecQuery(strQuery)
        intProcCount = objSet.count 

        ' Wscript.Echo "IPC: " & intProcCount
        ' If the process has exited, pop out of the loop and return 0.
        if (intProcCount <= 0) Then
            intRetVal = 0    ' process has exited; return 0
            Exit For
        End If
        ' objOutFile.WriteLine "Wait: count " & intProcCount
        Wscript.Sleep Tenth_Second
    Next
    ' If we exited the loop after exhausting the entire wait period, 
    ' intRetVal is still set to 1; if we exited early, we set it to 0.

    if (blnVerbose = True And blnSilent = False) Then
        if (intRetVal = 0) Then
            objOutFile.WriteLine("#     -> Process " & intPID & " on host " & _
	                         strHostname & " has finished")
        Else
            objOutFile.WriteLine("#     -> Process " & intPID & " on host " & _
	                         strHostname & " hung; return 1")
        End If
    End If

    Wait = intRetVal

End Function 'Wait


'
' Map_Filesystem() maps the admin$ filesystem of the host specified in
' 'strHostname'.  Before executing the 'net use' command, it checks whether
' it can 'dir' the admin$\Temp filesystem; if it can, the filesystem is
' already mapped and it returns.  Else it runs 'net use' to map it.
' If blnSilent is true, no messages are logged even if blnVerbose is True.
'
' Map_Filesystem() returns 0 if the mapping succeeds, else non-zero.
'
Function Map_Filesystem(strHostname, blnSilent)

On Error Resume Next

    DIM strCmd, intRes, intReturnVal
    intReturnVal = 0  ' default to success unless mapping is performed & fails

    ' Check whether we can access the remote admin$\Temp directory.
    ' If so, we've already mapped it; return success.  If not, map it now.

    ' Create a new Output class for the output of the cmd, which we'll discard
    Dim cDiscard
    Set cDiscard = New Output
    strCmd = "dir \\" & strHostname & "\admin$\Temp"
    intRes = RunCommand_Local(strCmd, cDiscard, blnSilent, True)
    if (blnVerbose = True And blnSilent = False) Then
        objOutFile.WriteLine "# Result of admin$\Temp dir: " & intRes
    End If
    if (intRes <> 0) Then   ' it isn't mapped, so map it.
        strCmd = "%SystemRoot%\system32\net.exe use \\" & strHostname
	strCmd = strCmd & "\admin$ " & strPassword & " /user:" & strUser
        if (blnVerbose = True And blnSilent = False) Then
	    objOutFile.WriteLine "# Map the drive, command: '" & strCmd & "'"
        End If
        cDiscard.Reset()
        intReturnVal = RunCommand_Local(strCmd, cDiscard, blnSilent, True)
        if (intReturnVal <> 0 And blnSilent = False) Then
            objOutFile.WriteLine("# Map failed, error " & intReturnVal)
        End If
    Else   ' 'dir' cmd succeeded, it's already mapped for us.
        if (blnVerbose = True And blnSilent = False) Then
            objOutFile.WriteLine("# Drive is already mapped")
        End If
    End If

    Map_Filesystem = intReturnVal

End Function ' Map_Filesystem


'
' UnMap_Filesystem() executes the 'net use /delete' command to unmap the
' admin$ filesystem.
' If blnSilent is true, no messages are logged even if blnVerbose is True.
' 
' It returns the exit status of the 'net use' command.
'
Function UnMap_Filesystem(strHostname, blnSilent)

On Error Resume Next

    DIM strCmd, intRes
    strCmd = "%SystemRoot%\system32\net.exe use \\" & strHostname
    strCmd = strCmd & "\admin$ /delete"
    if (blnVerbose = True And blnSilent = False) Then
        Wscript.Echo "# Unmap command: '" & strCmd & "'"
    End If

    ' Create a new Output class for the output of the cmd, which we'll discard
    Dim cDiscard
    Set cDiscard = New Output
    intRes = RunCommand_Local(strCmd, cDiscard, blnSilent, True)
    if (intRes <> 0 And blnSilent = False) Then
        Wscript.Echo("# Unmap failed, error " & intRes)
    End If

    UnMap_Filesystem = intRes

End Function ' UnMap_Filesystem


'
' Dump_Output() displays as comment lines (i.e., prepended by "# ") the
' contents of the stdout and stderr arrays that are filled by 
' RunCommand_Local()
'
Function Dump_Output(ByRef pOutput)

On Error Resume Next

    Dim l
    objOutFile.WriteLine("#  Dump_Output: Output line count: " & _
                          pOutput.intOLineCount)
    for l = 0 To pOutput.intOLineCount-1
        objOutFile.WriteLine("#   '" & pOutput.strOLines(l) & "'")
    Next

    objOutFile.WriteLine("#  Dump_Output: Stderr line count: " & _
                          pOutput.intELineCount)
    for l = 0 To pOutput.intELineCount-1
        objOutFile.WriteLine("#   '" & pOutput.strELines(l) & "'")
    Next

End Function 'Dump_Output


'
' Do_Probe() is run when RHWindowsScript.vbs is invoked with the /i (inquire)
' command-line option.  It checks that the supplied hostname, username, and
' password will allow the CIMV2 and WMI namespaces to be opened, and fetches
' the Win32_OperatingSystem class both to retrieve the OS version (2000, 
' 2003, 2008, etc.) and to verify that the class can be fetched, and then
' fetches the Win32_DiskDrive class, because fetching that class requires
' admin privileges (which Win32_OperatingSystem class does not).
' Finally, checks that the host's admin$ filesystem can be mapped IFF
' performing a remote probe, because some of the discovery will require
' that data written to a remote file by executing utilities can be fetched
' back to the Collector host.
'
Function Do_Probe(strSystem, strUser, strPassword)

On Error Resume Next

    Dim objSWbemLocator, intCOError, intCIMV2Error, intWMIError
    Dim strNameSpace, strStart, strEnd
    Dim intDiff

    ' Time the probe.  Slow probes mean slow discovery.
    strStart = now

    Set objSWbemLocator = CreateObject("WbemScripting.SWbemLocator")
    if (Err <> 0) Then
        intCOError = Err.Number
        objOutFile.WriteLine("CreateLocator=NOK(" & hex(intCOError) & ")")
    Else
        objOutFile.WriteLine("CreateLocator=OK")
    End If

    ' Try connecting to the hosts's CIMV2 namespace
    strNameSpace = "root\CIMV2"
    Set objCIMV2Services = objSWbemLocator.ConnectServer(strSystem, _
                                                       strNameSpace, _
                                                       strUser, _
                                                       strPassword)
    if (Err <> 0) Then
        intCIMV2Error = Err.Number
        objOutFile.WriteLine("CIMV2_Namespace=NOK(" & hex(intCIMV2Error) & ")")
    Else
        objOutFile.WriteLine("CIMV2_Namespace=OK")
    End If
    objCIMV2Services.Security_.ImpersonationLevel = 3

    ' ...And to the WMI namespace
    strNameSpace = "root\WMI"
    Set objWMIServices = objSWbemLocator.ConnectServer(strSystem, _
                                                       strNameSpace, _
                                                       strUser, _
                                                       strPassword)
    if (Err <> 0) Then
        intWMIError = Err.Number
        objOutFile.WriteLine("WMI_Namespace=NOK(" & hex(intWMIError) & ")")
    Else
        objOutFile.WriteLine("WMI_Namespace=OK")
    End If
    objWMIServices.Security_.ImpersonationLevel = 3

    ' GetOSType() fetches Win32_OperatingSystem, which any WMI priviledge 
    ' can do.
    Dim intOSRes
    intOSRes = GetOSType()
    If (intOSRes = 0) Then
        objOutFile.WriteLine("OSClass=OK")
        objOutFile.WriteLine("OSVersion=" & strOSType)
    Else
        objOutFile.WriteLine("OSClass=NOK(" & hex(intOSRes) & ")")
        objOutFile.WriteLine("OSVersion=UNKNOWN")
    End If

    ' Try fetching Win32_DiskDrive; that won't fetch unless full admin creds.
    ' The DiskDrive class also contains the information we need to determine
    ' whether the host is a VMWare Virtual Machine, so set that with this
    ' call, too.
    Dim intFetchDDRes, blnIsVM
    blnIsVM = False  ' default to no.

    intFetchDDRes = Fetch_DiskDrive(blnIsVM)   ' blnIsVM is passed by ref.
    If (intFetchDDRes = 0) Then
        objOutFile.WriteLine("DiskDriveClass=OK")
	objOutFile.WriteLine("IsVirtualMachine=" & blnIsVM)
    Else
        objOutFile.WriteLine("DiskDriveClass=NOK(" & hex(intFetchDDRes) & ")")
    End If

    ' If we are doing a remote exploration, map the admin$ filesystem.  
    ' Don't move this step out of this routine, or the verbose output will
    ' violate the constraints of the data file.
    Dim intMapResult
    if (blnLocalExploration = False) Then
        intMapResult = Map_Filesystem(strSystem, True)
        if (intMapResult = 0) Then
            objOutFile.WriteLine("adminShare=OK")
        Else
            objOutFile.WriteLine("adminShare=NOK(" & intMapResult & ")")
        End If
    End If

    ' If the host we're exploring is Windows 2000 or 2003, we use fcinfo.exe
    ' to obtain the FCTargetMapping information.  So, if the host is one of
    ' those platforms, check whether fcinfo.exe is installed and indicate
    ' that in the probe results.  HOWEVER, if the host is a VMWare Virtual
    ' Machine, it can't have FC data, so we don't need fcinfo.
    ' NOTE: If probing a remote host, this check requires that Map_Filesystem()
    ' has been called, and if it failed, our test for fcinfo is invalid.
    If (strOSType = "2000" OR strOSType = "2003") Then
        ' If VM, don't bother checking for fcinfo; no FC discovery to do.
	if (blnIsVM = True) Then
            objOutFile.WriteLine("fcinfo=NotApplicable")
	ElseIf (intMapResult <> 0) Then
            ' If we couldn't map admin$ share, we can't determine whether our
            ' remote 'dir' of fcinfo.exe succeeded, so test is inconclusive.
            objOutFile.WriteLine("fcinfo=INCONCLUSIVE")
        Else
            Dim blnfcinfoInstalled
            blnfcinfoInstalled = File_Found("%SystemRoot%\system32", _
	                                    "fcinfo.exe", True)
            if (blnfcinfoInstalled = True) Then
                objOutFile.WriteLine("fcinfo=OK")
            Else
                objOutFile.WriteLine("fcinfo=NOK")
            End If
        End If
    Else
        objOutFile.WriteLine("fcinfo=WMI")
    End If

    Dim intUNMapResult
    if (blnLocalExploration = False) Then
        intUNMapResult = UnMap_Filesystem(strSystem, True)
    End If

    strEnd = now
    intDiff = Int(DateDiff("s", strStart, strEnd))
    objOutFile.WriteLine("ProbeDurationInSecs=" & intDiff)

    objOutFile.Close
    Wscript.Quit 0

End Function 'Do_Probe


'
' File_Found() determines whether the file specified by strFile is found in
' the directory specified by strPath on the host being explored.  It 
' determines this by running a 'dir' command and wading through the returned
' filenames.  Note that this 'dir' turns into a local command if 
' RHWindowsScript.vbs is running on the host being explored, and a remote
' command if it is running on behalf of the Collector.
'
Function File_Found(strPath, strFile, blnSilent)

On Error Resume Next

    Dim strCmd, cOutput, intRetVal
    Dim blnFileFound
    blnFileFound = False

    Set cOutput = New Output
    strCmd = "dir " & chr(34) & strPath & "\" & strFile & chr(34)

    ' We want to determine whether the file is installed *on the host being
    ' explored*, so run the 'dir' command locally if the script is exploring
    ' the host on which it is running; else remotely.
    intRetVal = RunCommand(strCmd, cOutput, blnSilent, False)
    ' Wscript.Echo "FileFound dir intRetVal: " & intRetVal

    ' Wscript.Echo "Elinecnt: '" & cOutput.intELineCount
    Dim I
    For I = 0 To cOutput.intOLineCount-1
        ' Wscript.Echo "line: '" & cOutput.strOLines(I) & "'"
        If (InStr(cOutput.strOLines(I), strFile)) Then
	    blnFileFound = True
            ' objOutFile.WriteLine "# VVVV MATCH: " & cOutput.strOLines(I)
            Exit For
        End If
    Next ' For I

    File_Found = blnFileFound

End Function 'File_Found


Const VSTART =       "|Start|"
Const VOUTPUT =      "|Output|"
Const VOUTPUT_END =  "|Output End|"
Const VERROR =       "|Error|"
Const VERROR_END =   "|Error End|"
'
' Write_Output writes the stdout and stderr contents of the specified Output
' class into the data file, using the following VxVM and VxDMP convention:
'
' |Start|<command>|<integer Result>|
' |Output|
' <stdout lines>
' |Output End|
' |Error|
' <stderr lines>
' |Error End|
'
' <command> is specified by 'strCommand', and the stdout and stderr lines
' are from the Output class's strOLines() and strELines() arrays.
'
' Note that the above sub-sections are all nested within the 
' "|##|Output|##| and |##|Output End|##| super-sections.
'
' Write_Output() returns the number of lines in the stderr buffer 
' (cOutput.intELineCount) to indicate whether the command being written
' succeeded or failed.
'
Function Write_Output(strCommand, cOutput, blnSilent)

On Error Resume Next

    ' Write the header including command and number of lines in stderr buffer
    objOutFile.WriteLine VSTART & strCommand & "|" & cOutput.intELineCount & "|"
    objOutFile.WriteLine VOUTPUT

    Dim i
    For I = 0 To cOutput.intOLineCount-1
        ' Wscript.Echo "xOline: '" & cOutput.strOLines(I) & "'"
        objOutFile.WriteLine cOutput.strOLines(I)
    Next ' For I

    objOutFile.WriteLine VOUTPUT_END
    objOutFile.WriteLine VERROR

    For I = 0 To cOutput.intELineCount-1
        ' Wscript.Echo "xEline: '" & cOutput.strELines(I) & "'"
        objOutFile.WriteLine cOutput.strELines(I)
    Next ' For I

    objOutFile.WriteLine VERROR_END

    ' Wscript.Echo "WO res: " & cOutput.intELineCount

    Write_Output = cOutput.intELineCount

End Function 'Write_Output


'
' Local administrator privileges are required in order to fully discover hosts.
' WMI's Win32_DiskDrive class requires Administrator privileges to fetch,
' unlike many other classes, which non-admin accounts can access.  
' Fetch_DiskDrive(), called by Do_Probe(), uses this fact to determine 
' whether the specified account has the privileges required to discover 
' the host in order to make the probe command more comprehensive.
' In addition, we use the Model attr of the DiskDrive class to determine
' whether this is a VM.  Fetch_DiskDrive() sets blnIsVM to True if it matches
' the "VMWare" string for any DiskDrive Model value (parameters default to
' pass-by-ref).  
'
Function Fetch_DiskDrive(blnIsVM)

On Error Resume Next

    Dim strQuery
    Dim objSet, objWbemObj
    Dim intMyError

    intMyError = 0

    strQuery = "SELECT * FROM Win32_DiskDrive"
    Set objSet = objCIMV2Services.ExecQuery(strQuery)
        
    Dim strModel
    For Each objWbemObj In objSet
        if (Err <> 0) Then
            intMyError = Err.Number
	    Exit For  ' And that's all she wrote...
        End If
	' If we can fetch the DiskDrive class, check whether this is a VM
        ' A VMWare disk contains this value for at least one Model attr:
        ' Model = "VMware Virtual disk SCSI Disk Device";
	strModel = objWbemObj.Model
	if (InStr(strModel, "VMware")) Then
            blnIsVM = True
	    Exit For   ' we just need to match one
        End If 
    Next
    Fetch_DiskDrive = intMyError

End Function 'Fetch_DiskDrive


'
' TimeStamp() fetches the current date and time, and returns it to caller
' as a string value timestamp.
' It also writes that timestamp into the data file as a comment (i.e.,
' prefaced with "# ") IFF blnVerbose is True.
' It accepts a string that is written between the comment string and the
' date-time string.  This string allows descriptive Timestamp lines.
' E.g.: "# Start Win32_DiskDrive: 12/1/2009 1:49:08 PM"
'
Function TimeStamp(strStr)

On Error Resume Next

    Dim strNow
    strNow = now
    TimeStamp = strNow

    if (blnVerbose = True) Then
        objOutFile.WriteLine "# TS: " & strStr & ": " & strNow
        ' objOutFile.Flush
    End If

End Function 'Timestamp


'
' TimeStampDiff() computes the difference in seconds between strStart and
' strEnd, and writes that integer value into the output file IFF either
' blnVerbose or blnAlwaysPrint are True.  (We may want to display the
' duration of a command but not the absolute starting and ending times).
' It accepts a string that is written between the comment string (as above)
' and the duration value.
' E.g.: "# Win32_DiskDrive duration 1 second(s)".
'
Function TimeStampDiff(strStr, strStart, strEnd, blnAlwaysPrint)

On Error Resume Next

    Dim intDiff
    intDiff = Int(DateDiff("s", strStart, strEnd))
    TimeStampDiff = intDiff

    if (blnVerbose = True Or blnAlwaysPrint = True) Then
        if (blnVerbose = True) Then
            objOutFile.WriteLine "# TSDS: " & strStr & " S: " & strStart & _
                                 ", E: " & strEnd
        End If
        objOutFile.WriteLine "# TSD: " & strStr & intDiff & " second(s)"
        ' objOutFile.Flush
    End If

End Function 'TimestampDiff



const HKEY_LOCAL_MACHINE = &H80000002

'
' RegEnumKeys() returns in an array the enumerated keys of the specified 
' registry section.
' objService is object obtained when we connected to default WMI namespace
' strKeyPath is the path in the registry to the section we wish to enumerate.
'
Function RegEnumKeys(objService, strKeyPath)

    Dim oReg, arrSubKeys

    set oReg = objService.get("StdRegProv")
    oReg.EnumKey HKEY_LOCAL_MACHINE, strKeyPath, arrSubKeys
    RegEnumKeys = arrSubKeys

End Function 'RegEnumValues


'
' RegRead() reads and returns the value of the specified key.
' objService is object obtained when we connected to default WMI namespace
' strKeyPath is the path in the registry to the section we want.
' strKeyName is the name of the key whose value we wish to fetch.
'
Function RegRead(objService, strKeyPath, strKeyName)

    Dim oReg, strValue

    Set oReg = objService.get("StdRegProv")
    oReg.GetStringValue HKEY_LOCAL_MACHINE, strKeyPath, strKeyName, strValue
    ' StdOut.WriteLine "Name: " & strKeyName & ", value: " & strValue

    RegRead = strValue

End Function 'RegRead


'
' Is64BitOS() returns True if the target host is running a 64-bit OS, False
' otherwise.  
'
Function Is64BitOS()

    Dim strKeyPath, subKey, strKeyValue
    ' Look for key "Identifier" under HARDWARE\DESCRIPTION\System\CentralProcessor\0
    ' Once found the key, check the value to see if "x86" string exists.
    ' If string "x86" exists, then its a 32bit machine and all the required
    ' keys will be under "Software".
    ' If not the host is running a 64-bit OS, and all of the keys we expect will
    ' be under "Software\Wow6432Node" instead of "Software".
    strKeyPath = "HARDWARE\DESCRIPTION\System\CentralProcessor\0"
    subKey = "Identifier"

    strKeyValue = RegRead(objDefServices, strKeyPath, subkey)
    ' Wscript.Echo strKeyValue
    If (InStr(strKeyValue, "x86")) Then
       ' Wscript.Echo "32-bit OS"
    Else
       ' Wscript.Echo "64-bit OS"
       Is64BitOS = True
       Exit Function
    End If

    ' If we are here, the key doesn't exist, so this is a 32-bit host.
    ' Wscript.Echo "Host is 32-bit!"
    Is64BitOS = False
    Exit Function

End Function 'Is64BitOS


'
' Vbscript's split() function can only split on single instances of characters.
' This means that splitting ascii values on " " results in a bunch more
' entries than we want.  FixForSplit() iteratively substitutes one space char
' for every instance of two successive space chars until only single space
' chars exist in the string 'strStr'.
' FixForSplit() returns the modified string.
'
Function FixForSplit(strStr)

    DIM tmpStr

    tmpStr = strStr
    ' Wscript.Echo "Fix: instr: '" & tmpStr & "'"
    tmpStr = Trim(tmpStr)
    'iteratively replace every sequence of two spaces with 1.
    Do While InStr(1, tmpStr, "  ")
      tmpStr = Replace(tmpStr, "  ", " ")
    Loop

    ' Wscript.Echo "Fix: outstr: '" & tmpStr & "'"
    FixForSplit = tmpStr

End Function 'FixForSplit



