#########################################################################
# @(#) $Id: make_hotfix.pl,v 1.21 2011/05/19 05:26:13 marya Exp $
# $Source: /project/vasco/src/hal/cli/diag/upgrade/hotfix/make_hotfix.pl,v $
# Hotfix archive main.pl
# This file was created automatically - do not modify.
#########################################################################

use strict;

###########################################################################
# dependent modules and standard initialization
###########################################################################
BEGIN {
    unshift @INC, ".";
    use mkpath;
    my $jobs_dir = &construct_path("", "jobs");
    unshift @INC, $jobs_dir;
    use lib qw($jobs_dir);
};

use warnings;
use Switch;
use Archive::Tar;
use File::Find;
use IO::Zlib;
use File::Path;
use File::Basename;
use File::Copy;
use FileHandle;
use Cwd;
use Time::localtime;

use mkpath;
use ValidateArchive;

# The first thing that should always be done is to redirect output
# from the script to a file in the maintenance output directory.
# The script is called with the maintenance directory as the
# current directory.
if (!open STDOUT, '>>./output/stdout.txt') {
    print &halmsg("-t 999", "Unable to write to output/stdout.txt: $!");
    exit(1);
}
if (!open STDERR, '>>./output/stderr.txt') {
    print &halmsg("-t 999 -W", "Unable to write to output/stderr.txt: $!");
    exit(1);
}
if (!open STDIN, '/dev/null') {
    print &halmsg("-t 999", "Unable to redirect stdin: $!");
    # could still continue
}

my @hf_env_to_cleanup = qw(EAT_HOME_DIR EAT_CATALOG_DIR ETAZ_HOME_DIR ETAZ_CATALOG_DIR ETAZ_BIN_DIR ETAZ_LIB_DIR ETAZ_DTD_DIR PERL5LIB);

my $hf_cc = {};  # hotfix context closure (global state)
my $hf_sep = '/'; # global name separator
&load_desc_file($hf_cc);  # load description file data

# Determine the OS discriminator in the archive.
# May need to adjust this if we support multiple variations
# of the same OS.
if ( $is_windows ) {
    $hf_cc->{OS} = "Windows_NT";  # Name expected in the archive
    #$hf_sep = '\\';  '/' works on windows too
}
else {
    my($ostype) = JobManager::getOsType();
    my(@unamefield) = split /\s+/,$ostype;
    $hf_cc->{OS} = $unamefield[0];
}
$hf_cc->{MaintenanceDir} = getcwd;
$hf_cc->{ArchiveTop} = join($hf_sep,$hf_cc->{MaintenanceDir},"data",$hf_cc->{OS});
$hf_cc->{PID} = $$;

# Set ProductTop, ProductTopParent, VersionDir, VersionFile, vxccs
&set_runtime_locations($hf_cc);

# Set if 1 if SCM enabled
$hf_cc->{_SCMInstalled} = &check_scm_installed();

########################################################
# Hotfix specific data will be inserted below this line

my @hf_workflow = (
	"Validate Product Version",
	"Validate Fix Can Be Installed",
	"Check If Obsolete",
	"Check PreRequisites",
	"Enable Exit Logic",
	"Stop xprtld",
	"Save Old Files",
	"Install New Files",
	"Start Of Exit Logic",
	"Start xprtld",
	"Create Status File If Successful",
	"Show CC Status",
	"Cleanup",
);
my @hf_pre_upgrade_hooks = (
);

# End of hotfix specific data
########################################################

########################################################
# Main program starts
########################################################
&dump_ref($hf_cc, "Context");
$hf_cc->{_CleanupDone} = 0;  # don't display
$hf_cc->{_Failed} = 0; # don't display
$hf_cc->{_EnableExitLogic} = 0; # enable after version processing
$hf_cc->{_compute_destf} = \&hf_default_dest_file;

# if there is an extension, call its init sub if it exists
my $pkg = __PACKAGE__;
my $init_code = $pkg->can("hf_ext_init");
if ($init_code) {
    print("Calling extension library initialization.");
    $init_code->($hf_cc);
}

# ignore errors

# run the workflow
my ($wfi) = 1;
my $exit_logic_reached=0;
foreach my $wf (@hf_workflow) {
    my $wf_sub = lc($wf);  # lower case
    $wf_sub =~ s/\s+/_/g; # strip spaces

    # if there is an error, skip interveaning steps until we
    # reach the exit logic step
    $exit_logic_reached = 1 if $wf_sub eq "start_of_exit_logic";
    my $n = $wfi++;
    next if $hf_cc->{_Failed} && !$exit_logic_reached;

    print join('', "\n", &get_ts(), sprintf(" %2d. ", $n), $wf, "\n");
    eval "{\&wf_${wf_sub}(\$hf_cc);}";
    if ( $@ ) {
	$hf_cc->{_Failed} = 1;  # had a failure
	my $msg = join('', &get_ts(), " Error: ", $@);
	warn $msg; # send to stderr
	print $msg; # also send to stdout
	if ( $hf_cc->{_EnableExitLogic} == 0 ) {
	    last;
	}
	# else don't bail out, but try all exit logic steps,
	# but first call the pre-upgrade hooks in reverse order
	if ( scalar(@hf_pre_upgrade_hooks) ) {
	    my @h;
	    foreach my $wf (@hf_pre_upgrade_hooks) {
		 push @h, $wf;
	    }
	    foreach my $wf (@h) {
		 my $wf_sub = lc($wf);  # lower case
		 $wf_sub =~ s/\s+/_/g; # strip spaces
		 print join('', "\n", &get_ts(), "  ", $wf, " (undo)\n");
		 eval "{\&wf_${wf_sub}(\$hf_cc);}";
		 # ignore errors
	    }
	}
    }
}

# we may not have cleaned up if we aborted
&wf_cleanup($hf_cc) if !$hf_cc->{_CleanupDone};

if ( $hf_cc->{_Failed} ) {
    warn  join(' ', &get_ts(), "Upgrade failed\n");
    print join(' ', &get_ts(), "Upgrade failed\n");
}
else {
    print join(' ', &get_ts(), "Upgrade completed"); # reqd
}
print "\n";

exit(0);

# Helper subroutines
sub dump_ref {
    my($hr,$name,$depth) = @_;
    $depth ||= 0;
    my $tab = "  " x $depth;

    my $krt = ref ($hr);
    if ( $krt eq "ARRAY" ) {
	print join('', $tab, $name, " = [\n") if $name;
	foreach my $e ( @{$hr} ) {
	    &dump_ref($e, undef, $depth+1);
	}
	print join('', $tab, "]\n") if $name;
    }
    elsif ( $krt eq "HASH" ) {
	print join('', $tab, $name, " = {\n") if $name;
	foreach my $k (sort keys %{$hr}) {
	    &dump_ref($hr->{$k}, $k, $depth+1);
	}
	print join('', $tab, "}\n") if $name;
    }
    elsif ( $name ) {
	print join('',$tab, $name, " = ",  $hr, "\n");
    }
    else {
	print join('',$tab, $hr, "\n");
    }
}

# Load the values from the description file
sub load_desc_file {
    my $cc = shift or die "Context closure not provided\n";
    do {
	my($fh) = new FileHandle;
	die "Unable to open desc.txt" if !$fh->open("< desc.txt");
	while ( <$fh> ) {
	    m/\s*"(\w+)"\s*=\s*"(.*)"\s*$/o && do {
		$cc->{$1} = $2;
	    };
	}
	print join(' ', &get_ts(), "Loaded desc.txt\n");
    };
}

sub clean_env_run {
    my $cmd = shift;
    # unset significant environment variables
    my $zl = ""; # zap list
    foreach my $ev (@hf_env_to_cleanup) {
	next if !exists($ENV{$ev}); # do not create them if not present
	$zl = join('', $zl, 'local $ENV{',$ev,'}; ');
	$zl = join('', $zl, 'delete $ENV{',$ev,'};', "\n");
    }
    eval "{ $zl; $cmd(\@_); }";
}

sub clean_run_cmd {
    &clean_env_run("run_cmd", @_);
    die $@ if $@;
}

sub clean_run_cmd_w {
    &clean_env_run("run_cmd_w", @_);
}

#wrapper sub routine for &run_cmd(cmd) to double quote command
#and arguments before executing
sub run_cmd_with_args {
    my($cmd, @arg) = @_;
    my $argv = &double_quotes_arguments(@arg);
    my $cmd_with_arg = join(' ', "\"$cmd\"", $argv);
    print join('', &get_ts(), " # ",$cmd_with_arg, "\n");
    &run_cmd($cmd_with_arg);
}

#wrapper sub routine for &run_cmd_w(cmd) to double quote command
#and arguments before executing
sub run_cmd_warn_with_args {
    my($cmd, @arg) = @_;
    my $argv = &double_quotes_arguments(@arg);
    my $cmd_with_arg = join(' ', "\"$cmd\"", $argv);
    print join('', &get_ts(), " # ",$cmd_with_arg, "\n");
    &run_cmd_w($cmd_with_arg);
}

#wrapper sub routine for &clean_run_cmd(cmd) to double quote command
#and arguments before executing
sub clean_run_cmd_with_args {
    my($cmd, @arg) = @_;
    my $argv = &double_quotes_arguments(@arg);
    my $cmd_with_arg = join(' ', "\"$cmd\"", $argv);
    print join('', &get_ts(), " # ",$cmd_with_arg, "\n");
    &clean_run_cmd($cmd_with_arg);
}

#wrapper sub routine for &clean_run_cmd_w(cmd) to double quote command
#and arguments before executing
sub clean_run_cmd_warn_with_args {
    my($cmd, @arg) = @_;
    my $argv = &double_quotes_arguments(@arg);
    my $cmd_with_arg = join(' ', "\"$cmd\"", $argv);
    print join('', &get_ts(), " # ",$cmd_with_arg, "\n");
    &clean_run_cmd_w($cmd_with_arg);
}

#sub to double quote the arguments
#it takes input as array of arguments
sub double_quotes_arguments {
    my $argv = "";
    my @array = @_;
    my $arg;
    if (@array == 0) {
        print "array list is null..\n";
        return;
    }

    foreach $arg ( @array ) {
        $argv = join(" ",$argv, "\"$arg\"");
    }
    return $argv;
}

sub set_runtime_locations {
    my $cc = shift or die "Context closure not provided\n";

    my $relpt = &construct_path("", "bin/../..");
    my $here = getcwd;
    chdir $relpt or die "Unable to change dir to [$relpt]\n";
    my $pt = getcwd;
    chdir $here or die "Unable to change dir to [$here]\n";
    $cc->{ProductTop} = $pt;

    # now get the product top parent
    my $ptp = $pt;
    $ptp =~ s|\A(.*)/[^/]+\Z|$1|o;
    if ( $is_windows ) {
	if ( $ptp =~ m/\A\w:\Z/ ) {
	    $ptp .= "/"; # root of drive
	}
    }
    elsif ( $cc->{OS} =~ m/AIX/o ) {
	# AIX is a special case where there is no /opt/VRTSccs
	# directory
	# Default $ptp to /opt itself
	$ptp = "/opt";
    }
    else {
	if ( $ptp =~ m/\A\s*\Z/ ) {
	    $ptp = "/"; # root of file system
	}
    }
    $cc->{ProductTopParent} = $ptp;

    my $vfd;
    if ( $is_windows ) {
	# On Windows the file will be created in the product top directory
        if ( $cc->{PRODUCT_VERSION} eq "5.0MP1RP2" ) {
            # On Windows 5.0MP1RP2 don't search cc_version.txt in version directory 
            $vfd = join($hf_sep, $pt);
        }
        else {
            # On Windows other than 5.0MP1RP2 go into version directory
	    $vfd = join($hf_sep, $pt, "version");
        }
    }
    else {
	# On UNIX systems the file will be created in a fixed location
	$vfd = "/opt/VRTS/version";
    }
    $cc->{VersionDir} = $vfd;
    $cc->{VersionFile} = join($hf_sep, $vfd, "cc_version.txt");

    # setup the location of the vxccs script
    if ( $is_windows ) {
	# on windows this is a vbs script with a .bat front end
	my $win_vxccs = join($hf_sep, $pt, "Support/Tools/VxCCS/VxCCS.bat");
	$cc->{vxccs} = &short_path($win_vxccs);
    }
    else {
	$cc->{vxccs} = "/opt/VRTS/bin/vxccs"; # fixed location
    }
}

sub locate_webserver {
    my $cc = shift or die "Context closure not provided\n";
    return $cc->{_webserver} if exists($cc->{_webserver});
    if ( $is_windows ) {
	my $hive = "HKEY_LOCAL_MACHINE\\SOFTWARE\\VERITAS\\VRTSweb";
	my $hvp = &windows_regquery($hive);
	$cc->{_webserver} = $hvp->{"Path"};
    }
    else {
	$cc->{_webserver} = "/opt/VRTSweb";
    }
    return $cc->{_webserver};
}

sub locate_smweb {
    my $cc = shift or die "Context closure not provided\n";
    return $cc->{_smweb} if exists($cc->{_smweb});
    if ( $is_windows ) {
	my $hive = "HKEY_LOCAL_MACHINE\\SOFTWARE\\VERITAS\\VRTSsmweb\\CurrentVersion";
	my $hvp = &windows_regquery($hive);
	$cc->{_smweb} = $hvp->{"INSTALLDIR"};
    }
    else {
	$cc->{_smweb} = "/opt/VRTSsmweb";
    }
    return $cc->{_smweb};
}

sub locate_dbms {
    my $cc = shift or die "Context closure not provided\n";
    return $cc->{_dbms} if exists($cc->{_dbms});
    if ( $is_windows ) {
	my $hive = "HKEY_LOCAL_MACHINE\\SOFTWARE\\VERITAS\\VxDBMS3";
	my $hvp = &windows_regquery($hive);
	$cc->{_dbms} = $hvp->{"INSTALL_PATH"};
    }
    else {
	$cc->{_dbms} = "/opt/VRTSdbms3";
    }
    return $cc->{_dbms};
}

sub get_ts {
    return join('', '(', ctime(), ')');
}

# sub to compute the destination to copy a given file.
# Extensions can provide alternative locations for some
# files.
# The default is set up to copy to the CCS product tree.
sub hf_default_dest_file {
    my $cc = shift or die "Context closure not provided\n";
    my $f = shift;
    return join($hf_sep, $cc->{ProductTopParent}, $f);
}

# sub to load the version file
# It returns a hash with the contents
sub load_version_file {
    my $verfile = shift;

    die "$verfile not found\n" if ! -f $verfile;

    # parse the version file
    my $fh = new FileHandle();
    $fh->open($verfile, "r") or die "Unable to open $verfile: $!\n";

    my $ver;
    my $hfs = {};
    my $curhf;
    print "Parsing $verfile\n";
    print "==============================================\n";
    my $lno=1;
    while ( <$fh> ) {
	print join(' ', sprintf("[%03d]", $lno++), $_);
	chomp();
	next if m/\A\s*\Z/o;# empty lines
	next if m/\A#/o;    # comments
	m/version\s*=\s*(.*)\s*\Z/io && do {
	    if ( defined($ver) ) {
		warn "VERSION spectified multiple times in $verfile\n";
		next; # first only
	    }
	    $ver = $1;
	    next;
	};
	m/fixid\s*=\s*(.*)\s*\Z/io && do {
	    $curhf = $1;
	    if ( exists($hfs->{$curhf}) ) {
		warn "FIXID $curhf specified multiple times in $verfile\n";
		next; # first only
	    }
	    $hfs->{$curhf} = {};
	    next;
	};
	m/date\s*=\s*(.*)\s*\Z/io && do {
	    next if !defined($curhf);
	    $hfs->{$curhf}->{DATE} = $1;
	    next;
	};
	m/description\s*=\s*(.*)\s*\Z/io && do {
	    next if !defined($curhf);
	    $hfs->{$curhf}->{DESCRIPTION} = $1;
	    next;
	};
	m/requires\s*=\s*(.*)\s*\Z/io && do {
	    next if !defined($curhf);
	    $hfs->{$curhf}->{REQUIRES} = $1;
	    next;
	};
	m/obsoletes\s*=\s*(.*)\s*\Z/io && do {
	    next if !defined($curhf);
	    $hfs->{$curhf}->{OBSOLETES} = $1;
	    next;
	};
    }
    undef($fh);
    print "==============================================\n";

    die "Version not found in $verfile\n" if !defined($ver);
    print "Installed product version is $ver\n";

    $hfs->{VERSION} = $ver;

    return $hfs;
}

# Returns a hash with the status of configured explorers
#  {  
#     ExplorerName => {  
#         process_name => Name
#         process_state => state
#         explorer_type => type
#     }
#     ...
#   }
# Input Args:
#   (1|0) to print the output
sub get_hal_explorer_states {
    my $print_output = shift;

    my $es = {};
    my $halagentcfg = 
	&construct_exe_path("", join('/', $product_bindir, "halagentcfg"));
    my($cmd) = &double_quotes_arguments($halagentcfg,"list-processes", "-e", "-d");
    print join('', &get_ts(), " # ", $cmd, "\n") if $print_output;
    my @lines = qx/$cmd/;

    my $current_process;
    my $current_process_state;
    my $current_explorer;
    foreach my $l (@lines) {
	$_ = $l;
	chomp();
	next if m/\A\s*\Z/o;
	m/Process\s+\[(\w+)\]\s*$/o && do {
	    $current_process = $1;
	    $current_process_state = 'STOPPED';
	    print $l if $print_output;
	    next;
	};
	m/Process\s+\[(\w+)\]\s+(\w+)\*?\s+.*$/o && do {
	    $current_process = $1;
	    $current_process_state = $2;
	    print $l if $print_output;
	    next;
	};
	m/\s*Explorer\s+\[(\w+)\]\s*$/o && do {
	    $current_explorer = $1;
	    $es->{$current_explorer} = { 
		process_name => $current_process,
		process_state => $current_process_state
	    };
	    print $l if $print_output;
	    next;
	};
	m/\s*Explorer\s+\[(\w+)\]\s+(\w+)\*?\s+.*$/o && do {
	    $current_explorer = $1;
	    $es->{$current_explorer} = { 
		process_name => $current_process,
		process_state => $current_process_state
	    };
	    print $l if $print_output;
	    next;
	};
	m/\s*ExplorerType\s*=\s*(\w+)\s*$/o && do {
	    $es->{$current_explorer}->{explorer_type} = $1;
	    next;
	};
	m/\s*\w+\s*=/o && do { next; }; # expected
	die "UNEXPECTED PATTERN: $_";
    }

    #&dump_ref($es) if $print_output;

    return $es;
}

# Input Args: 
#   (1|0) to print the output
#   1 to also include the LHM, 0 otherwise
# Returns 1 if any HAL explorer is running (depends on lhm flag)
# Returns 0 otherwise.
sub check_any_hal_explorer_running {
    my $print_output = shift;
    my $lhm_too = shift || 0;
    my $es = &get_hal_explorer_states($print_output);
    foreach my $e (keys %{$es}) {
	next if !$lhm_too && $e eq "HALLHM";  # don't count
	return 1 if $es->{$e}->{process_state} =~ m/RUNNING|INITIALIZING|HUNG/o;
    }
    return 0;
}

sub check_is_ms_host {
    my $cc = shift or die "Context closure not provided\n";
    if ( exists($cc->{_IsMSHost}) ) {
	return $cc->{_IsMSHost};
    }
    # look for the presence of the dbstage program => CHM installed => MS
    # Note: it is never wrapped, so it will be in the bin directory on UNIX
    my $dbstage = &construct_exe_path("",join('/',$product_bindir, "dbstage"));
    $cc->{_IsMSHost} = 0;
    if ( -x $dbstage ) {
	$cc->{_IsMSHost} = 1;
    }
    return $cc->{_IsMSHost};
}

# Return non-zero if DM is installed, 0 otherwise
sub check_dm_installed {
    print "TBD: check_dm_installed needs to be validated\n";
    if ( $is_windows ) {
	# check if the explorer is present
	my $chk = join($hf_sep,  $hf_cc->{ProductTop}, "Data Module",
		       "bin", "dmexplorer.exe");
	return -f $chk;
    }
    else {
	# check if the explorer is present
	my $chk = join($hf_sep,  $hf_cc->{ProductTop}, "VRTSccdam",
		       "bin", "dmexplorer");
	return -f $chk;
    }
}

# Return 1 if SICL is installed, 0 otherwise
sub check_sicl_installed {
    print "TBD: check_sicl_installed needs to be validated\n";
    if ( $is_windows ) {
	# check if the explorer is present
	my $chk = join($hf_sep,  $hf_cc->{ProductTop}, "SICL",
		       "bin", "vxsicld.exe");
	return -f $chk;
    }
    else {
	# check if the explorer is present
	my $chk = join($hf_sep,  $hf_cc->{ProductTop}, "VRTSsicls",
		       "bin", "vxsicld");
	return -f $chk;
    }
}

# Return 1 if SCM is installed, 0 otherwise
sub check_scm_installed {
    if ( $is_windows ) {
        # check if the explorer is present
        my $chk = join($hf_sep,  $hf_cc->{ProductTopParent}, ".ccs.enabled");
        return ! -f $chk;
    }
    else {
        # check if the explorer is present
        return ! -f "/opt/VRTS/install/.ccs.enabled";
    }
}

# restore copied files
sub restore_files {
    my $cc = shift or die "Context closure not provided\n";

    my $fcu = $cc->{_FileCopyUndo};
    if ( defined($fcu) ) {
	eval {
	    foreach my $f (keys %{$fcu}) {
		print "Restoring $f\n";
		my $orig = $fcu->{$f};
		do {
		    print "$f not renamed\n"; next;
		} if ! -f $orig;
		unlink $f if -f $f; # remove the newer version
		File::Copy::move($orig, $f) or 
		    warn "Unable to restore $f: $!\n";
	    }
	};
	if ( $@ ) {
	    warn $@;
	}
    }
    $cc->{_FileCopyUndo} = undef;  # only do this once
}

########################################################
# Workflow subroutines
########################################################
# These are constructed from the workflow state names as 
# follows:
#       - subroutine names are in lower case
#       - all contiguous spaces collapsed into a single _
#       - A leading wf_ prefix applied
# For example:
#     Validate Product Version
# becomes
#      wf_validate_product_version
#
# These are all void subroutines. They must "die" on error.
# They are all invoked with a reference to a hash defining
# the "state".
########################################################

# Die if this is the wrong version of the product
sub wf_validate_product_version {
    my $cc = shift or die "Context closure not provided\n";

    my $want_ver = $cc->{PRODUCT_VERSION};
    my $ver;
    my $hfs;

    if ( $want_ver =~ m/\A(5\.0|5\.0RP1|5\.0MP1|5\.0MP1RP1)\Z/o ) {
	die "Do not yet support releases earlier than 5.0MP1RP2\n";

	# determine the product version somehow
	$ver = "VERSION";
	$hfs = {};
	$hfs->{VERSION} = $ver;
	# no information on hot fixes available
    }
    else {
	# The version file is present from 5.0MP1RP2 onward
	my $verfile = $cc->{VersionFile};
	$hfs = &load_version_file($verfile);
	$ver = $hfs->{VERSION};
    }
    die "Invalid product version - expected $want_ver\n" 
	if $ver ne $want_ver;

    # remember the contents for later
    $cc->{_HotFixStatus} = $hfs;
    return;
}

# validate a hot fix
sub wf_validate_fix_can_be_installed {
    my $cc = shift or die "Context closure not provided\n";

    my $hfs = $cc->{_HotFixStatus};

    # check if fix already installed
    foreach my $hf (keys %{$hfs}) {
	next if $hf eq "VERSION";
	next if $hf ne $cc->{FIXID};
	my $hfi = $hfs->{$hf};
	my $dt = $hfi->{DATE};
	die join("","This fix has already been installed",
		 (defined($dt) ? "\n" : "on '$dt'\n"));
    }
    print "This fix has not yet been installed\n";
}

sub wf_check_if_obsolete {
    my $cc = shift or die "Context closure not provided\n";

    my $hfs = $cc->{_HotFixStatus};

    # Has this fix been obsoleted?
    foreach my $hf (keys %{$hfs}) {
	next if $hf eq "VERSION";
	my $hfi = $hfs->{$hf};
	my $obs = $hfi->{OBSOLETES};
	next if !defined($obs);

	my @obl = split(/[\s,]+/, $obs);
	foreach my $ohf (@obl ) {
	    next if $ohf ne $cc->{FIXID};
	    die "This fix has been made obsolete by fix $hf\n";
	}
    }
    print "This fix is not obsolete\n";
}

sub wf_check_prerequisites {
    my $cc = shift or die "Context closure not provided\n";

    my $hfs = $cc->{_HotFixStatus};

    # Are pre-requisite fixes installed?
    my $prl = $cc->{RequiresFixes};
    my $errs=0;
    foreach my $pr ( @{$prl} ) {
	if ( !defined( $hfs->{$pr} ) ) {
	    print "Pre-requisite fix '$pr' is not installed\n";
	    $errs++;
	}
	else {
	    print "Pre-requisite fix '$pr' has been installed\n";
	}
    }
    die "Pre-requisite fixes are missing\n" if $errs;
    print "Pre-requisites constraints are satisified\n";
}

sub wf_enable_exit_logic {
    my $cc = shift or die "Context closure not provided\n";
    $cc->{_EnableExitLogic} = 1;
}

# Die if not the MS host
sub wf_validate_ms_host {
    my $cc = shift or die "Context closure not provided\n";
    die "This hotfix applies only to the Management Server host\n"
	if !check_is_ms_host($cc);
    print "This is the management server host\n";
    return;
}

# Die if the MS host
sub wf_validate_is_not_ms_host {
    my $cc = shift or die "Context closure not provided\n";
    die "This hotfix does not apply to the Management Server host\n"
	if check_is_ms_host($cc);
    print "This is not a management server host\n";
    return;
}

# start the LHM
# do not fail
sub wf_start_lhm() {
    my $cc = shift or die "Context closure not provided\n";

    my($lhmcmd) = &construct_exe_path("", join('/',$product_bindir,"lhm"));
    &run_cmd_warn_with_args($lhmcmd, "--start-service");
}

# stop the LHM
sub wf_stop_lhm() {
    my $cc = shift or die "Context closure not provided\n";

    # first stop all HAL explorers (safer)
    &wf_stop_all_hal_explorers($cc);

    my $count=0;
    while ( &check_any_hal_explorer_running(1, 1) ) { # print always
	if ( $count == 0 ) {
	    my($lhmcmd) = &construct_exe_path("", 
					      join('/',$product_bindir,"lhm"));
	    &run_cmd_with_args($lhmcmd,"--stop-service");
	}
	sleep(1);
	$count++;
	die "Unable to stop all HAL explorers and the LHM\n" if $count > 120;
    }
    print "All HAL explorers and the LHM have stopped\n";
}

# start all HAL explorers (LHM is running)
# do not fail
sub wf_start_all_hal_explorers {
    my $cc = shift or die "Context closure not provided\n";
    my($execmd) = &construct_exe_path("", join('/',$product_bindir,"halagentcfg"));
    &run_cmd_warn_with_args($execmd, "start-all");
    print "HAL explorers were re-started\n";
}

# stop all HAL explorers (but not the LHM)
sub wf_stop_all_hal_explorers {
    my $cc = shift or die "Context closure not provided\n";
    my $count=0;
    while ( &check_any_hal_explorer_running(($count ? 0 : 1), 0) ) {
	if ( $count == 0 ) {
	    my($execmd) = 
		&construct_exe_path("", join('/',$product_bindir,"halagentcfg"));
	    &run_cmd_with_args($execmd, "stop-all");
	}
	sleep(1);
	$count++;
	die "Unable to stop all HAL explorers\n" if $count > 120;
    }
}

# stop specific HAL explorers described in @hf_hal_explorer_types
sub wf_stop_hal_explorers_by_type {
    my $cc = shift or die "Context closure not provided\n";

    # get current explorer configuration (print config)
    my $es = &get_hal_explorer_states(1);

    # get the reference to the hash of explorer types to stop
    my $et = $cc->{StopHALExplorersByType};

    # find the processes to stop. Don't care if not running
    my $pts = {};
    foreach my $en (keys %{$es}) {
	my $e = $es->{$en};
	my $expt = $e->{explorer_type};
	next if !defined($expt);
	next if !exists($et->{$e->{explorer_type}});
	$pts->{$e->{process_name}} = 1;
	#printf("Found explorer type %s in process %s (%s)\n",
	#       $e->{explorer_type},$e->{process_name}, $en); 
    }
    my $np = scalar(keys %{$pts});
    if ( $np == 0 ) {
	print "No explorers of the specified type are configured.\n";
	return;
    }
    print "Number of processes to stop: $np\n";
    foreach my $p (sort keys %{$pts}) {
	print "\t$p\n";
    }

    foreach my $p (sort keys %{$pts}) {
	my($execmd) = 
	    &construct_exe_path("", join('/',$product_bindir,"halagentcfg"));
	&run_cmd_with_args($execmd, "stop-process", "-p", $p);
    }

    my $count = 0;
    while ( $count < 120 ) {
	sleep 5 if $count == 0;
	sleep 2 if $count;
	$es = &get_hal_explorer_states(1);
	my $running = 0;
	foreach my $en (keys %{$es}) {
	    my $e = $es->{$en};
	    my $expt = $e->{explorer_type};
	    next if !defined($expt);
	    next if !exists($et->{$e->{explorer_type}});
	    $running = 1 if $e->{process_state} =~ m/RUNNING|INITIALIZING|HUNG/o;
	    last if $running;
	}
	last if $running == 0;
	$count++;
    }
    die "Unable to stop all processes\n" if $count == 120;

    # remember the stopped processes
    $cc->{_StoppedHALProcesses} = $pts;
}

# restart the specific HAL explorers that were stopped
# do not fail
sub wf_start_stopped_hal_explorers {
    my $cc = shift or die "Context closure not provided\n";
    
    my $pts = $cc->{_StoppedHALProcesses};

    foreach my $p (sort keys %{$pts}) {
	my($execmd) = 
	    &construct_exe_path("", join('/',$product_bindir,"halagentcfg"));
	&run_cmd_warn_with_args($execmd, "start-process", "-p", $p);
    }
}

# recreate the HAL wrappers
sub wf_recreate_hal_wrappers {
    my $cc = shift or die "Context closure not provided\n";

    my $hal_perl = &construct_exe_path("",join('/',$product_bindir,"perl"));
    my $wrapper_script = &construct_path("", "inst/WrapCommands.pl");
    # For Wrapping executables in VRTShal bin as management server.
    # pass option --chm to WrapperCommands.pl else it will be 
    # considered as Managed Host.
    my $chm = "";
    if ( check_is_ms_host($cc) ) {
        $chm = "--chm";
    }
    my $cmd = &double_quotes_arguments($hal_perl, $wrapper_script, $chm);
    print join('', &get_ts(), " # ", $cmd, "\n");
    my $output = `$cmd 2>&1`;
}

# show CC status (OS specific)
# Don't die
sub wf_show_cc_status {
    my $cc = shift or die "Context closure not provided\n";

    eval {
	&clean_run_cmd_with_args($cc->{vxccs}, "status");

	# show HAL status
	&get_hal_explorer_states(1);
    };
    if ( $@ ) {
	print "Error: $@\n";
	print "Error ignored.\n";
    }
}

# start DM
# do not fail
sub wf_start_dm {
    my $cc = shift or die "Context closure not provided\n";
    $cc->{_DMInstalled} = &check_dm_installed()	if ( !exists($cc->{_DMInstalled}) );
    return if !$cc->{_DMInstalled};
    if ( $is_windows ) {
	my $cmd = join(' ', "net", "start", "dmexplorer");
	print join('', &get_ts(), " # ", $cmd, "\n");
	my $output = `$cmd 2>&1`;
	print $output;
    }
    else {
	&run_cmd_warn_with_args($cc->{vxccs}, "start", "dmexplorer");
    }
}

# stop DM
sub wf_stop_dm {
    my $cc = shift or die "Context closure not provided\n";
    $cc->{_DMInstalled} = &check_dm_installed()	if ( !exists($cc->{_DMInstalled}) );
    return if !$cc->{_DMInstalled};
    if ( $is_windows ) {
	my $cmd = join(' ', "net", "stop", "dmexplorer");
	print join('', &get_ts(), " # ", $cmd, "\n");
	my $output = `$cmd 2>&1`;
	print $output;
    }
    else {
	&run_cmd_with_args($cc->{vxccs}, "stop", "dmexplorer");
    }
}

# start Monitoring
# do not fail
sub wf_start_monitoring {
    my $cc = shift or die "Context closure not provided\n";
    $cc->{_SICLInstalled} = &check_sicl_installed() 
	if ( !exists($cc->{_SICLInstalled}) );
    return if !$cc->{_SICLInstalled};
    if ( $is_windows ) {

	    if ( check_is_ms_host($cc) ) {
     	    my $cmd = join(' ', "net", "start", "vxamccs");
    	    print join('', &get_ts(), " # ", $cmd, "\n");
	        my $output = `$cmd 2>&1`;
	        print $output;

	        $cmd = join(' ', "net", "start", "vxasd");
	        print join('', &get_ts(), " # ", $cmd, "\n");
	        $output = `$cmd 2>&1`;
	        print $output;
	    }
   	    my $cmd = join(' ', "net", "start", "vxsicld");
	    print join('', &get_ts(), " # ", $cmd, "\n");
	    my $output = `$cmd 2>&1`;
	    print $output;
    }
    else {
	    if ( check_is_ms_host($cc) ) {
            &run_cmd_with_args($cc->{vxccs}, "start", "vxamd");
            &run_cmd_with_args($cc->{vxccs}, "start", "vxasd");
	    }
    	&run_cmd_warn_with_args($cc->{vxccs}, "start", "vxsicld");
    }
}

# stop Monitoring
sub wf_stop_monitoring {
    my $cc = shift or die "Context closure not provided\n";
    $cc->{_SICLInstalled} = &check_sicl_installed() 
	if ( !exists($cc->{_SICLInstalled}) );
    return if !$cc->{_SICLInstalled};
    if ( $is_windows ) {
        my $cmd = join(' ', "net", "stop", "vxsicld");
	    print join('', &get_ts(), " # ", $cmd, "\n");
	    my $output = `$cmd 2>&1`;
	    print $output;
	    if ( check_is_ms_host($cc) ) {
    	    my $cmd = join(' ', "net", "stop", "vxasd");
    	    print join('', &get_ts(), " # ", $cmd, "\n");
    	    my $output = `$cmd 2>&1`;
    	    print $output;
        
    	    $cmd = join(' ', "net", "stop", "vxamccs");
    	    print join('', &get_ts(), " # ", $cmd, "\n");
    	    $output = `$cmd 2>&1`;
    	    print $output;
	    }
    }
    else {
        &run_cmd_with_args($cc->{vxccs}, "stop", "vxsicld");
        if ( check_is_ms_host($cc) ) {
            &run_cmd_with_args($cc->{vxccs}, "stop", "vxasd");
            &run_cmd_with_args($cc->{vxccs}, "stop", "vxamd");
        }
    }
}

# save files that would get over written
sub wf_save_old_files {
    my $cc = shift or die "Context closure not provided\n";

    my $osbase = $cc->{ArchiveTop};
    my $filelist = [];
    my $find_wanted = sub {
	return if -d $File::Find::name; # skip dirs
	$_ = $File::Find::name;
	s|\A\./(.*)\Z|$1|; # strip the leading ./
	push @{$filelist}, $_;
    };

    # add the files in the OS specific branch of the data tree
    # chdir to get relative paths
    my $here = getcwd;
    chdir $osbase or die "Unable to change dir to [$osbase]\n or hotfix does not exists for this OS\n";
    find({wanted=>$find_wanted, no_chdir=>1}, ".");
    chdir $here or die "Unable to change dir to [$here]\n";

    print "Archive contents for this platform:\n  ";
    print join("\n  ", @{$filelist});
    print "\n\n";

    # get the product top parent
    my $ptp = $cc->{ProductTopParent};
    
    # Rename each file in-place with the hotfix id.
    # Rename may help on Windows if a file is in use.
    # Keep a hash of undo information.
    my $hfid = $cc->{FIXID};
    my $f_undo = {};
    foreach my $f (@{$filelist}) {
	my $srcf = $cc->{_compute_destf}($cc, $f);
	my $destf = $cc->{_compute_destf}($cc, join('.',$f,$hfid,"hfbak"));
	next if ! -f $srcf;
	print join('', "Renaming ", $srcf, "\n  => ", $destf, "\n");
	if ( -f $destf ) {
	    # is this a re-run of the same hotfix?
	    print join(' ', "  * File already present - not overwritten!\n");
	    # Will be added to the undo so we can restore it upon failure.
	    # This will help in recovery if rerun after a failure.
	}
	else {
	    # Big assumption that this rename operation is atomic
	    # otherwise the 'next' above skips possible cleanup of
	    # previous failure during exit of move.
	    File::Copy::move($srcf, $destf) or
		die "Unable to rename $srcf: $!\n";
	}
	$f_undo->{$srcf} = $destf;
    }
    $cc->{_FilesToCopy} = $filelist;
    $cc->{_FileCopyUndo} = $f_undo;
}

# copy out files from the hotfix archive
sub wf_install_new_files {
    my $cc = shift or die "Context closure not provided\n";

    my $osbase = $cc->{ArchiveTop};
    my $filelist = $cc->{_FilesToCopy};
    my $ptp = $cc->{ProductTopParent};
    foreach my $f (@{$filelist}) {
	my $srcf = join($hf_sep, $osbase, $f);
	my $destf = $cc->{_compute_destf}($cc, $f);
	print "Updating $destf\n";
	# allows reinstallation during testing
	unlink($destf) if -f $destf;
	my $dirname = dirname($destf);
	File::Path::mkpath($dirname) if (!-d $dirname);
	File::Copy::syscopy($srcf, $destf) or
	    die "Unable to copy to $destf: $!\n";
	# this variable could be put in a configurable context
	my $preserve_perms = 1;
	if ($preserve_perms) {
	    # make sure permissions are copied as well as data
	    # stat the file
	    my $mode = [stat($srcf)]->[2];
	    # mask off file type
	    $mode &= 07777;
	    # set permissions of target file
	    chmod($mode, $destf);
	}
    }
}

# Create a file indicating that this hotfix is installed
sub wf_create_status_file_if_successful {
    my $cc = shift or die "Context closure not provided\n";

    if ( $cc->{_Failed} ) {
	print "Hotfix failed - version file is not updated\n";
	return;
    }
    #print " replace version = $cc->{ReplaceVersion} \n" ;
    # update the version file
    my $fh;
    if( defined($cc->{ReplaceVersion}))
    {
        $fh = new FileHandle(); 
	my $verfile = $cc->{VersionFile};
        my $verfile_bkp = "$verfile".".bak";
        print "version backup file = $verfile_bkp\n";
        File::Copy::move($cc->{VersionFile},$verfile_bkp) or
                    warn "Unable to copy $verfile_bkp $!\n";
        $fh->open(">$verfile") or die "Unable to open $verfile for writing: $!\n";
        my $fhb = new FileHandle();
        $fhb->open("<$verfile_bkp") or die "Unable to open $verfile_bkp: $!\n";

       print $fh "VERSION=$cc->{ReplaceVersion}\n";
       while(<$fhb>)
       {
           my $line = $_;

           if ($line =~ /^VERSION=*/)
           {
               next;  
           }
           else
           {
              print $fh $line;
           }
       }
       undef $fhb;
       undef $fh;
   }
   $fh = new FileHandle();
   $fh->open(">>$cc->{VersionFile}") or die "Unable to open $cc->{VersionFile} verfile for writing: $!\n";

    print $fh "\n";
    print $fh join('', "FIXID=", $cc->{FIXID}, "\n");
    print $fh join('', "DATE=", ctime(), "\n");
    print $fh join('', "DESCRIPTION=", $cc->{DESCRIPTION}, "\n");

    my $rqf = $cc->{RequiresFixes};
    if ( defined($rqf) ) {
	print $fh join('', "REQUIRES=");
	print $fh join(',', @{$rqf});
	print $fh "\n";
    }

    my $obf = $cc->{ObsoletesFixes};
    if ( defined($obf) ) {
	print $fh join('', "OBSOLETES=");
	print $fh join(',', @{$obf});
	print $fh "\n";
    }
    undef $fh;
    print "Version file updated\n";

    $cc->{_FileCopyUndo} = undef; # commit
}

# Commit a custom archive if successful
sub wf_commit_custom_archive_if_successful {
    my $cc = shift or die "Context closure not provided\n";

    if ( $cc->{_Failed} ) {
	print "Archive installation failed\n";
	return;
    }

    print "Archive installation committed\n";
    $cc->{_FileCopyUndo} = undef; # commit
}

# Cleanup intermediate files and the extracted archive - no errors.
sub wf_cleanup {
    my $cc = shift or die "Context closure not provided\n";
    return if $cc->{_CleanupDone};

    eval {
	&restore_files($cc);
	my($datadir) = join('/', $cc->{MaintenanceDir}, "data");
	print join(' ', &get_ts(), "Removing the", $datadir, "directory\n");
	File::Path::rmtree($datadir);
	unlink "main.pl";
	# Do not unlink desc.txt as that is needed for ArchiveJob --list-patches
	###unlink "desc.txt";
    };
    if ( $@ ) {
	print join("\n", 
		   "There were errors during cleanup which were ignored:",
		   $@, "");
    }

    $cc->{_CleanupDone} = 1;
    return;
}

sub wf_start_of_exit_logic {
    my $cc = shift or die "Context closure not provided\n";

     if ( $cc->{_Failed} ) {
	 # restore files first so servers can restart
	 &restore_files($cc);
	 print "Attempting to restart servers if necessary\n";
     }
}

sub wf_stop_web_server {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    my $webdir = &locate_webserver($cc);
    my $cmd;
    if ( $is_windows ) {
	# We need to first stop dependent services (recursively).
	my $svc_list_ref = [];
	$svc_list_ref = &compute_svc_stop_order("VRTSweb");
	print join(' ', "Stop order:", @{$svc_list_ref},"\n");
	foreach my $svc (@{$svc_list_ref}) {
	    if ( exists($cc->{_vrtsweb_skip_svcs}->{$svc}) ) {
		print "Service $svc already stopped\n";
		next;
	    }
	    $cmd = join(' ', "net", "stop", $svc);
	    print join('', &get_ts(), " # ", $cmd, "\n");
	    &run_cmd_w(join(' ',$cmd,'2>&1'));
	}
	$cc->{_vrtsweb_svcs} = $svc_list_ref;
    }
    else {
	# We need to remember if there are any other web-apps
	# registered.  If so, we must use the "restart" option
	# when starting the web server later.
	# If we blindly call restart, the launcher will block
	# forever.
	#  - Carl 12/9/08
	$cmd = join('', $webdir, "/bin/webgui listapps");
	print join('', &get_ts(), " # ", $cmd, "\n");
	my @lines = qx/$cmd/;
	print join('', @lines);
	my $needs_restart = 0;
	foreach my $l (@lines) {
	    $_ = $l;
	    chomp();
	    m/server is offline/o && do { $needs_restart = 0; last; };
	    m/ROOT/o && do { next; }; # ignore this one
	    $needs_restart = 1;
	    last;
	}
	$cc->{_vrtsweb_needs_restart} = $needs_restart;

	#$cmd = join(' ', $cc->{vxccs}, "stop", "web");
	# Per 1501987 cannot do that, since the newly capable vxccs
	# does not support stop web (force or otherwise), and even
	# if it did, it is only part of the MEGA hotfix and not
	# already available at this point in the workflow.
	$cmd = join('', $webdir, "/bin/webgui");
	&clean_run_cmd_warn_with_args($cmd, "stop","force");
    }
}

sub wf_start_web_server {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    my $webdir = &locate_webserver($cc);
    my $cmd;
    if ( $is_windows ) {
	&clean_run_cmd_with_args($cc->{vxccs}, "start", "VRTSweb");
        return if !$cc->{_SCMInstalled};
	&clean_run_cmd_with_args($cc->{vxccs}, "start", "VRTSscmui");
    }
    else {
	# use "start" or "restart" as remembered prior to shutdown
	#$cmd = join(' ', $cc->{vxccs},
	#    ($cc->{_vrtsweb_needs_restart} ? "restart" : "start"), "web");
	#print join('', &get_ts(), " # ", $cmd, "\n");
	#&clean_run_cmd_w($cmd);
	&clean_run_cmd_warn_with_args($cc->{vxccs}, ($cc->{_vrtsweb_needs_restart} ? "restart" : "start"), "web");
    }
}

sub wf_stop_scm_server {
	my $cc = shift or die "Context closure not provided\n";
	return if !$cc->{_SCMInstalled};
	my $op = "stop";
	my $svc_name =
		(
		 $is_windows
		 ? "VRTSscms"
		 : "scms"
		);
	&clean_run_cmd_with_args($cc->{vxccs}, $op, $svc_name);
}

sub wf_start_scm_server {
	my $cc = shift or die "Context closure not provided\n";
        return if !$cc->{_SCMInstalled};
	my $op = "start";
	my $svc_name =
		(
		 $is_windows
		 ? "VRTSscms"
		 : "scms"
		);
	&clean_run_cmd_with_args($cc->{vxccs}, $op, $svc_name);
}

####################################
sub wf_restart_spc_service {
	
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }

    if ( $is_windows ) {
	
        my $reg_key_vrtsweb = "HKEY_LOCAL_MACHINE\\SOFTWARE\\VERITAS\\VRTSweb";
	my $hvp_vrtsweb = &windows_regquery($reg_key_vrtsweb);
	my $path = $hvp_vrtsweb->{"Path"};
	my $vrtsweb_path = "$path\\bin";
	
	my $reg_key_cc = "HKEY_LOCAL_MACHINE\\SOFTWARE\\VERITAS\\CommandCentral Storage";
	my $hvp_cc = &windows_regquery($reg_key_cc);
	my $install_dir_path = $hvp_cc->{"InstallDir"};
	my $web_engine_path = "$install_dir_path\\Web Engine";
        system(" echo Y | \"$vrtsweb_path\\stopApp.exe\" spc \"$web_engine_path\" ");
        system(" \"$vrtsweb_path\\startApp.exe\" spc \"$web_engine_path\" ");

    }
    else {
        my $install_dir_path = `cat /etc/VRTSccs/ccs.conf | grep CCS_INSTALL_DIR | cut -f 2 -d=`;
        chomp $install_dir_path;
        my $web_engine_path = "$install_dir_path/VRTSweb/VERITAS/";
        system("/opt/VRTSweb/bin/stopApp spc");
        system("/opt/VRTSweb/bin/startApp spc $web_engine_path");
    }
}


sub wf_stop_cc_gui {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    print "TBD: stop_cc_gui not implemented\n";
    if ( $is_windows ) {
	# ignore these services when stopping/starting vrtsweb
	$cc->{_vrtsweb_skip_svcs}->{"VRTScclogin"} = 1;
	$cc->{_vrtsweb_skip_svcs}->{"VRTSscmui"} = 1;
	$cc->{_vrtsweb_skip_svcs}->{"VRTSccsweb"} = 1;
    }
}

sub wf_start_cc_gui {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    print "TBD: start_cc_gui not implemented\n";
}

sub wf_stop_database_server {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    print "TBD: stop_database_server not implemented\n";
}

sub wf_start_database_server {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    print "TBD: start_database_server not implemented\n";
}

sub wf_stop_cc_database {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    print "TBD: stop_cc_database not implemented\n";
}

sub wf_start_cc_database {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    print "TBD: start_cc_database not implemented\n";
}

sub wf_stop_dm_database {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    print "TBD: stop_dm_database not implemented\n";
}

sub wf_start_dm_database {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    print "TBD: start_dm_database not implemented\n";
}

# start everything
sub wf_start_ccs {
    my $cc = shift or die "Context closure not provided\n";
    # Use the product control script
    &clean_run_cmd_warn_with_args($cc->{vxccs}, "start");
}

# stop everything
sub wf_stop_ccs {
    my $cc = shift or die "Context closure not provided\n";
    # Use the product control script
    &clean_run_cmd_with_args($cc->{vxccs}, "stop");

    if ( $is_windows && check_is_ms_host($cc) ) {
	# ignore these services when stopping/starting vrtsweb
	$cc->{_vrtsweb_skip_svcs}->{"VRTScclogin"} = 1;
	$cc->{_vrtsweb_skip_svcs}->{"VRTSscmui"} = 1;
	$cc->{_vrtsweb_skip_svcs}->{"VRTSccsweb"} = 1;
    }
}

sub wf_start_smweb {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    my $smwebdir = &locate_smweb($cc);
    my $cmd;
    if ( $is_windows ) {
	$cmd = join(' ', "net", "start", "VRTSsmweb");
	print join('', &get_ts(), " # ", $cmd, "\n");
	my $rc = &run_cmd_w(join(' ',$cmd,'2>&1'));

	# It has been observed that this command fails if
	# called immediately after VRTSweb is started.
	# However, a subsequent call shows that the service
	# is started.
	#  - Carl 12/11/08
	if ( $rc != 0 ) {
	    print join('', &get_ts(), " # ", $cmd, " (Retrying)\n");
	    &run_cmd_w(join(' ',$cmd,'2>&1'));
	}
    }
    else {
	$cmd = join('', $smwebdir, "/webgui/veaw");
	&clean_run_cmd_warn_with_args($cmd, "start");
    }
}

sub wf_stop_smweb {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    my $smwebdir = &locate_smweb($cc);
    my $cmd;
    if ( $is_windows ) {
	$cmd = join(' ', "net", "stop", "VRTSsmweb");
	# ignore this service when stopping/starting vrtsweb
	$cc->{_vrtsweb_skip_svcs}->{"VRTSsmweb"} = 1;
	print join('', &get_ts(), " # ", $cmd, "\n");
	&run_cmd_w(join(' ',$cmd,'2>&1'));
    }
    else {
	$cmd = join('', $smwebdir, "/webgui/veaw");
	&clean_run_cmd_warn_with_args($cmd, "stop");
    }
}

sub wf_start_esmweb {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    
    if ( $is_windows ) {
        system("sc start CCSGUI");
	my $count = 0;
        while (1) {
	    my $output = `"sc query CCSGUI | findstr \"STATE\""`;
            if ($output !~ /RUNNING/) {
		sleep(10);    
                $count++;
            }
	    if($count == 6 || $output =~ /RUNNING/) {
	        last;
	    }
        }
    }
    else {
	system("/opt/VRTS/bin/vxccs start esmweb");

    }
}

sub wf_stop_esmweb {
    my $cc = shift or die "Context closure not provided\n";
    if ( !check_is_ms_host($cc) ) {
	print "N/A\n";
	return;
    }
    if ( $is_windows ) {
	system("sc stop CCSGUI");
    }
    else {
	system("/opt/VRTS/bin/vxccs stop esmweb");
    }
}

# wf_start_xprtld will start xprtld
# This module is applicable only for CCS MS
# and CH.
sub wf_start_xprtld {
    my $cc = shift or die "Context closure not provided\n";
    if ( $is_windows ) {
        my $cmd = join(' ', "net", "start", "xprtld");
        print join('', &get_ts(), " # ", $cmd, "\n");
        my $output = `$cmd 2>&1`;
        print $output;
    }
    else {
        system("/etc/init.d/xprtld start");
    }
}

# wf_stop_xprtld will stop xprtld
# This module is applicable only for CCS MS
# and CH.
sub wf_stop_xprtld {
    my $cc = shift or die "Context closure not provided\n";
    if ( $is_windows ) {
        my $cmd = join(' ', "net", "stop", "xprtld");
        print join('', &get_ts(), " # ", $cmd, "\n");
        my $output = `$cmd 2>&1`;
        print $output;
    }
    else {
        system("/etc/init.d/xprtld stop");
    }
}


########################################################
# VRTSweb start/stop is a pain, and on windows it is even
# more painful as the SCM adds its contribution...
# We have to start/stop dependent servers, so need to
# go through hoops to figure this out.
# Code contributed by David Elson from the SCM installer.
#
# Windows only.
#
# Build forward dependency adjacency matrix,
# based on dependencies specified in the
# registry.
#
# Implement this as a 2-level hash ref, where
# $adj->{X}{Y} exists iff Y is immediately
# dependent on X.
#
# Dependencies are specified in the registry
# as backward dependencies, thus each entry
# is put into the adjacency matrix as a
# predecessor, with the successor being the
# key currently under examination.
#
##################################################
sub win_enum_svcs {
	# initialize registry interface
	eval "use Win32API::Registry 0.21 qw( :ALL )";
	# this will be under HKEY_LOCAL_MACHINE
	my $path = "SYSTEM\\CurrentControlSet\\Services";
	# since we use the package dynamically, we need to eval these.
	my $key_read = eval "KEY_READ";
	my $hklm = eval "HKEY_LOCAL_MACHINE";
	# handle to services, in registry
	my $services_key;
	# get handle to services
	my $rok_rc = RegOpenKeyEx( $hklm, $path, 0, $key_read, $services_key );
	if (!$rok_rc) {
		# (:-(
		die "Can't open HKEY_LOCAL_MACHINE\\${path}: ", regLastError();
	}
	# (:-)
	# we'll have to walk registry to populate hashref ccs
	my $ccs = {};
	my $ix = 0;
	my $rc = 1;
	while ($rc) {
		# mostly dummies, since all we we use is osValName
		my ($osValName, $iolValName, $ouType, $opValData, $iolValData);
		$rc = RegEnumKeyEx($services_key, $ix, $osValName, $iolValName, [], $ouType, $opValData, $iolValData);
		if (!$rc) {
			last;
		}
		$ccs->{$osValName} = 1;
		$ix++;
	}
	RegCloseKey( $services_key );

	# Initialize search pattern.
	# Forward dependents might not start with VRTS, soooo ...
	#my $pattern = "^VRTS";
	# ... walk them all.
	my $pattern = "\\S";
	# start small (:-)
	my $adj = {};
	# now that we have services hash, we can build up adjacency graph
	for my $svc_key (sort keys %$ccs) {
		# only consider what we specify
		next if $svc_key !~ /${pattern}/;
		#print("${svc_key}\n");
		my $subpath = $path."\\${svc_key}";
		my $subkey;
		$rok_rc = RegOpenKeyEx($hklm, $subpath, 0, $key_read, $subkey);
		if (!$rok_rc) {
			# (:-(
			my $errmsg = "Unable to open subpath == ${subpath}.";
			die $errmsg;
		}
		# (:-)
		# now look for dependency values.
		my ($pType, $pData, $dataLen);
		my $rqv_rc = RegQueryValueEx($subkey, "DependOnService", [], $pType, $pData, $dataLen);
		# nonzero return code ==> success
		if ($rqv_rc) {
			# could be empty, but probably not ... why would anyone bother?
			if ($pData) {
				# got some.
				# they are multi-valued (sz_multi)
				# and, therefore, separated by char(0).
				# so, if we have:
				#   deps(y) = [x1,x2,x3]
				# then:
				#   x1 ==> y
				#   x2 ==> y
				#   x3 ==> y
				# which means that in order to shut down y,
				# we must first shut down x1, x2, and x3.
				my $deps = [split(/\0/, $pData)];
				for my $dep (@$deps) {
					next if $dep !~ /\S/;
					#print("  ${dep}\n");
					my $pred = $dep;
					$pred =~ s/^\s+|\s+$//g;
					my $succ = $svc_key;
					$succ =~ s/^\s+|\s+$//g;
					$succ =~ s/\///;

					if (! exists $adj->{$pred}) {
						$adj->{$pred} = {};
					}
					$adj->{$pred}{$succ} = 1;
				}
			}
		}

		RegCloseKey($subkey);
	}

	# make this (1) to print out adjacencies.
	if (0) {
		print("\nDependency Adjacency for VRTS Services:\n");
		print("--------------------------------------\n");
		for my $pred (sort keys %$adj) {
			my $succs = $adj->{$pred};
			for my $succ (sort keys (%$succs)) {
				print("${pred} ==> ${succ}\n");
			}
		}
	}

	return $adj;
}

sub win_compute_depends {
    my ($svc_adj, $svc_name) = @_;
    my $deps_href = $svc_adj->{$svc_name};
    my $deps_aref = [sort keys %$deps_href];

    return $deps_aref;
}

sub win_compute_adj_graph {
    my ($svc_adj, $svc1) = @_;
    my $ag = {};
    my $todo = {
	aref => [$svc1],
	href => {
	    $svc1 => 1,
	},
    };
    my $todo_aref = $todo->{aref};
    my $todo_href = $todo->{href};

    while (scalar(@$todo_aref)) {
	my $svc_name = pop(@$todo_aref);
	delete $todo_href->{$svc_name};
	
	$ag->{$svc_name} = {};
	
	my $deps = win_compute_depends($svc_adj, $svc_name);
	
	if (!$deps) {
	    return undef;
	}
	
	for my $dep (@$deps) {
	    #print("  ${dep}\n");
	    if (!$todo_href->{$dep} && !$ag->{$svc_name}{$dep}) {
		push(@$todo_aref, $dep);
		$todo_href->{$dep} = 1;
	    }
	    $ag->{$svc_name}{$dep} = 1;
	}
    }

    return $ag;
}

sub win_analyze_adj_graph {
    my ($ag) = @_;
    my $order = [];
    my $todo_href = {};

    # initially, must process all names
    for my $key (sort(keys(%$ag))) {
	$todo_href->{$key} = 1;
    }

    # initialize to get into loop
    my $changed = 1;

    while ($changed) {
	# only test what changed during this pass
	$changed = 0;
	my $keys_done = [];
	
	# keeps separate array for clean delete
	my $keys_todo = [sort(keys(%$todo_href))];
	for my $key (@$keys_todo) {
	    # compute number of (unprocessed) dependents.
	    # note: first time through will be names
	    # having no dependents to begin with.
	    my $eff_num_deps = 0;
	    my $dep_href = $ag->{$key};
	    for my $dep_key (sort(keys(%$dep_href))) {
		if ($todo_href->{$dep_key}) {
		    $eff_num_deps++;
		}
	    }
	
	    if (!$eff_num_deps) {
		#print("${key} has no unprocessed dependents.\n");
		# partial ordering
		push(@$order, $key);
		delete $todo_href->{$key};
		$changed = 1;
	    }
	}
    }

    return $order;
}

# windows-specific "compute_stop_order()"
sub win_compute_stop_order {
    my ($svc1) = @_;

    # deps
    my $svc_adj = win_enum_svcs();

    my $ag = win_compute_adj_graph($svc_adj, $svc1);
    my $order = undef;

    if ($ag) {
	$order = win_analyze_adj_graph($ag);
    }

    return $order;
}

# Passed a service name, returns a reference to
# an array containing a stop order for the
# specified service, taking into account
# dependencies.
sub compute_svc_stop_order {
    my ($svc1) = @_;
    my $stop_order_aref = undef;
    if ($is_windows) {
	$stop_order_aref = win_compute_stop_order($svc1);
    } else {
	# unix
	$stop_order_aref = [$svc1];
    }

    return $stop_order_aref;
}

########################################################
# Hotfix specific extensions will be inserted below this line

